#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/moduleparam.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/cdev.h>
#include <linux/string.h>
#include <linux/list.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/sched.h>

#include <asm/uaccess.h>

#include "pacdmp-types.h"
#include "pacdmp-regs2.h"
#include "pacdmp-pci.h"

#define GET_FREE_PAGES    (10)
#define MAX_BUF_SIZE      (640*1024*4)

#ifdef __arm__

#define CARD_MAX_BUF_SIZE (640*1024*4)
#define PACDMP_MAJOR    240
#define PACDMP_MINOR    0
#define PACDMP_MAX_NUM  1
#define __W90X900__
#else

#define CARD_MAX_BUF_SIZE (640*1024*4 + 704*576*2*12)
#define PACDMP_MAJOR    0
#define PACDMP_MINOR    0
#define PACDMP_MAX_NUM  1
#endif

#ifdef __W90X900__
#define IRQ_FLAGS       SA_INTERRUPT
#elif defined(IRQF_SHARED)
#define IRQ_FLAGS       IRQF_SHARED
#else
#define IRQ_FLAGS       SA_SHIRQ
#endif

/*********************************************************\
 * Global data
 \*********************************************************/
static struct pacdmp_dev g_pacdmps[PACDMP_MAX_NUM];
static unsigned int g_pacdmp_num;
static int g_pacdmp_major = PACDMP_MAJOR;
static int g_pacdmp_minor = PACDMP_MINOR;

static int irq_debug = 0;
static int pacdmp_debug = 0;
#ifdef __W90X900__
static unsigned int pacdmp_reserved_addr = 61*1024*1024;
#else
static unsigned int pacdmp_reserved_addr = 0;
#endif

struct file_info {
    struct file_info *next;

    struct file *filp;
    struct pacdmp_dev *dmp_dev;

    int streamon;
    int wakeup;
    wait_queue_head_t wait_queue;

    int live_streamon;
    int live_wakeup;
    wait_queue_head_t live_wait_queue;

    int joined_list;
};
static struct file_info *g_list[PACDMP_MAX_NUM];

struct device_info {
    int status;
    TIME_CTRL time_ctrl;
    int standard[4];
    int resolution[4];
    int preview_enable[4];
    int encoder_enable[4];
    int audio_enable[4];
    int vlost_enable[4];
    RATE_CTRL rate_ctrl[4];
    GOP gop[4];
    MDECT mdect[4];
    int mdect_enable[4];
    int fps[4];
    COLOR color[4];
    OSD osd[4];
    MASK mask[4];
    POSITION position[4];
};
static struct device_info g_info[PACDMP_MAX_NUM];
static struct mutex g_info_lock[PACDMP_MAX_NUM];

/*********************************************************\
 *  insmod args
 \*********************************************************/
module_param( irq_debug,            int,  S_IRUGO );
module_param( pacdmp_debug,         int,  S_IRUGO );
module_param( pacdmp_reserved_addr, uint, S_IRUGO );

MODULE_PARM_DESC( irq_debug, "irq handler debug messages, default is 0 (no)" );
MODULE_PARM_DESC( pacdmp_debug, "debug messages, default is 0 (no)" );
MODULE_PARM_DESC( pacdmp_reserved_addr, "reserved address, default is 0" );

static void pacdmp_write_bar3( struct pacdmp_dev *dmp_dev, u32 address )
{
    unsigned long tflag;

    spin_lock_irqsave(&dmp_dev->reg_lock, tflag);
    pci_write_config_dword( dmp_dev->core.pci, PCI_BASE_ADDRESS_3, address );
    spin_unlock_irqrestore( &dmp_dev->reg_lock, tflag );
}

static void pacdmp_read_bar3( struct pacdmp_dev *dmp_dev, u32 *address )
{
    unsigned long tflag;

    spin_lock_irqsave(&dmp_dev->reg_lock, tflag);
    pci_read_config_dword( dmp_dev->core.pci, PCI_BASE_ADDRESS_3, address );
    spin_unlock_irqrestore( &dmp_dev->reg_lock, tflag );
}

static inline u32 mmio_write32( void __iomem *ioaddr, u32 offset, u32 value )
{
    u32 retval = -1;

    // offset: byte unit.
    iowrite32( value, ioaddr + offset );
    retval = ioread32( ioaddr + offset ); // handle PCI write posting problem
    udelay(1);
    return retval;
}

static inline u32 mmio_read32( void __iomem *ioaddr, u32 offset )
{
    u32 value = -1;

    // offset: byte unit
    value = ioread32( ioaddr + offset );
    return value;
}

static inline u16 mmio_write16( void __iomem *ioaddr, u32 offset, u16 value )
{
    u16 retval = -1;

    iowrite16( value, ioaddr + offset );
    retval = ioread16( ioaddr + offset );
    udelay(1);
    return retval;
}

static inline u16 mmio_read16( void __iomem *ioaddr, u32 offset )
{
    u16 value = -1;

    value = ioread16( ioaddr + offset );
    return value;
}

static inline u8 mmio_write8( void __iomem *ioaddr, u32 offset, u8 value )
{
    u8 retval = -1;

    iowrite8( value, ioaddr + offset );
    retval = ioread8( ioaddr + offset );
    udelay(1);
    return retval;
}

static inline u8 mmio_read8( void __iomem *ioaddr, u32 offset )
{
    u8 value = -1;

    value = ioread8( ioaddr + offset );
    return value;
}

static struct pci_device_id pacdmp_pci_tbl[] = {
    { PCI_DEVICE(PACDMP_PCI_VENDOR_ID, PACDMP_PCI_DEVICE_ID) },
    { 0,},
};

MODULE_DEVICE_TABLE(pci, pacdmp_pci_tbl);

MODULE_AUTHOR("Sunplus CT modell.wang/shengguo.tan");
MODULE_DESCRIPTION("Sunplus CT PACDMP PCI Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(PACDMP_PCI_VERSION);

#ifdef __arm__
static irqreturn_t pacdmp_irq(int irq, void * dev_id, struct pt_regs *regs)
#else
static irqreturn_t pacdmp_irq(int irq, void * dev_id)
#endif
{
    int handled = 0;
    struct pacdmp_dev * dmp_dev = dev_id;
    int card_id = dmp_dev->core.nr;
    void __iomem * mmio_vaddr = dmp_dev->mmio_vaddr;
    volatile u32 vic1st = 0, vic2st = 0;
    u32 tmask1 = 0, tmask2 = 0, tbase = 0;
    u32 DMAFlag = 0, TimerFlag = 0, VPUFlag = 0;
    u32 tmp_bar3 = 0;
    u32 tflag;
    u8 *taddr;
    u32 tlen, tnum;
    u32 towner, tpid;
    u32 channel;
    u32 buf_ptr, buf_usg;
    struct file_info *info;

    spin_lock_irqsave( &dmp_dev->irq_lock, tflag );

    pacdmp_read_bar3( dmp_dev, &tmp_bar3 ); // save old BAR3 in tmp_bar3

    pacdmp_write_bar3( dmp_dev, PACDMP_EVB_HST_VIC1 );
    vic1st = mmio_read32( mmio_vaddr, 0 ); // Read Host VIC1 status
    vic2st = mmio_read32( mmio_vaddr, 0x1000 ); // Read Host VIC2 status

    if ( vic1st == 0x0 && vic2st == 0x0 ) {
        // not our interrupt, ignoring...
        handled = 0;
        goto irq_end;
    }

    // ISR body

    tmask1 = mmio_read32( mmio_vaddr, VIC_INT_ENABLE );
    tmask2 = mmio_read32( mmio_vaddr, VIC_INT_ENABLE + 0x1000 );
    // Handler interrupt issued by PACDMP
    if ( irq_debug )
        printk(KERN_DEBUG "%s: received interrupt.[VIC Lv1/Mask = 0x%08X/0x%08X]"
               "[VIC Lv2/Mask = 0x%08X/0x%08X]\n",
                dmp_dev->core.name, vic1st, tmask1, vic2st, tmask2);
    handled = 1;

    // VIC1
    if ((vic1st & 0x000FFFFF) != 0x0) { // Video port
        int prepare_live_data = 0;

        VPUFlag = vic1st & 0x000FFFFF;
        if ( irq_debug )
            printk(KERN_DEBUG "%s: received Video port 0 interrupt.\n", dmp_dev->core.name);

        dmp_dev->live_streamon = 0;
        dmp_dev->live_wakeup = 0;
        for (info = g_list[card_id]; info != NULL; info = info->next) {
            if (info->live_streamon) {
                dmp_dev->live_streamon = 1;
                if (info->live_wakeup) {
                    dmp_dev->live_wakeup = 1;
                }
            }
        }
        if (dmp_dev->live_streamon && !dmp_dev->live_wakeup) {
            dmp_dev->live_data = 0;
            prepare_live_data = 1;
        }

        tbase = 0x1E131000;
        pacdmp_write_bar3( dmp_dev, tbase );
        for (channel = 0; channel < 4; channel++) {
            if (VPUFlag & (0x1 << (channel*5))) {
                mmio_write32( mmio_vaddr, 0x1000*channel + 0x1C, 0x0 );
                if (prepare_live_data) {
                    buf_ptr = mmio_read32( mmio_vaddr, 0x1000*channel + 0x58 );
                    buf_usg = (buf_ptr == 0x1) ? 0x4 : (buf_ptr/2);
                    mmio_write32( mmio_vaddr, 0x1000*channel + 0x5C, buf_usg );
                    dmp_dev->live_data |= 1 << (3*channel + buf_usg/2);
                }
            }
            if (VPUFlag & (0x8 << (channel*5))) {
                printk(KERN_INFO "%s: received CH-%d hourly Interrupt.\n", dmp_dev->core.name, channel);
                mmio_write32( mmio_vaddr, 0x1000*channel + 0x7000 + 0x140, 0x0 );
            }
            if (VPUFlag & (0x10 << (channel*5))) {
                printk(KERN_INFO "%s: received CH-%d daily Interrupt.\n", dmp_dev->core.name, channel);
                mmio_write32( mmio_vaddr, 0x1000*channel + 0x7000 + 0x160, 0x0 );
            }
        }
        if (prepare_live_data && dmp_dev->live_data) {
            for (info = g_list[card_id]; info != NULL; info = info->next) {
                if (info->live_streamon) {
                    info->live_wakeup = 1;
                    wake_up_interruptible(&info->live_wait_queue);
                }
            }
        }
    } else if ((vic1st & 0x01000000) != 0x0) { // DSP Interrupt
        if ( irq_debug )
            printk(KERN_ERR "%s: received DSP Interrupt.\n", dmp_dev->core.name);
        tbase = 0x1E050000;
        pacdmp_write_bar3( dmp_dev, tbase );
        mmio_write32( mmio_vaddr, 0xC , 0x1 );
    }

    // VIC2
    if ((vic2st & 0x0000FFFF) != 0x0) { // System DMA
        DMAFlag = 1;
        tbase = PACDMP_EVB_SYSDMA;
        while ((vic2st & 0x1) == 0x0) {
            tbase += 0x20;
            vic2st >>= 1;
        }

        /* DSP Stub Test Area */
        if (tbase == 0x1E1101C0) {
            if ( dmp_dev->hst_len == 0x0 ) {
                pacdmp_write_bar3( dmp_dev, 0x01C80000 );
                tlen = mmio_read32( mmio_vaddr, 0x1C );
                dmp_dev->hst_len = tlen;
            } else {
                tlen = dmp_dev->hst_len;
            }

            if ( dmp_dev->hst_num == 0x0 ) {
                pacdmp_write_bar3( dmp_dev, 0x01C80000 );
                tnum = mmio_read32( mmio_vaddr, 0x18 );
                dmp_dev->hst_num = tnum;
            } else {
                tnum = dmp_dev->hst_num;
            }

            dmp_dev->streamon = 0;
            dmp_dev->wakeup = 0;
            for (info = g_list[card_id]; info != NULL; info = info->next) {
                if (info->streamon) {
                    dmp_dev->streamon = 1;
                    if (info->wakeup) {
                        dmp_dev->wakeup = 1;
                    }
                }
            }

            while (1) {
                taddr = dmp_dev->master_buf + (tlen * dmp_dev->hst_rptr);
                towner = *(u16 *)taddr;
                if (towner != 0x86) break;

                tpid = *(u16 *)(taddr + 4);

                if (dmp_dev->streamon) {
                    if (!dmp_dev->wakeup) {
                        for (info = g_list[card_id]; info != NULL; info = info->next) {
                            if (info->streamon) {
                                info->wakeup = 1;
                                wake_up_interruptible(&info->wait_queue);
                            }
                        }
                    } else {
#ifndef __arm__
                        printk("[%s]IRQ waiting user qbuf: %d......\n",
                                dmp_dev->core.name, ++dmp_dev->delay_count);
#endif
                    }
                    break;
                }
                *(u16 *)taddr = 0x61;
                dmp_dev->hst_rptr = (dmp_dev->hst_rptr + 1) % tnum;
            }
        }
        /* DSP Stub Test Area End */

        pacdmp_write_bar3( dmp_dev, tbase );
        mmio_write32( mmio_vaddr, 0x18, 0x1 );
    } else if ((vic2st & 0x003E0000) != 0x0) { // DSP DMA
        DMAFlag = 1;
        tbase = 0x1E050070;
        vic2st >>= 17;
        while ((vic2st & 0x1) == 0x0) {
            tbase += 0x20;
            vic2st >>= 1;
        }
        pacdmp_write_bar3( dmp_dev, tbase );
        mmio_write32( mmio_vaddr, 0x18, 0x1 );
    } else if ((vic2st & 0x7E000000) != 0x0) { // Timer
        TimerFlag = 1;
        tbase = PACDMP_EVB_TIMER0;
        vic2st >>= 25;
        while ((vic2st & 0x1) == 0x0) {
            tbase += 0x1000;
            vic2st >>= 1;
        }
        pacdmp_write_bar3( dmp_dev, tbase );
        mmio_write32( mmio_vaddr, 0x0, 0x04000000);
    }

    if (DMAFlag) dmp_dev->irq_flag |= 0xde11 | ( tbase << 16 );
    if (TimerFlag) dmp_dev->irq_flag |= 0xde11 | ( tbase << 16 );
    if (VPUFlag) dmp_dev->irq_flag |= 0xde11 | ( vic1st << 16 );

irq_end:

    pacdmp_write_bar3( dmp_dev, tmp_bar3 ); // restore old BAR3

    /* irq statistics */
    dmp_dev->irq_total++;
    if ( handled )
        dmp_dev->irq_self++;

    if ( irq_debug )
        printk( KERN_DEBUG "%s: ISR statistic(self:total)--> %u:%u \n", dmp_dev->core.name, dmp_dev->irq_self, dmp_dev->irq_total );

    spin_unlock_irqrestore( &dmp_dev->irq_lock, tflag );

    return IRQ_RETVAL(handled);
}

static int pacdmp_pci_open(struct inode *inode, struct file *filp)
{
    struct file_info *finfo;
    struct pacdmp_dev *dmp_dev;

    dmp_dev = container_of( inode->i_cdev, struct pacdmp_dev, cdev );
    finfo = (struct file_info *)vmalloc( sizeof(struct file_info) );
    if (finfo == NULL) {
        return -ENOMEM;
    }
    finfo->filp = filp;
    finfo->dmp_dev = dmp_dev;
    finfo->streamon = 0;
    finfo->wakeup = 0;
    init_waitqueue_head( &finfo->wait_queue );
    finfo->live_streamon = 0;
    finfo->live_wakeup = 0;
    init_waitqueue_head( &finfo->live_wait_queue );
    finfo->joined_list = 0;
    finfo->next = NULL;
    filp->private_data = finfo;

    dmp_dev->users++;

    printk( "%s: opened (%u)\n", dmp_dev->core.name, dmp_dev->users );
    return 0;
}

static int pacdmp_pci_release( struct inode *inode, struct file *filp )
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int card_id = dmp_dev->core.nr;
    struct file_info *info, *prev;

    finfo->streamon = 0;
    finfo->wakeup = 0;
    finfo->live_streamon = 0;
    finfo->live_wakeup = 0;

    if (finfo->joined_list) {
        mutex_lock( &dmp_dev->file_lock );
        prev = NULL;
        for (info = g_list[card_id]; info != NULL; prev = info, info = info->next) {
            if (info == finfo) {
                break;
            }
        }
        if (info == NULL) {
            printk( KERN_ERR "error: file info joined list, but not found!\n" );
        } if (prev == NULL) {
            g_list[card_id] = finfo->next;
        } else {
            prev->next = finfo->next;
        }
        mutex_unlock( &dmp_dev->file_lock );
    }
    filp->private_data = NULL;
    vfree( finfo );

    /* free resources */
    dmp_dev->users--;

    printk( "%s: released (%u)\n", dmp_dev->core.name, dmp_dev->users );

    return 0;
}

static int pacdmp_pci_mmap (struct file *filp, struct vm_area_struct *vma)
{
    vma->vm_flags |= VM_RESERVED;
#ifdef __arm__
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif

    if (remap_pfn_range(vma,
                vma->vm_start,
                vma->vm_pgoff,
                vma->vm_end - vma->vm_start,
                vma->vm_page_prot))
    {
        printk("mmap error occured\n");
        return -EAGAIN;
    }
    vma->vm_ops = NULL;
    return 0;
}

static unsigned int pacdmp_pci_poll (struct file *filp, poll_table *wait)
{
    struct file_info *finfo = filp->private_data;
    unsigned int mask = 0;

    poll_wait(filp, &(finfo->wait_queue), wait);

    if (finfo->wakeup) {
        mask |= POLLIN | POLLRDNORM;
    }
    return mask;
}

static ssize_t pacdmp_pci_write(struct file *filp, const char *bp, size_t count, loff_t *ppos)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int ret = 0;
    char * tbp;

    if (count > MAX_BUF_SIZE) return -1;

    tbp = memcpy(dmp_dev->master_buf, bp, count);

    return ret = ((u8 *)tbp == dmp_dev->master_buf) ? count : -1;
}

static ssize_t pacdmp_pci_read(struct file *filp, char *bp, size_t count, loff_t *ppos)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int ret = 0;
    char * tbp;

    if (count > MAX_BUF_SIZE) return -1;

    tbp = memcpy(bp, dmp_dev->master_buf, count);

    return ret = (tbp == bp) ? count : -1;
}

static int pacdmp_streamon(struct file *filp)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int card_id = dmp_dev->core.nr;

    mutex_lock( &dmp_dev->file_lock );
    if (!finfo->joined_list) {
        finfo->joined_list = 1;
        finfo->next = g_list[card_id];
        g_list[card_id] = finfo;
    }
    mutex_unlock( &dmp_dev->file_lock );
    finfo->streamon = 1;
    return 0;
}

static int pacdmp_streamoff(struct file *filp)
{
    struct file_info *finfo = filp->private_data;
    finfo->streamon = 0;
    return 0;
}

static int pacdmp_dqbuf(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev;
    int retval;
    u32 data32;

    if (!finfo->streamon)
        return -EACCES;

    if (finfo->wakeup == 0) {
        if (filp->f_flags & O_NONBLOCK) {
            return -EAGAIN;
        }

        retval = wait_event_interruptible(finfo->wait_queue, (finfo->wakeup != 0));
        if (retval < 0) {
            return retval;
        }
    }
    dmp_dev = finfo->dmp_dev;
    data32 = dmp_dev->hst_rptr;
    if (copy_to_user((void __user *) arg, (u32 *) &data32, sizeof(u32)))
        return -EFAULT;
    return 0;
}

static int pacdmp_qbuf(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev= finfo->dmp_dev;
    int card_id = dmp_dev->core.nr;
    int wakeup_count;
    u8 *taddr;
    u32 tlen, tnum;
    struct file_info *info;

    if (!finfo->streamon)
        return -EACCES;

    if (!finfo->wakeup)
        return 0;

    wakeup_count = 0;
    mutex_lock( &dmp_dev->file_lock );
    for (info = g_list[card_id]; info != NULL; info = info->next) {
        if (info->streamon && info->wakeup) {
            wakeup_count++;
        }
    }
    mutex_unlock( &dmp_dev->file_lock );

    if (wakeup_count == 1) {
        tlen = dmp_dev->hst_len;
        tnum = dmp_dev->hst_num;
        taddr = dmp_dev->master_buf + (tlen * dmp_dev->hst_rptr);
        *(u16 *)taddr = 0x61;
        dmp_dev->hst_rptr = (dmp_dev->hst_rptr + 1) % tnum;
    }
    finfo->wakeup = 0;
    return 0;
}

static int pacdmp_live_streamon(struct file *filp)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int card_id = dmp_dev->core.nr;

    mutex_lock( &dmp_dev->file_lock );
    if (!finfo->joined_list) {
        finfo->joined_list = 1;
        finfo->next = g_list[card_id];
        g_list[card_id] = finfo;
    }
    mutex_unlock( &dmp_dev->file_lock );
    finfo->live_streamon = 1;
    return 0;
}

static int pacdmp_live_streamoff(struct file *filp)
{
    struct file_info *finfo = filp->private_data;
    finfo->live_streamon = 0;
    return 0;
}

static int pacdmp_live_dqbuf(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev;
    int retval;
    u32 data32;

    if (!finfo->live_streamon)
        return -EACCES;

    if (finfo->live_wakeup == 0) {
        if (filp->f_flags & O_NONBLOCK) {
            return -EAGAIN;
        }

        retval = wait_event_interruptible(finfo->live_wait_queue, (finfo->live_wakeup != 0));
        if (retval < 0) {
            return retval;
        }
    }

    dmp_dev = finfo->dmp_dev;
    data32 = dmp_dev->live_data;
    if (copy_to_user((void __user *) arg, (u32 *) &data32, sizeof(u32)))
        return -EFAULT;
    return 0;
}

static int pacdmp_live_qbuf(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;

    if (!finfo->live_streamon)
        return -EACCES;

    finfo->live_wakeup = 0;
    return 0;
}

static int pacdmp_init_queue(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    QUEUE queue;

    if ( copy_from_user( &queue, (void __user *) arg, sizeof( queue ) ) )
        return -EFAULT;

    mutex_lock( &dmp_dev->queue_lock );
    dmp_dev->cmdque_base = queue.base;
    dmp_dev->cmdque_len = queue.len;
    dmp_dev->cmdque_num = queue.num;
    dmp_dev->cmdque_wptr = 0;
    dmp_dev->hst_rptr = 0;
    dmp_dev->hst_len = 0;
    dmp_dev->hst_num = 0;
    mutex_unlock( &dmp_dev->queue_lock );
    return 0;
}

static void init_all_properties(int card_id)
{
    int channel;

    g_info[card_id].status = 0;

    g_info[card_id].time_ctrl.ctrl_type = 1;
    g_info[card_id].time_ctrl.time = 0;

    for (channel = 0; channel < 4; channel++) {
        g_info[card_id].standard[channel] = DEFAULT_STANDARD;
        g_info[card_id].resolution[channel] = DEFAULT_RESOLUTION;
        g_info[card_id].preview_enable[channel] = 1;
        g_info[card_id].encoder_enable[channel] = 1;
        g_info[card_id].audio_enable[channel] = 1;
        g_info[card_id].vlost_enable[channel] = 1;

        g_info[card_id].rate_ctrl[channel].ctrl_type = 2;
        g_info[card_id].rate_ctrl[channel].qpi = 29;
        g_info[card_id].rate_ctrl[channel].qpp = 30;
        g_info[card_id].rate_ctrl[channel].qpb = 31;

        g_info[card_id].gop[channel].gopp = 30;
        g_info[card_id].gop[channel].gopb = 2;

        g_info[card_id].mdect[channel].mthreshold = 0x2000;
        g_info[card_id].mdect[channel].minterval = 2;
        g_info[card_id].mdect_enable[channel] = 1;

        g_info[card_id].fps[channel] = (DEFAULT_RESOLUTION == STD_PAL) ? 25 : 30;

        g_info[card_id].color[channel].hue = 0x80;
        g_info[card_id].color[channel].contrast = 0x80;
        g_info[card_id].color[channel].saturation = 0x80;
        g_info[card_id].color[channel].brightness = 0x80;

        g_info[card_id].osd[channel].color_index = 1;
        g_info[card_id].osd[channel].osd0_enable = 1;
        g_info[card_id].osd[channel].osd0_x = 0x20;
        g_info[card_id].osd[channel].osd0_y = 0x20;
        strcpy(g_info[card_id].osd[channel].osd0_str,
                "\217\216\215\214/\213\212/\211\210 \205\204:\203\202:\201\200");
        g_info[card_id].osd[channel].osd1_enable = 0;
        g_info[card_id].osd[channel].osd1_x = 0;
        g_info[card_id].osd[channel].osd1_y = 0;
        g_info[card_id].osd[channel].osd1_str[0] = '\0';

        memset(&g_info[card_id].mask, 0, sizeof(g_info[card_id].mask));
        memset(&g_info[card_id].position, 0, sizeof(g_info[card_id].position));
    }
}

static int pacdmp_set_property(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int card_id = dmp_dev->core.nr;
    PACDMP_PROPERTY dmp_prop;

    if( copy_from_user( &dmp_prop, (void __user *) arg, sizeof( dmp_prop ) ) ) {
        printk("%s: line %d error\n", __FUNCTION__, __LINE__);
        return -EFAULT;
    }

    if( dmp_prop.buffer ) {
        unsigned char *buffer;
        unsigned int write_addr, write_len, offset;
        unsigned int cmdque_full;
        u32 data32;

        if( dmp_prop.len < 16 || dmp_prop.len > dmp_dev->cmdque_len ) {
            printk("%s: line %d error\n", __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        write_len = (dmp_prop.len + 3) & (~3);
        buffer = vmalloc( write_len );
        if( buffer == NULL ) {
            printk("%s: line %d error\n", __FUNCTION__, __LINE__);
            return -ENOMEM;
        }

        if( copy_from_user( buffer, (void __user *) dmp_prop.buffer, dmp_prop.len ) ) {
            vfree(buffer);
            printk("%s: line %d error\n", __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        if( dmp_prop.retry > 100 ) {
            printk( KERN_INFO "dmp_prop retry(%d) > 100, reset it to 100\n", dmp_prop.retry );
            dmp_prop.retry = 100;
        }

        mutex_lock( &dmp_dev->queue_lock );
        write_addr = dmp_dev->cmdque_base + dmp_dev->cmdque_wptr * dmp_dev->cmdque_len;
        pacdmp_write_bar3( dmp_dev, write_addr );

        cmdque_full = 1;
        do {
            data32 = mmio_read32( dmp_dev->mmio_vaddr, 0 );
            if( ((u16)data32)  == 0x86 ) {
                cmdque_full = 0;
                break;
            }
            msleep(50);
        } while( dmp_prop.retry-- > 0 );

        if( cmdque_full ) {
            mutex_unlock( &dmp_dev->queue_lock );
            vfree(buffer);
            printk( KERN_ERR "cmdque full\n" );
            return -EFAULT;
        }

        *(u16 *)buffer = 0x86;
        for( offset = 0; offset < write_len; offset += 4 ) {
            mmio_write32( dmp_dev->mmio_vaddr, offset, *(u32 *)(buffer + offset) );
        }
        *(u16 *)buffer = 0x61;
        mmio_write32( dmp_dev->mmio_vaddr, 0, *(u32 *)buffer );
        dmp_dev->cmdque_wptr = (dmp_dev->cmdque_wptr + 1) % dmp_dev->cmdque_num;
        mutex_unlock( &dmp_dev->queue_lock );
        vfree(buffer);
    }

    if( dmp_prop.prop ) {
        struct spct6100_property prop;
        int type, channel;

        if( copy_from_user( &prop, (void __user *) dmp_prop.prop, sizeof( prop ) ) ) {
            printk("%s: line %d error\n", __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        type = prop.type;
        channel = prop.channel;

        mutex_lock(&g_info_lock[card_id]);
        switch(type) {
            case -1:
                if (g_info[card_id].status && prop.context.value) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].status = prop.context.value;
                if (g_info[card_id].status == 0) {
                    init_all_properties(card_id);
                }
                break;

            case CARD_TIME_CTRL:
                g_info[card_id].time_ctrl = prop.context.time_ctrl;
                break;

            case CHANNEL_STANDARD:
                if (channel < 0 || channel > 3) {
                    return -EFAULT;
                }
                if (g_info[card_id].status) {
                    if (g_info[card_id].standard[channel] != prop.context.value) {
                        mutex_unlock(&g_info_lock[card_id]);
                        return -EFAULT;
                    }
                }
                g_info[card_id].standard[channel] = prop.context.value;
                break;

            case CHANNEL_RESOLUTION:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                if (g_info[card_id].status) {
                    if (g_info[card_id].resolution[channel] != prop.context.value) {
                        mutex_unlock(&g_info_lock[card_id]);
                        return -EFAULT;
                    }
                }
                g_info[card_id].resolution[channel] = prop.context.value;
                break;

            case CHANNEL_PREVIEW_ENABLE:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                if (g_info[card_id].status) {
                    if (g_info[card_id].preview_enable[channel] != prop.context.value) {
                        mutex_unlock(&g_info_lock[card_id]);
                        return -EFAULT;
                    }
                }
                g_info[card_id].preview_enable[channel] = prop.context.value;
                break;

            case CHANNEL_ENCODER_ENABLE:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                if (g_info[card_id].status) {
                    if (g_info[card_id].encoder_enable[channel] != prop.context.value) {
                        mutex_unlock(&g_info_lock[card_id]);
                        return -EFAULT;
                    }
                }
                g_info[card_id].encoder_enable[channel] = prop.context.value;
                break;

            case CHANNEL_AUDIO_ENABLE:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                if (g_info[card_id].status) {
                    if (g_info[card_id].audio_enable[channel] != prop.context.value) {
                        mutex_unlock(&g_info_lock[card_id]);
                        return -EFAULT;
                    }
                }
                g_info[card_id].audio_enable[channel] = prop.context.value;
                break;

            case CHANNEL_VLOST_ENABLE:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                if (g_info[card_id].status) {
                    if (g_info[card_id].vlost_enable[channel] != prop.context.value) {
                        mutex_unlock(&g_info_lock[card_id]);
                        return -EFAULT;
                    }
                }
                g_info[card_id].vlost_enable[channel] = prop.context.value;
                break;

            case CHANNEL_RATE_CTRL:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].rate_ctrl[channel] = prop.context.rate_ctrl;
                break;

            case CHANNEL_GOP:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].gop[channel] = prop.context.gop;
                break;

            case CHANNEL_MDECT:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].mdect[channel] = prop.context.mdect;
                break;

            case CHANNEL_MDECT_ENABLE:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].mdect_enable[channel] = prop.context.value;
                break;

            case CHANNEL_FPS:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].fps[channel] = prop.context.value;
                break;

            case CHANNEL_COLOR:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].color[channel] = prop.context.color;
                break;

            case CHANNEL_OSD:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].osd[channel] = prop.context.osd;
                break;

            case CHANNEL_MASK:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].mask[channel] = prop.context.mask;
                break;

            case CHANNEL_POSITION:
                if (channel < 0 || channel > 3) {
                    mutex_unlock(&g_info_lock[card_id]);
                    return -EFAULT;
                }
                g_info[card_id].position[channel] = prop.context.position;
                break;

            default:
                mutex_unlock(&g_info_lock[card_id]);
                return -EFAULT;
        }
        mutex_unlock(&g_info_lock[card_id]);
    }
    return 0;
}

static int pacdmp_get_property(struct file *filp, unsigned long arg)
{
    struct file_info *finfo = filp->private_data;
    struct pacdmp_dev *dmp_dev = finfo->dmp_dev;
    int card_id = dmp_dev->core.nr;
    struct spct6100_property prop;
    int type, channel;

    if( copy_from_user( &prop, (void __user *) arg, sizeof( prop ) ) ) {
        printk("%s: line %d error\n", __FUNCTION__, __LINE__);
        return -EFAULT;
    }

    type = prop.type;
    channel = prop.channel;

    switch(type) {
        case -1:
            prop.context.value = g_info[card_id].status;
            break;

        case CARD_TIME_CTRL:
            prop.context.time_ctrl = g_info[card_id].time_ctrl;
            break;

        case CHANNEL_STANDARD:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].standard[channel];
            break;

        case CHANNEL_RESOLUTION:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].resolution[channel];
            break;

        case CHANNEL_PREVIEW_ENABLE:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].preview_enable[channel];
            break;

        case CHANNEL_ENCODER_ENABLE:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].encoder_enable[channel];
            break;

        case CHANNEL_AUDIO_ENABLE:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].audio_enable[channel];
            break;

        case CHANNEL_VLOST_ENABLE:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].vlost_enable[channel];
            break;

        case CHANNEL_RATE_CTRL:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.rate_ctrl = g_info[card_id].rate_ctrl[channel];
            break;

        case CHANNEL_GOP:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.gop = g_info[card_id].gop[channel];
            break;

        case CHANNEL_MDECT:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.mdect = g_info[card_id].mdect[channel];
            break;

        case CHANNEL_MDECT_ENABLE:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].mdect_enable[channel];
            break;

        case CHANNEL_FPS:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.value = g_info[card_id].fps[channel];
            break;

        case CHANNEL_COLOR:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.color = g_info[card_id].color[channel];
            break;

        case CHANNEL_OSD:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.osd = g_info[card_id].osd[channel];
            break;

        case CHANNEL_MASK:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.mask = g_info[card_id].mask[channel];
            break;

        case CHANNEL_POSITION:
            if (channel < 0 || channel > 3) {
                return -EFAULT;
            }
            prop.context.position = g_info[card_id].position[channel];
            break;

        default:
            return -EFAULT;
    }

    if( copy_to_user( (void __user *) arg, &prop, sizeof( prop ) ) ) {
        printk("%s: line %d error\n", __FUNCTION__, __LINE__);
        return -EFAULT;
    }
    return 0;
}

static int pacdmp_pci_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    u32 data32 = -1;

    struct pci_dev *pci_dev;
    struct file_info *finfo;
    struct pacdmp_dev *dmp_dev;
    REG32 reg32;
    REG16 reg16;
    REG8 reg8;

    finfo = filp->private_data;
    dmp_dev = finfo->dmp_dev;
    pci_dev = dmp_dev->core.pci;

    switch (cmd) {
        case PACDMP_CTL_SET_WINDOW_BASE:
            data32 = (u32) arg;
            if ( pacdmp_debug > 1 )
                printk( "%s (CTL_SET_WINDOW_BASE):\t(0x%08X)\n", dmp_dev->core.name, data32 );
            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, data32 );
            pacdmp_read_bar3( dmp_dev, &data32 );
            mutex_unlock( &dmp_dev->m_lock );
            return data32;

        case PACDMP_CTL_READ_REG32: // NOTE: BAR3 is modified after calling this .
            if ( copy_from_user( &reg32, (void __user *) arg, sizeof( REG32 ) ) )
                return -EFAULT;

            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, reg32.base );
            reg32.value = mmio_read32( dmp_dev->mmio_vaddr, reg32.offset );
            if ( pacdmp_debug > 1 )
                printk( "%s (CTL_READ_REG32):\t(0x%08X:0x%08X, 0x%08X)\n",
                        dmp_dev->core.name, reg32.base, reg32.offset, reg32.value );
            mutex_unlock( &dmp_dev->m_lock );

            if ( copy_to_user( (void __user *) arg, (REG32 *) &reg32, sizeof( REG32 ) ) )
                return -EFAULT;

            return 0;

        case PACDMP_CTL_WRITE_REG32: // NOTE: BAR3 is modified after calling this .
            if ( copy_from_user( &reg32, (void __user *) arg, sizeof( REG32 ) ) )
                return -EFAULT;

            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, reg32.base );
            mmio_write32( dmp_dev->mmio_vaddr, reg32.offset, reg32.value );
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "%s (CTL_WRITE_REG32):\t(0x%08X:0x%08X, 0x%08X)\n", dmp_dev->core.name, reg32.base, reg32.offset, reg32.value );
            mutex_unlock( &dmp_dev->m_lock );

            // FIXME: assume always write success !! (is this true ??)
            if ( copy_to_user( (void __user *) arg, (REG32 *) &reg32, sizeof( REG32 ) ) )
                return -EFAULT;

            return 0;

        case PACDMP_CTL_READ_REG16:

            if (copy_from_user(&reg16, (void __user *)arg, sizeof(REG16)))
                return -EFAULT;

            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, reg16.base );
            reg16.value = mmio_read16( dmp_dev->mmio_vaddr, reg16.offset );
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "%s (CTL_READ_REG16):\t(0x%08X:0x%08X, 0x%08X)\n", dmp_dev->core.name, reg16.base, reg16.offset, reg16.value );
            mutex_unlock( &dmp_dev->m_lock );

            if (copy_to_user( (void __user *) arg, (REG16 *) &reg16, sizeof(REG16) ))
                return -EFAULT;

            return 0;

        case PACDMP_CTL_WRITE_REG16:

            if (copy_from_user(&reg16, (void __user *)arg, sizeof(REG16)))
                return -EFAULT;

            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, reg16.base );
            mmio_write16(dmp_dev->mmio_vaddr, reg16.offset, reg16.value);
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "%s (CTL_WRITE_REG16):\t(0x%08X:0x%08X, 0x%08X)\n", dmp_dev->core.name, reg16.base, reg16.offset, reg16.value );
            mutex_unlock( &dmp_dev->m_lock );

            if (copy_to_user((void __user *)arg, (REG16 *)&reg16, sizeof(REG16)))
                return -EFAULT;

            return 0;

        case PACDMP_CTL_READ_REG8:

            if (copy_from_user(&reg8, (void __user *)arg, sizeof(REG8)))
                return -EFAULT;

            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, reg8.base );
            reg8.value = mmio_read8( dmp_dev->mmio_vaddr, reg8.offset );
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "%s (CTL_READ_REG8):\t(0x%08X:0x%08X, 0x%08X)\n", dmp_dev->core.name, reg8.base, reg8.offset, reg8.value );
            mutex_unlock( &dmp_dev->m_lock );

            if (copy_to_user( (void __user *) arg, (REG8 *) &reg8, sizeof(REG8) ))
                return -EFAULT;

            return 0;

        case PACDMP_CTL_WRITE_REG8:

            if (copy_from_user(&reg8, (void __user *)arg, sizeof(REG8)))
                return -EFAULT;

            mutex_lock( &dmp_dev->m_lock );
            pacdmp_write_bar3( dmp_dev, reg8.base );
            mmio_write8(dmp_dev->mmio_vaddr, reg8.offset, reg8.value);
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "%s (CTL_WRITE_REG8):\t(0x%08X:0x%08X, 0x%08X)\n", dmp_dev->core.name, reg8.base, reg8.offset, reg8.value );
            mutex_unlock( &dmp_dev->m_lock );

            if (copy_to_user((void __user *)arg, (REG8 *)&reg8, sizeof(REG8)))
                return -EFAULT;

            return 0;

        case PACDMP_CTL_GET_PHYBAR:

            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(CTL_GET_PHYBAR):\t\n" );
            mutex_lock( &dmp_dev->m_lock );
            data32 = pci_resource_start( pci_dev, 0 );
            mutex_unlock( &dmp_dev->m_lock );
            if ( copy_to_user( (void __user *) arg, (u32 *) &data32, sizeof( u32 ) ) )
                return -EFAULT;

            return 0;

        case PACDMP_CTL_GET_ISRFLG:

            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(CTL_GET_ISRFLG):\t\n" );
            spin_lock_irqsave( &dmp_dev->reg_lock, data32 );
            if ( copy_to_user( (void __user *) arg, (u32 *) &dmp_dev->irq_flag, sizeof( u32 ) ) ) {
                spin_unlock_irqrestore( &dmp_dev->reg_lock, data32 );
                return -EFAULT;
            }
            if ( ( dmp_dev->irq_flag & 0xFFFF ) == 0xde11 ) dmp_dev->irq_flag = 0x0;
            spin_unlock_irqrestore( &dmp_dev->reg_lock, data32 );

            return 0;

        case PACDMP_CTL_QUERYBUF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(CTL_QUERYBUF):\t\n" );
            data32 = dmp_dev->master_buf_phy;

            if ( copy_to_user( (void __user *) arg, (u32 *) &data32, sizeof( u32 ) ) )
                return -EFAULT;
            return 0;

        case PACDMP_CTL_STREAMON:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_STREAMON):\t\n" );
            return pacdmp_streamon( filp );

        case PACDMP_CTL_STREAMOFF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_STREAMOFF):\t\n" );
            return pacdmp_streamoff( filp );

        case PACDMP_CTL_QBUF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_QBUF):\t\n" );
            return pacdmp_qbuf( filp, arg );

        case PACDMP_CTL_DQBUF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_DQBUF):\t\n" );
            return pacdmp_dqbuf( filp, arg );

        case PACDMP_CTL_LIVE_STREAMON:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_LIVE_STREAMON):\t\n" );
            return pacdmp_live_streamon( filp );

        case PACDMP_CTL_LIVE_STREAMOFF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_LIVE_STREAMOFF):\t\n" );
            return pacdmp_live_streamoff( filp );

        case PACDMP_CTL_LIVE_QBUF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_LIVE_QBUF):\t\n" );
            return pacdmp_live_qbuf( filp, arg );

        case PACDMP_CTL_LIVE_DQBUF:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_LIVE_DQBUF):\t\n" );
            return pacdmp_live_dqbuf( filp, arg );

        case PACDMP_CTL_INIT_QUEUE:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_INIT_QUEUE):\t\n" );
            return pacdmp_init_queue( filp, arg );

        case PACDMP_CTL_SET_PROPERTY:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_SET_PROPERTY):\t\n" );
            return pacdmp_set_property( filp, arg );

        case PACDMP_CTL_GET_PROPERTY:
            if ( pacdmp_debug > 1 )
                printk( KERN_DEBUG "(PACDMP_CTL_GET_PROPERTY):\t\n" );
            return pacdmp_get_property( filp, arg );

        default:
            break;
    }

    return -ENOIOCTLCMD;
}

static struct file_operations pacdmp_pci_dev_ops = {
    .owner    = THIS_MODULE,
    .open     = pacdmp_pci_open,
    .release  = pacdmp_pci_release,
    .mmap     = pacdmp_pci_mmap,
    .poll     = pacdmp_pci_poll,
    .read     = pacdmp_pci_read,
    .write    = pacdmp_pci_write,
    .ioctl    = pacdmp_pci_ioctl,
};

static int __devinit pacdmp_pci_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
    int rc;
    struct pacdmp_dev *dmp_dev;
    char tmpFlag[64];
    unsigned long prnsize, resflag;
    u32 data32;
    u8 data8;
    dev_t devno;

    if ( g_pacdmp_num == PACDMP_MAX_NUM ) {
        rc = -ENOMEM;
        goto out;
    }

    /*********************************************************************************
     * FIXME: Ugly solution !!
     *    make pci_enable_device() happy !! Because of obscure info of BAR3, which
     *    will be complained during pcibios_init()... (dmesg for more information)
     *********************************************************************************/
#if 0
    for ( i = 1; i < 6; i++ ) {
        pdev->resource[i].start = pdev->resource[0].start;
        pdev->resource[i].end = pdev->resource[0].end;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
        pdev->resource[i].parent = pdev->resource[0].parent;
#endif
    }
#endif
    pdev->resource[3].flags &= ~(0x200); // hacking to prevent resource collisions of BAR3
    pdev->resource[4].flags &= ~(0x200); // hacking to prevent resource collisions of BAR4

    dmp_dev = &g_pacdmps[g_pacdmp_num];
    dmp_dev->core.nr = g_pacdmp_num;
    sprintf(dmp_dev->core.name, "%s%d(0x%x)", PACDMP_NAME, dmp_dev->core.nr, pdev->device);

    /* init dmp_dev resources and set to defaut values */
    spin_lock_init( &dmp_dev->irq_lock );
    spin_lock_init( &dmp_dev->reg_lock );
    mutex_init( &dmp_dev->m_lock );
    mutex_init( &dmp_dev->file_lock );
    mutex_init( &dmp_dev->queue_lock );
    mutex_init( &g_info_lock[g_pacdmp_num] );
    dmp_dev->hst_rptr = 0;
    dmp_dev->hst_len = 0;
    dmp_dev->hst_num = 0;

    /* pci stuffs (init, irq, mmio, bus mastering... */
    rc = pci_enable_device( pdev );
    if ( rc < 0 ) {
        printk( "%s: PCI enable failure\n", dmp_dev->core.name );
        rc = -EIO;
        goto out;
    }
    dmp_dev->core.pci = pdev;

    /**********************************************************************************
     *  Restore the hacking ....
     **********************************************************************************/
    pdev->resource[3].flags |= 0x200;
    pdev->resource[4].flags |= 0x200;
    /**********************************************************************************/

    /* print pci info */
    prnsize = 0;
    resflag = pci_resource_flags( pdev, 0 );

    if ( resflag & IORESOURCE_IO ) prnsize += sprintf( tmpFlag, "IO" );
    if ( resflag & IORESOURCE_MEM ) prnsize += sprintf( tmpFlag+prnsize, "|MEM" );
    if ( resflag & IORESOURCE_PREFETCH ) prnsize += sprintf( tmpFlag+prnsize, "|PREFETCH" );
    if ( resflag & IORESOURCE_READONLY ) prnsize += sprintf( tmpFlag+prnsize, "|READONLY" );
    printk( KERN_INFO "%s: found at %s, mmio(addr, size): 0x%08X, 0x%08X, resources(type, flag): %s, 0x%lX\n",
            dmp_dev->core.name, pci_name( pdev ), (unsigned int) pci_resource_start( pdev, 0 ),
            (unsigned int) pci_resource_len( pdev, 0 ), tmpFlag, resflag );

    /* PCI MMIO region (BAR0) */
    rc = pci_request_region( pdev, 0, dmp_dev->core.name );
    if ( rc < 0 )
    {
        dev_err( &pdev->dev, "%s: could not request PCI region.\n", dmp_dev->core.name );
        goto err_out_disable_pci;
    }

    dmp_dev->mmio_vaddr = pci_iomap( pdev, 0, 0 );
    if ( !dmp_dev->mmio_vaddr )
    {
        dev_err( &pdev->dev, "%s: cannot remap MMIO memory, aborting.\n", dmp_dev->core.name );
        rc = -EIO;
        goto err_out_release_res;
    }
    dmp_dev->mmio_size  = pci_resource_len( pdev, 0);
    printk("dmp_dev->mmio_size = 0x%08x\n", dmp_dev->mmio_size);

    /* initial MMIO base address to SDRAM start address on EVB */
    data32 = PACDMP_EVB_SDRAM_BAR;
    pacdmp_write_bar3( dmp_dev, data32 );
    data32 = 0x0;
    pacdmp_read_bar3( dmp_dev, &data32 );
    printk( "%s: init BAR_3 to EVB SDRAM base address (0x%08X)\n", dmp_dev->core.name, data32);


    // PCI Bus Mastering stuff...
    pci_set_master( pdev );
    if ( pci_set_dma_mask( pdev, DMA_32BIT_MASK ) ) {
        printk( "%s: Oops: no 32bit PCI DMA ???\n", dmp_dev->core.name );
        rc = -EIO;
        goto err_out_unmap;
    }

    if ( pacdmp_reserved_addr != 0 ) {
        dmp_dev->master_buf_phy = (dma_addr_t)(pacdmp_reserved_addr
                + CARD_MAX_BUF_SIZE * g_pacdmp_num);
        dmp_dev->master_buf = (u8 *)ioremap_nocache(dmp_dev->master_buf_phy, MAX_BUF_SIZE);
        if (dmp_dev->master_buf == NULL) {
            printk(KERN_ERR "%s: ioremap_nocache failure\n", dmp_dev->core.name);
            rc = -ENOMEM;
            goto err_out_unmap;
        }
        memset(dmp_dev->master_buf, 0xbe, MAX_BUF_SIZE);
    } else {
#ifdef __arm__
        printk(KERN_ERR "%s: error, pacdmp_reserved_addr is NULL\n", dmp_dev->core.name);
        rc = -ENOMEM;
        goto err_out_unmap;
#else
#ifdef GET_FREE_PAGES
        dmp_dev->master_free_pages = __get_free_pages(GFP_KERNEL|GFP_ATOMIC, GET_FREE_PAGES);
        if (dmp_dev->master_free_pages == 0) {
            printk(KERN_ERR "%s: Could not __get_free_pages\n", dmp_dev->core.name);
            rc = -ENOMEM;
            goto err_out_unmap;
        }
        dmp_dev->master_buf_phy = (dma_addr_t)__pa(dmp_dev->master_free_pages);
        dmp_dev->master_buf = (u8 *)dmp_dev->master_free_pages;
#else
        dmp_dev->master_buf = (u8 *)kzalloc(MAX_BUF_SIZE, GFP_ATOMIC);
        if (dmp_dev->master_buf == NULL) {
            printk(KERN_ERR "%s: Could not kzalloc BUF_SIZE=%d\n", dmp_dev->core.name, MAX_BUF_SIZE);
            rc = -ENOMEM;
            goto err_out_unmap;
        }
        dmp_dev->master_buf_phy = virt_to_phys(dmp_dev->master_buf);
#endif
#endif
    }

    printk("%s: PCI MASTER DATA vir: 0x%08X, phy: 0x%08X\n",
            dmp_dev->core.name, (u32)dmp_dev->master_buf, (u32)dmp_dev->master_buf_phy);
    pci_write_config_dword(pdev, PCI_BASE_ADDRESS_4, dmp_dev->master_buf_phy);
    data32 = 0x0;
    pci_read_config_dword(pdev, PCI_BASE_ADDRESS_4, &data32);
    printk("%s: init BAR_4 to Host Mem address (0x%08X)\n", dmp_dev->core.name, data32);

    // Disable ALL VIC First
    pacdmp_write_bar3( dmp_dev, PACDMP_EVB_HST_VIC2+VIC_INT_ENABLE_CLEAR );
    mmio_write32(dmp_dev->mmio_vaddr, 0, 0xFFFFFFFF);
    pacdmp_write_bar3( dmp_dev, PACDMP_EVB_HST_VIC1+VIC_INT_ENABLE_CLEAR );
    mmio_write32(dmp_dev->mmio_vaddr, 0, 0xFFFFFFFF);

    // Request IRQ
    if ((data32 = request_irq(pdev->irq, &pacdmp_irq, IRQ_FLAGS, dmp_dev->core.name, dmp_dev))) {
        printk(KERN_ERR "%s: Could not reserve IRQ %d\n", dmp_dev->core.name, pdev->irq);
        pci_read_config_byte(pdev, PCI_INTERRUPT_PIN, &data8);
        printk(KERN_ERR "%s: PCI_INTERRUPT_PIN = %d\n", dmp_dev->core.name, data8);
        pci_read_config_byte(pdev, PCI_INTERRUPT_LINE, &data8);
        printk(KERN_ERR "%s: PCI_INTERRUPT_LINE = %d\n", dmp_dev->core.name, data8);
        rc = -EIO;
        goto err_out_free_master_buf;
    } else {
        printk(KERN_ERR "%s: Reserve IRQ %d\n", dmp_dev->core.name, pdev->irq);
        pci_read_config_byte(pdev, PCI_INTERRUPT_LINE, &data8);
        printk(KERN_ERR "%s: PCI_INTERRUPT_LINE = %d\n", dmp_dev->core.name, data8);
    }

    /* init pacdmp with some default values (for exmpale, for file handler ... ) */
    devno = MKDEV(g_pacdmp_major, g_pacdmp_minor + g_pacdmp_num);
    cdev_init(&dmp_dev->cdev, &pacdmp_pci_dev_ops);
    dmp_dev->cdev.owner = THIS_MODULE;
    rc = cdev_add(&dmp_dev->cdev, devno, 1);
    if (rc != 0) {
        printk(KERN_ERR "%s: Unable to register driver.\n", dmp_dev->core.name);
        goto err_out_free_master_buf;
    }

    /* everything is fine */
    g_pacdmp_num++;
    pci_set_drvdata( pdev, dmp_dev );

    init_all_properties( dmp_dev->core.nr );
    printk( "%s: init ok!\n", dmp_dev->core.name );

out:
    return rc;

err_out_free_master_buf:
    if ( pacdmp_reserved_addr != 0 ) {
        iounmap( dmp_dev->master_buf );
    } else {
#ifdef GET_FREE_PAGES
        free_pages( dmp_dev->master_free_pages, GET_FREE_PAGES );
#else
        kfree( dmp_dev->master_buf );
#endif
    }
err_out_unmap:
    pci_iounmap( pdev, dmp_dev->mmio_vaddr );
err_out_release_res:
    pci_release_region( pdev, 0 );
err_out_disable_pci:
    pci_disable_device( pdev );
    goto out;
}

static void __devexit pacdmp_pci_remove_one(struct pci_dev *pdev)
{
    int i;
    struct pacdmp_dev *dmp_dev;

    dmp_dev = pci_get_drvdata( pdev );

    // Disable ALL VIC before leaving driver module
    pacdmp_write_bar3( dmp_dev, PACDMP_EVB_HST_VIC2+VIC_INT_ENABLE_CLEAR );
    mmio_write32(dmp_dev->mmio_vaddr, 0, 0xFFFFFFFF);
    pacdmp_write_bar3( dmp_dev, PACDMP_EVB_HST_VIC1+VIC_INT_ENABLE_CLEAR );
    mmio_write32(dmp_dev->mmio_vaddr, 0, 0xFFFFFFFF);

    /* free irq */
    free_irq( dmp_dev->core.pci->irq, dmp_dev );

    /* release resources */
    pci_iounmap( pdev, dmp_dev->mmio_vaddr );
    pci_disable_device( pdev );
    pci_release_region( pdev, 0 );

    /* ensure that the driver data is no longer valid. */
    pci_set_drvdata( pdev, NULL );

    /* free memory */
    for ( i = 0; i < g_pacdmp_num; i++ ) {
        if ( pacdmp_reserved_addr != 0 ) {
            iounmap( dmp_dev->master_buf );
        } else {
#ifdef GET_FREE_PAGES
            free_pages( dmp_dev->master_free_pages, GET_FREE_PAGES );
#else
            kfree( dmp_dev->master_buf );
#endif
        }
    }
}

static struct pci_driver pacdmp_pci_driver = {
    .name           = MODULENAME,
    .id_table       = pacdmp_pci_tbl,
    .probe          = pacdmp_pci_init_one,
    .remove         = __devexit_p(pacdmp_pci_remove_one),
};

static int __init pacdmp_pci_init_module(void)
{
    int result;
    dev_t dev;

    g_pacdmp_num = 0;
    if (g_pacdmp_major) {
        dev = MKDEV(g_pacdmp_major, g_pacdmp_minor);
        result = register_chrdev_region(dev, PACDMP_MAX_NUM, PACDMP_NAME);
    } else {
        result = alloc_chrdev_region(&dev, g_pacdmp_minor, PACDMP_MAX_NUM, PACDMP_NAME);
        g_pacdmp_major = MAJOR(dev);
    }
    if (result < 0) {
        printk(KERN_WARNING "%s: can't get major %d\n", PACDMP_NAME, g_pacdmp_major);
        return result;
    }

    return pci_register_driver(&pacdmp_pci_driver);
}

static void __exit pacdmp_pci_cleanup_module(void)
{
    int i;
    dev_t devno = MKDEV(g_pacdmp_major, g_pacdmp_minor);

    for (i = 0; i < g_pacdmp_num; i++) {
        cdev_del(&g_pacdmps[i].cdev);
    }
    g_pacdmp_num = 0;

    unregister_chrdev_region(devno, PACDMP_MAX_NUM);

    pci_unregister_driver(&pacdmp_pci_driver);
}

module_init(pacdmp_pci_init_module);
module_exit(pacdmp_pci_cleanup_module);
