/*
 * =====================================================================================
 *       Copyright (C), 2010-2020, Bridge Team
 *       Filename:  stream_server.c
 *       Compiler:  gcc
 *
 *    Description:  implement IP cam stream server
 *         Others:  none
 *        History:
 *
 *        Version:  1.0
 *        Created:  09/20/10 17:11:54
 *         Author:  liukai , liukai@sunnorth.com.cn
 *        Company:  Sunplus Core Technology
 *
 *   Modification:
 *
 *
 * =====================================================================================
 */


#include "streaming_server.h"
#include "ip_cam_dev.h"
#include "file_opt_asf.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "spct_dataqueue.h"

typedef struct server_context
{
	// streaming
	Streaming_Server server;
	Streaming_Session * session;
	Ip_Cam_Device ipcam;
	int streamtype;
	int rtpport;

	// file opt
	int bsavefile;
	File_Property  pro;
	File_Opt fileopt;
	int fileoptready;
	int filesize;
	char basename[256];
	char outfilename[256];
	int beginwrite;
	Data_Queue dq;
	int stop;
	struct timeval timeold;
	struct timeval timenew;
	int framecount;
	int timecount;
} Server_Context;



static void * save_file(void * arg)
{
	Server_Context * context;
	Spct_Data * data;
	context = (Server_Context *) arg;

	// set data queue used
	dq_set_used(&context->dq);

	while(1)
	{
		// get data from data queue
		data = dq_get(&context->dq);
		if(data == NULL)
		{
			continue;
		}
		else
		{
			if(data->type == DATA_VIDEO)
			{
				if(context->beginwrite == 0 && data->flags == DATA_VIDEO_I)
					context->beginwrite = 1;
			}
			if(context->beginwrite && (context->fileoptready == 0))
			{
				fileopt_write_data(&context->fileopt, data);
				context->filesize += data->size;
				if(context->filesize >= 200*1024*1024) // file len 200M
				{
					fileopt_close(&context->fileopt);
					sprintf(context->outfilename,"%s_%d.asf", context->basename, context->framecount);
					context->fileoptready = fileopt_create(&context->fileopt, context->outfilename, &context->pro);
					if(context->fileoptready != 0)
					{
						printf("create file fail %s\n", context->outfilename);
					}
					context->filesize = 0;
					context->beginwrite = 0;
				}
			}
			// free data from data queue
			dq_pop(&context->dq);
		}
	}
}

static void send_data(void * handler, void * arg)
{
	Server_Context * context;
	Streaming_Session * session;
	Spct_Data * data ;
	int passsecond;
	if(handler == NULL || arg == NULL)
		return;
	context = (Server_Context *) handler;
	session = context->session;
	data = (Spct_Data * ) arg;
	if(data->type == DATA_VIDEO)
	{
		// frame rate calculate
		context->framecount ++;
		if(context->framecount == 1)
		{
			gettimeofday(&context->timeold, NULL);
			context->timecount = 0;
		}
		else if((context->framecount %1000) == 0)
		{
			gettimeofday(&context->timenew, NULL);
			passsecond = context->timenew.tv_sec - context->timeold.tv_sec;
			context->timecount += passsecond;
			printf("Get send frame %d, pass time %d, pertime %d, framerate %d\n",
					context->framecount,
					context->timecount ,
					passsecond,
					1000/passsecond);
		  context->timeold =  context->timenew;
		}
	}

	// send data to save file data queue
	if(context->bsavefile)
		dq_in(&context->dq, data);

	// streaming data
	streaming_send_data(context->session, data);
}

int showhelp(char * program)
{
	printf("%s [-h] [-s filename] [-t stream type] [-p port]\n", program);
	printf("         -h: show help\n");
	printf("         -s: save asf file at the save time streaming\n");
	printf("         -t: streaming type 0 uncast, 1 multicast, default 0\n");
	printf("         -p: rtp port number\n");
}


void pasarg(int argc, char * argv[], Server_Context * context)
{
	int i;
	if(argc == 1)
	{
		strncpy(context->basename, "test.asf", sizeof(context->basename));
		strncpy(context->outfilename,"test.asf", sizeof(context->basename));
		return ;
	}

	context->rtpport = 900; // default portnum

	for(i = 1; i < argc; i++)
	{
		if(strncmp(argv[i], "-s", 2) == 0)
		{
			context->bsavefile = 1;
			strncpy(context->basename, argv[++i], sizeof(context->basename));
			strncpy(context->outfilename, argv[i], sizeof(context->basename));
		}
		else if(strncmp(argv[i], "-h", 2) == 0)
		{
			showhelp(argv[0]);
			exit(0);
		}else if(strncmp(argv[i], "-t", 2) == 0)
		{
			context->streamtype = atoi(argv[++i]);
			printf("stream type %d\n", context->streamtype);
		}else if(strncmp(argv[i], "-p", 2) == 0)
		{
			context->rtpport = atoi(argv[++i]);
		}else
		{
			printf("error args\n");
			showhelp(argv[0]);
			exit(0);
		}
	}
	return 0;
}

int main(int argc, char * argv[])
{
	Server_Context context;
	char url[256];
	pthread_attr_t attr;
	pthread_t threadId1;
	int res;

	memset(&context , 0 , sizeof(context));

	// parser argv
	pasarg(argc, argv, &context);

	// perpare file property
	context.pro.audio_codec = CODEC_CT_G723;
	context.pro.video_codec = CODEC_H264;
	context.pro.width = 704; 	// PAL D1
	context.pro.height = 576; 	// PAL D1
	context.pro.audioen = 1; 	// not save audio
	context.pro.videoen = 1; 	// save video
	context.pro.channels = 0; 	// must 0
	context.pro.sample_rate = 8000; // audio sample rate
	context.pro.bit_rate = 200; 	// bitrate

	// device init
	if(ip_cam_construct(&context.ipcam, "/dev/video1"))
	{
		fprintf(stderr, "Open device fail\n");
		return -1;
	}

	// init streaming server
	if(streaming_server_construct(&context.server, 554)) // use default RTSP port
	{
		fprintf(stderr, "Create srever fail\n");
		return -1;
	}

	// create stream session
	context.session = streaming_new_session(&context.server,
					0,	//channel 0
					context.rtpport, // rtp port number
					context.streamtype,	  // is not multicast
					NULL, // use radom multicast address
					1,	  // stream aduio
					1	  // stream video
					);
	if(context.session == NULL)
	{
		fprintf(stderr, "Create streaming session fail\n");
		streaming_server_destruct(&context.server);
		return -1;
	}
	streaming_get_session_url(context.session, url, sizeof(url));
	printf("spct streaming session URL: %s\n", url);

	// register data handle function to ip cam device
	context.ipcam.datahandler = &context;
	context.ipcam.fun = send_data;
	streaming_server_start(&context.server);

	if(context.bsavefile) // save file
	{
		// data queue construct
		res = dq_construct(&context.dq,
				800*1024, // buffer size 800k , must enough to hold 3 or more data
				DQ_TYPE_WAIT // wait queue
				);
		if(res != 0)
		{
			printf("construct data queue fail\n");
			return -1;
		}

		// file opt construct
		res = fileopt_asf_construct(&context.fileopt);
		if(res != 0)
		{
			printf("construct fileopt fail\n");
			return -1;
		}
		// create file
		context.fileoptready = fileopt_create(&context.fileopt, context.outfilename, &context.pro);

		// create save file  thread
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
		if((res = pthread_create(&threadId1, &attr, save_file, &context)) != 0)
		{
			printf("create save data thread fail\n");
			return -1;
		}
	}

	// start device to work
	cam_start_work(&context.ipcam);

	while(context.stop == 0)
	{
		usleep(4000000);
	}

	if(context.bsavefile)
	{
		pthread_join(threadId1, NULL);
	}

	// destruct ipcam
	ip_cam_destruct(&context.ipcam);

	//
	streaming_server_stop(&context.server);
	streaming_del_session(&context.server, context.session);
	streaming_server_destruct(&context.server);

}

