/*****
 *   This program shows how to display 4 channel live path yuv422 frame of one 
 *   SPCT6100 video capture card using Gtk library.
 *
 *   Gtk can display frames with RGB24 format, so we use libswscale to convert 
 *   yuv422 frame to RGB24 format.
 *
 *   This is a multi thread program, we create a new thread to poll, process
 *   and display the live path frame, so no need to embed the poll event to 
 *   Gtk main event loop as that of single thread version.
 *   
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <linux/videodev2.h>
#include <gtk/gtk.h>
#include <X11/Xlib.h>
#define XUTIL_DEFINE_FUNCTIONS
#include <X11/Xutil.h>

#include "config.h"

/*****
 *  Inform libswscale use fast bilinear scale algorithm, it is fastest.
 *  Some other algorithms are defined in libswscale header file.
 */
#define SWS_FLAGS    SWS_FAST_BILINEAR
//#define SRC_PIX_FMT  PIX_FMT_UYVY422
#define SRC_PIX_FMT  PIX_FMT_YUV420P

#define CLEAR(x) memset (&(x), 0, sizeof (x))
int IMAGE_WIDTH = 352;
int IMAGE_HEIGHT = 240;

static GtkWidget *window;
static GtkWidget *darea[4]; // one channel one draw area.
guchar *rgbbuf = NULL;
int global_width = 0;
int global_height = 0;

void *global_data_addr[4] = {NULL, NULL, NULL, NULL};

gboolean 
on_darea_expose(GtkWidget *widget,
        GdkEventExpose *event,
        gpointer user_data);
static void 
destroy(GtkWidget*, gpointer);

struct buffer {
    void *start;
    size_t length;
};

struct header {
    int offset[4];
};
struct buffer *buffers = NULL;
AVPicture m_src_pic;
AVPicture m_dst_pic;
struct SwsContext *m_ctx;

    static void
errno_exit(const char *s)
{
    fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
    exit(EXIT_FAILURE);
}

gboolean xxx(gpointer id){
    gtk_widget_queue_draw(darea[(gint)id]);
    return FALSE;
}

void display_yuv(void *data_starting_addr, int ch_id)
{
    global_data_addr[ch_id] = data_starting_addr;
    gdk_threads_add_idle(xxx, (gpointer)ch_id);
}

void process_data(int fd)
{
    int i;
    unsigned int buf_index;
    unsigned int index;
    struct v4l2_buffer buf;
    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl(fd, VIDIOC_DQBUF, (void *)&buf))
        errno_exit("VIDIOC_DQBUF");

    buf_index = buf.index;
    for (i = 0; i < 4; i++) {
        index = buf_index & 0xff;
        buf_index >>= 8;
        if (index != 0xff) {
            display_yuv(buffers[index].start, i);
        }
    }
    if (-1 == ioctl(fd, VIDIOC_QBUF, &buf))
        errno_exit("VIDIOC_QBUF");
}

/*****
 *  live thread
 *
 *  1. set up V4L2 buffer queue.
 *  2. poll live path.
 *  3. process data.
 */
void live_thread(GtkWidget *widget)
{
    struct v4l2_requestbuffers req;
    enum v4l2_buf_type type;
    int index = 0, fd, result;
    int n_buffers = 0;
    struct pollfd pfd;
    CLEAR (req);

    fd = open("/dev/video0", O_RDWR);
    index = 4;
    ioctl(fd, VIDIOC_S_INPUT, &index);
    index = 0;

    req.count = 12;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl (fd, VIDIOC_REQBUFS, &req)) {
        fprintf (stderr, "SPCT6100 V4L2 Driver does not support "
                "memory mapping\n");
        exit (EXIT_FAILURE);
    }

    buffers = calloc (req.count, sizeof (*buffers));

    if (!buffers) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }
    for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
        struct v4l2_buffer buf;
        CLEAR (buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = n_buffers;
        if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buf))
            errno_exit ("VIDIOC_QUERYBUF");
        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start =
            mmap (NULL,
                    buf.length,
                    PROT_READ | PROT_WRITE, 
                    MAP_SHARED, 
                    fd, buf.m.offset);
        if (MAP_FAILED == buffers[n_buffers].start)
            errno_exit ("mmap");
    }

    for (index = 0; index < n_buffers; index++) {

        struct v4l2_buffer buf;
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = index;
        if (-1 == ioctl(fd, VIDIOC_QBUF, &buf))
            errno_exit("VIDIOC_QBUF");
    }
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl(fd, VIDIOC_STREAMON, &type))
        errno_exit("VIDIOC_STREAMON");

    pfd.fd = fd;
    pfd.events = POLLIN;
    while(1) {

        result = poll(&pfd, 1, -1);
        if(result < 0) {

            break;
        } else {

            process_data(fd);

        }
    }
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl(fd, VIDIOC_STREAMOFF, &type))
        errno_exit("VIDIOC_STREAMON");

}

    int 
main (int argc, char *argv[])
{
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *table;
    GtkWidget *button1, *button2, *button3;
    GtkWidget *button4, *button5, *button6;

    if (!g_thread_supported())
    {
        g_thread_init(NULL);
    }
    if (argc != 2) {
        printf("Usage: program [arg]\n"
                "       arg: 0 - NTSC\n"
                "       arg: 1 - PAL\n");
        exit(1);
    } else {

        if (atoi(argv[1]) == 1)
            IMAGE_HEIGHT = 288;
    }
    gdk_threads_init();
    gtk_init (&argc, &argv);
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    // Draw UI.
    darea[0] = gtk_drawing_area_new ();
    darea[1] = gtk_drawing_area_new ();
    darea[2] = gtk_drawing_area_new ();
    darea[3] = gtk_drawing_area_new ();
    button1 = gtk_button_new_with_mnemonic("_Setting");
    button2 = gtk_button_new_with_mnemonic("_Record");
    button3 = gtk_button_new_with_mnemonic("Sto_p");
    button4 = gtk_button_new_with_mnemonic("_playback");
    button5 = gtk_button_new_with_mnemonic("PushButton");
    button6 = gtk_button_new_with_mnemonic("PushButton");
    table   = gtk_table_new(2, 2, TRUE);
    gtk_table_attach(GTK_TABLE(table), darea[0], 0 ,1, 0, 1, GTK_FILL, GTK_FILL, 5, 5);
    gtk_table_attach(GTK_TABLE(table), darea[1], 1 ,2, 0, 1, GTK_FILL, GTK_FILL, 5, 5);
    gtk_table_attach(GTK_TABLE(table), darea[2], 0 ,1, 1, 2, GTK_FILL, GTK_FILL, 5, 5);
    gtk_table_attach(GTK_TABLE(table), darea[3], 1 ,2, 1, 2, GTK_FILL, GTK_FILL, 5, 5);
    gtk_widget_set_size_request(darea[0], IMAGE_WIDTH, IMAGE_HEIGHT);
    gtk_widget_set_size_request(darea[1], IMAGE_WIDTH, IMAGE_HEIGHT);
    gtk_widget_set_size_request(darea[2], IMAGE_WIDTH, IMAGE_HEIGHT);
    gtk_widget_set_size_request(darea[3], IMAGE_WIDTH, IMAGE_HEIGHT);
    vbox = gtk_vbox_new(FALSE, 5);
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start_defaults(GTK_BOX(vbox), table);
    gtk_box_pack_start_defaults(GTK_BOX(hbox), button1);
    gtk_box_pack_start_defaults(GTK_BOX(hbox), button2);
    gtk_box_pack_start_defaults(GTK_BOX(hbox), button3);
    gtk_box_pack_start_defaults(GTK_BOX(hbox), button4);
    gtk_box_pack_start_defaults(GTK_BOX(hbox), button5);
    gtk_box_pack_start_defaults(GTK_BOX(hbox), button6);
    gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);
    gtk_container_add(GTK_CONTAINER(window), vbox);
    // End of draw UI.

    g_signal_connect(G_OBJECT(darea[0]), 
            "expose-event",
            G_CALLBACK(on_darea_expose), 
            NULL);
    g_signal_connect(G_OBJECT(darea[1]), 
            "expose-event",
            G_CALLBACK(on_darea_expose), 
            NULL);
    g_signal_connect(G_OBJECT(darea[2]), 
            "expose-event",
            G_CALLBACK(on_darea_expose), 
            NULL);
    g_signal_connect(G_OBJECT(darea[3]), 
            "expose-event",
            G_CALLBACK(on_darea_expose), 
            NULL);

    g_signal_connect(G_OBJECT(window), 
            "destroy", 
            G_CALLBACK(destroy), 
            NULL);
    av_log_set_level(AV_LOG_QUIET);
    gtk_widget_show_all (window);

    // Create the new thread.
    g_thread_create((GThreadFunc)live_thread, NULL, FALSE, NULL);

    // gdk_threads_enter/leave is necessary to ensure thread security.
    gdk_threads_enter();
    gtk_main ();
    gdk_threads_leave();
    return 0;
}

    gboolean
on_darea_expose(GtkWidget *widget,
        GdkEventExpose *event,
        gpointer user_data)
{
    int channel_index = 0;

    if (widget == darea[0])
        channel_index = 0;
    else if (widget == darea[1])
        channel_index = 1;
    else if (widget == darea[2])
        channel_index = 2;
    else if (widget ==  darea[3])
        channel_index = 3;
    else 
        channel_index = 0;

    if (global_data_addr[channel_index] != NULL) {

        /*****
         *   rgb buffer is used to store the converted RGB24 format data.
         *   If rgb buffer has not been malloced or user scaled the display window,
         *   we should free the old rgb buffer and malloc a new rgb buffer.
         */
        if (rgbbuf == NULL 
                || widget->allocation.width != global_width 
                || widget->allocation.height != global_height) {

            if (rgbbuf == NULL) {

                rgbbuf = (guchar *)malloc(widget->allocation.width * widget->allocation.height * 3);
                global_width = widget->allocation.width;
                global_height = widget->allocation.height;
                m_ctx = sws_getCachedContext(m_ctx, IMAGE_WIDTH, IMAGE_HEIGHT, SRC_PIX_FMT,
                        widget->allocation.width, widget->allocation.height, PIX_FMT_RGB24,
                        SWS_FLAGS, NULL, NULL, NULL);

            } else { // we should free the old buffer first.

                free(rgbbuf);
                rgbbuf = (guchar *)malloc(widget->allocation.width * widget->allocation.height * 3);
                global_width = widget->allocation.width;
                global_height = widget->allocation.height;
                m_ctx = sws_getCachedContext(m_ctx, IMAGE_WIDTH, IMAGE_HEIGHT, SRC_PIX_FMT,
                        widget->allocation.width, widget->allocation.height, PIX_FMT_RGB24,
                        SWS_FLAGS, NULL, NULL, NULL);
            }

        }

        /*****
         *  Using libswscale to scale and convert the yuv422 frame in one step.
         *
         *  1. Initialize the source data
         *  2. Initialize the destination data.
         *  3. Call sws_scale to scale and convert source yuv422 data to destination RGB24 data.
         */
        avpicture_fill(&m_src_pic, global_data_addr[channel_index], SRC_PIX_FMT, IMAGE_WIDTH, IMAGE_HEIGHT);
        avpicture_fill(&m_dst_pic, rgbbuf, PIX_FMT_RGB24, widget->allocation.width, widget->allocation.height);
        sws_scale(m_ctx, 
                m_src_pic.data, 
                m_src_pic.linesize, 0, IMAGE_HEIGHT,
                m_dst_pic.data, m_dst_pic.linesize);

        gdk_draw_rgb_image(widget->window, 
                widget->style->fg_gc[GTK_STATE_NORMAL],
                0, 
                0, 
                widget->allocation.width, 
                widget->allocation.height,
                GDK_RGB_DITHER_MAX, 
                rgbbuf, 
                widget->allocation.width*3);
    }
    return TRUE;
}

    static void
destroy (GtkWidget *window,
        gpointer data)
{
    gtk_main_quit ();
}
