/*****
 *   This program shows how to display one channel live path yuv422 frame 
 *   using Gtk library.
 *
 *   Gtk can display frames with RGB24 format, so we use libswscale to convert 
 *   yuv422 frame to RGB24 format, libswscale can scale the frame also.
 *
 *   This is a single thread program, we use poll system call to poll live path 
 *   data, so the keypoint is how to embed our poll event to the Gtk main event 
 *   loop.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <linux/videodev2.h>
#include <gtk/gtk.h>

#include <X11/Xlib.h>
#define XUTIL_DEFINE_FUNCTIONS
#include <X11/Xutil.h>

#include "config.h"

/*****
 *  Inform libswscale use fast bilinear scale algorithm, it is fastest.
 *  Some other algorithms are defined in libswscale header file.
 */
#define SWS_FLAGS    SWS_FAST_BILINEAR
//#define SRC_PIX_FMT  SRC_PIX_FMT
#define SRC_PIX_FMT  PIX_FMT_YUV420P

#define CLEAR(x) memset (&(x), 0, sizeof (x))
int IMAGE_WIDTH = 352; 
int IMAGE_HEIGHT = 240;
int channelnum = -1;

struct spct6100_source  {
    GSource source;
    GPollFD pollfd;
};

static GtkWidget *window, *darea;
static int global_width = 0;
static int global_height = 0;
static void *global_data_addr;
guchar *rgbbuf;

gboolean 
on_darea_expose(GtkWidget *widget,
        GdkEventExpose *event,
        gpointer user_data);
static void 
destroy(GtkWidget*, gpointer);

struct buffer {
    void *start;
    size_t length;
};

struct buffer *buffers = NULL;

AVPicture m_src_pic;
AVPicture m_dst_pic;
struct SwsContext *m_ctx;

    static void
errno_exit(const char *s)
{
    fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
    exit(EXIT_FAILURE);
}

void display_yuv(void *data_starting_addr)
{
    global_data_addr = data_starting_addr;

    //Process and display the frame at memory address 'darea'.
    gtk_widget_queue_draw(darea);
}

/*****
 *  1. Get the memory address of the frame.
 *  2. Process and display the frame.
 *  3. Queue the used buffer again.
 */
void process_data(int fd)
{
    struct v4l2_buffer buf;
    unsigned int index;

    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl(fd, VIDIOC_DQBUF, (void *)&buf))
        errno_exit("VIDIOC_DQBUF");

    index = buf.index;
    index = (index >> (channelnum * 8)) & 0xff;
    if (index != 0xff) {
        display_yuv(buffers[index].start);
    }

    if (-1 == ioctl(fd, VIDIOC_QBUF, &buf))
        errno_exit("VIDIOC_QBUF");
}

/*****
 *  Initialize V4L2 buffer queue.
 */
void v4l2_init(int fd)
{
    struct v4l2_requestbuffers req;
    enum v4l2_buf_type type;
    int index = 0;
    int n_buffers = 0;
    CLEAR (req);

    ioctl(fd, VIDIOC_S_INPUT, &index);

    req.count = 12;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl (fd, VIDIOC_REQBUFS, &req)) {
        fprintf (stderr, "SPCT6100 V4L2 Driver does not support "
                "memory mapping\n");
        exit (EXIT_FAILURE);
    }

    buffers = calloc (req.count, sizeof (*buffers));

    if (!buffers) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }
    for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
        struct v4l2_buffer buf;
        CLEAR (buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = n_buffers;
        if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buf))
            errno_exit ("VIDIOC_QUERYBUF");
        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start =
            mmap (NULL,
                    buf.length,
                    PROT_READ | PROT_WRITE, 
                    MAP_SHARED, 
                    fd, buf.m.offset);
        if (MAP_FAILED == buffers[n_buffers].start)
            errno_exit ("mmap");
    }

    for (index = 0; index < n_buffers; index++) {

        struct v4l2_buffer buf;
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = index;
        if (-1 == ioctl(fd, VIDIOC_QBUF, &buf))
            errno_exit("VIDIOC_QBUF");
    }
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl(fd, VIDIOC_STREAMON, &type))
        errno_exit("VIDIOC_STREAMON");
}

/*****
 *  Every time our own polling be waken up, this function will be executed.
 *  When waken up, it means one live path frame has come in, we can display
 *  it.
 */
    static gboolean
spct6100_process(GSource *source, GSourceFuncs callback, gpointer data)
{
    struct spct6100_source *self;
    self = (struct spct6100_source *)source;
    process_data(self->pollfd.fd);
    return TRUE;
}

    static gboolean
spct6100_prepare(GSource *self, gint *timeout)
{
    *timeout = -1;
    return FALSE;
}

    static gboolean
spct6100_check(GSource *source)
{
    struct spct6100_source *self;
    self = (struct spct6100_source *)source;
    return self->pollfd.revents & G_IO_IN;
}

int main (int argc, char *argv[])
{
    int fd;
    struct spct6100_source *my_source;

    /*****
     *  To embed our own poll event into the Gtk main event loop, we
     *  must register this structure to Gtk and implement these three
     *  hook functions.
     */
    static GSourceFuncs funcs = {
        spct6100_prepare, 
        spct6100_check,
        spct6100_process,
        NULL,
    };

    if (argc != 3) {
        printf("Usage: %s <standard> <channel>\n"
                "standard:   NTSC|PAL\n"
                "channel:    0|1|2|3\n", argv[0]);
        return -1;
    }

    if (strcasecmp(argv[1], "NTSC") == 0) {
        IMAGE_HEIGHT = 240;
    } else if (strcasecmp(argv[1], "PAL") == 0) {
        IMAGE_HEIGHT = 288;
    } else {
        printf("Bad argument: %s\n", argv[1]);
        printf("Usage: %s <standard> <channel>\n"
                "standard:   NTSC|PAL\n"
                "channel:    0|1|2|3\n", argv[0]);
        return -1;
    }

    channelnum = atoi(argv[2]);
    if (channelnum < 0 || channelnum > 3) {
        printf("Bad argument: %s\n", argv[2]);
        printf("Usage: %s <standard> <channel>\n"
                "standard:   NTSC|PAL\n"
                "channel:    0|1|2|3\n", argv[0]);
        return -1;
    }

    gtk_init (&argc, &argv);
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);


    darea = gtk_drawing_area_new (); //Gtk draw area widget can display RGB24 images.
    gtk_widget_set_size_request(darea, IMAGE_WIDTH, IMAGE_HEIGHT);
    gtk_container_add(GTK_CONTAINER(window), darea);
    g_signal_connect(G_OBJECT(darea), 
            "expose-event",
            G_CALLBACK(on_darea_expose), 
            NULL);

    g_signal_connect(G_OBJECT(window), 
            "destroy", 
            G_CALLBACK(destroy), 
            NULL);
    gtk_widget_show_all (window);

    fd = open("/dev/video0", O_RDWR);
    v4l2_init(fd);

    //Create our own poll event.
    my_source = (struct spct6100_source *)g_source_new(&funcs, 
            sizeof(struct spct6100_source));

    //Specify which fd and event to listen.
    my_source->pollfd.fd = fd;
    my_source->pollfd.events = G_IO_IN;
    my_source->pollfd.revents = 0;

    //Add our poll event to Gtk main event loop.
    g_source_add_poll((GSource *)my_source, &my_source->pollfd);
    g_source_attach((GSource *)my_source, NULL);

    //Inform libswscale not to display log information on the screen.
    av_log_set_level(AV_LOG_QUIET);

    gtk_main ();
    return 0;
}

    gboolean
on_darea_expose(GtkWidget *widget,
        GdkEventExpose *event,
        gpointer user_data)
{
    if(global_data_addr != NULL) {

        /*****
         *   rgb buffer is used to store the converted RGB24 format data.
         *   If rgb buffer has not been malloced or user scaled the display window,
         *   we should free the old rgb buffer and malloc a new rgb buffer.
         */
        if (rgbbuf == NULL 
                || widget->allocation.width != global_width 
                || widget->allocation.height != global_height) {

            if (rgbbuf == NULL) {

                rgbbuf = (guchar *)malloc(widget->allocation.width * widget->allocation.height * 3);
                global_width = widget->allocation.width;
                global_height = widget->allocation.height;
                m_ctx = sws_getCachedContext(m_ctx, IMAGE_WIDTH, IMAGE_HEIGHT, SRC_PIX_FMT,
                        widget->allocation.width, widget->allocation.height, PIX_FMT_RGB24,
                        SWS_FLAGS, NULL, NULL, NULL);

            } else { // we should free the old buffer first.

                free(rgbbuf);
                rgbbuf = (guchar *)malloc(widget->allocation.width * widget->allocation.height * 3);
                global_width = widget->allocation.width;
                global_height = widget->allocation.height;
                m_ctx = sws_getCachedContext(m_ctx, IMAGE_WIDTH, IMAGE_HEIGHT, SRC_PIX_FMT,
                        widget->allocation.width, widget->allocation.height, PIX_FMT_RGB24,
                        SWS_FLAGS, NULL, NULL, NULL);
            }

        }
        /*****
         *  Using libswscale to scale and convert the yuv422 frame in one step.
         *
         *  1. Initialize the source data
         *  2. Initialize the destination data.
         *  3. Call sws_scale to scale and convert source yuv422 data to destination RGB24 data.
         */
        avpicture_fill(&m_src_pic, global_data_addr, SRC_PIX_FMT, IMAGE_WIDTH, IMAGE_HEIGHT);
        avpicture_fill(&m_dst_pic, rgbbuf, PIX_FMT_RGB24, widget->allocation.width, widget->allocation.height);
        sws_scale(m_ctx, 
                m_src_pic.data, 
                m_src_pic.linesize, 0, IMAGE_HEIGHT,
                m_dst_pic.data, m_dst_pic.linesize);

        gdk_draw_rgb_image(widget->window, 
                widget->style->fg_gc[GTK_STATE_NORMAL],
                0, 
                0, 
                widget->allocation.width,
                widget->allocation.height,
                GDK_RGB_DITHER_MAX, 
                rgbbuf,
                widget->allocation.width*3);
    }
    return TRUE;
}

    static void
destroy (GtkWidget *window,
        gpointer data)
{
    gtk_main_quit ();
}
