/*****
 *  This programe display YUV420P raw data using SDL library.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <linux/videodev2.h>

#include <SDL.h>

#define CLEAR(x) memset (&(x), 0, sizeof (x))

//#define SDL_OVERLAY_FMT  SDL_UYVY_OVERLAY
#define SDL_OVERLAY_FMT  SDL_IYUV_OVERLAY

static int HEIGHT = 240;

struct buffer {
    void *start;
    size_t length;
};
static struct buffer *buffers = NULL;

static int channelnum = -1;
static int fd = -1;
static int n_buffers = 0;
static int counter = 0;

static SDL_Surface *screen;
static SDL_Overlay *overlay;
static SDL_Rect rect;

    static void
errno_exit(const char *s)
{
    fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
    exit(EXIT_FAILURE);
}

// Convert and display the live path frame.
void display_yuv(void *addr, int index)
{
    int size = 352 * HEIGHT;

    switch (overlay->format) {
    case SDL_UYVY_OVERLAY:
        if (overlay->hw_overlay) {
            memcpy(overlay->pixels[0], addr, size * 2);
        } else {
            overlay->pixels[0] = (unsigned char *)addr;
        }
        break;

    case SDL_IYUV_OVERLAY:
        if (overlay->hw_overlay) {
            memcpy (overlay->pixels[0], addr, size);
            memcpy (overlay->pixels[1], addr + size, size/4);
            memcpy (overlay->pixels[2], addr + (size*5/4), size/4);
        } else {
            overlay->pixels[0] = addr;
            overlay->pixels[1] = addr + size;
            overlay->pixels[2] = addr + (size*5/4);
        }
        break;
    } 
    SDL_DisplayYUVOverlay(overlay, &rect);
}

void process_data(void)
{
    struct v4l2_buffer buf;
    unsigned int index;

    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl (fd, VIDIOC_DQBUF, &buf))
        errno_exit("VIDIOC_DQBUF");

    index = buf.index;
    index = (index >> (channelnum * 8)) & 0xff;
    if (index != 0xff) {
        display_yuv(buffers[index].start, index);
        counter++;
    }

    if (-1 == ioctl (fd, VIDIOC_QBUF, &buf))
        errno_exit("VIDIOC_QBUF");
}

//Set up V4L2 buffer queue.
static void local_init_mmap_and_buffer_queue(int fd)
{
    struct v4l2_requestbuffers req;
    enum v4l2_buf_type type;
    int index = 0;
    CLEAR (req);
    req.count = 1;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl (fd, VIDIOC_REQBUFS, &req)) {

        fprintf (stderr, "SPCT6100 V4L2 Driver does not support "
                "memory mapping\n");
        exit (EXIT_FAILURE);
    }

    buffers = calloc (req.count, sizeof (*buffers));

    if (!buffers) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }
    for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
        struct v4l2_buffer buf;
        CLEAR (buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = n_buffers;
        if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buf))
            errno_exit ("VIDIOC_QUERYBUF");
        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start =
            mmap (NULL /* start anywhere */,
                    buf.length,
                    PROT_READ | PROT_WRITE /* required */,
                    MAP_SHARED /* recommended */,
                    fd, buf.m.offset);
        if (MAP_FAILED == buffers[n_buffers].start)
            errno_exit ("mmap");
    }

    for (index = 0; index < n_buffers; index++) {

        struct v4l2_buffer buf;
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = index;
        if (-1 == ioctl(fd, VIDIOC_QBUF, &buf))
            errno_exit("VIDIOC_QBUF");
    }
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl(fd, VIDIOC_STREAMON, &type))
        errno_exit("VIDIOC_STREAMON");
}

int main (int argc, char *argv[])
{
    int n = 0;
    int err = 0;
    struct pollfd pfd;
    enum v4l2_buf_type type;
    struct v4l2_capability cap;
    struct timeval time_start, time_end;

    if (argc != 3) {
        printf("Usage: %s <standard> <channel>\n"
                "standard:   NTSC|PAL\n"
                "channel:    0|1|2|3\n", argv[0]);
        return -1;
    }

    if (strcasecmp(argv[1], "NTSC") == 0) {
        HEIGHT = 240;
    } else if (strcasecmp(argv[1], "PAL") == 0) {
        HEIGHT = 288;
    } else {
        printf("Bad argument: %s\n", argv[1]);
        printf("Usage: %s <standard> <channel>\n"
                "standard:   NTSC|PAL\n"
                "channel:    0|1|2|3\n", argv[0]);
        return -1;
    }

    channelnum = atoi(argv[2]);
    if (channelnum < 0 || channelnum > 3) {
        printf("Bad argument: %s\n", argv[2]);
        printf("Usage: %s <standard> <channel>\n"
                "standard:   NTSC|PAL\n"
                "channel:    0|1|2|3\n", argv[0]);
        return -1;
    }

    atexit(SDL_Quit);

    //Init SDL.
    if(SDL_Init(SDL_INIT_VIDEO)<0)
    {
        printf("SDL Initialize error\n");
        return 1;
    }

    //Create SDL screen and yuv422 overlay.
    screen = SDL_SetVideoMode(352, HEIGHT, 0, SDL_HWSURFACE|SDL_RESIZABLE);
    overlay = SDL_CreateYUVOverlay(352, HEIGHT, SDL_OVERLAY_FMT, screen);
    overlay->pitches[0] = 352;
    overlay->pitches[1] = 176;
    overlay->pitches[2] = 176;
    rect.x = 0;
    rect.y = 0;
    rect.w = 352;
    rect.h = HEIGHT;

    fd = open( "/dev/video0", O_RDWR);
    if (fd < 0) {

        printf( "Can not open SPCT6100 PCI device....\n" );
        exit(0);
    }

    err = ioctl(fd, VIDIOC_QUERYCAP, &cap);
    printf("Driver Name: %s\nCard Name: %s, PCI Info: %s\n",
            cap.driver,
            cap.card,
            cap.bus_info);

    local_init_mmap_and_buffer_queue(fd);
    ioctl(fd, VIDIOC_S_INPUT, &channelnum);

    pfd.fd =  fd;
    pfd.events = POLLIN;
    gettimeofday(&time_start, NULL);
    while(1) {

        n = poll(&pfd, 1, -1);
        if(n<0) {

            break;
        } else {

            process_data();
            if (counter == 150) {
                gettimeofday(&time_end, NULL);
                printf("fps: %f\n", 150.0/(time_end.tv_sec - time_start.tv_sec));
                break;
            }
        }
    }

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl(fd, VIDIOC_STREAMOFF, &type))
        errno_exit("VIDIOC_STREAMOFF");
    printf( "..... Done.\n" );

    // Close all file
    close(fd);
    return 0;
}
