#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <linux/videodev2.h>
#include "vo_sdl.h"
#include "dev.h"

struct DeviceInfo g_LiveDev[4];
int g_nLiveCount = 0;

struct DeviceInfo g_CompressDev[4];
int g_nCompressCount = 0;

#define LINE_SIZE 1

static int g_SDLInited = 0;
static int g_nWidth = 2 * 352 + LINE_SIZE;
static int g_nHeight = 2 * 288 + LINE_SIZE;
static int g_nRow, g_nCol;
static Display_Rect g_Rect[16];

static void init_sdl (void)
{
    int w, h, i;

    switch (g_nLiveCount) {
    case 1: g_nRow = 2; g_nCol = 2; break;
    case 2: g_nRow = 3; g_nCol = 3; break;
    case 3: g_nRow = 3; g_nCol = 4; break;
    case 4: g_nRow = 4; g_nCol = 4; break;

    default:
        return;
    }

    vo_init (g_nWidth, g_nHeight, 4*g_nLiveCount, "V4L2");
    w = (g_nWidth - LINE_SIZE * (g_nCol - 1)) / g_nCol;
    h = (g_nHeight - LINE_SIZE * (g_nRow - 1)) / g_nRow;
    for (i = 0; i < 4*g_nLiveCount; i++) {
        g_Rect[i].x = (i % g_nCol) * (w + LINE_SIZE);
        g_Rect[i].y = (i / g_nCol) * (h + LINE_SIZE);
        g_Rect[i].w = w;
        g_Rect[i].h = h;
    }
}

static void find_buddy (void)
{
    int i, j;

    for (i = 0; i < g_nLiveCount; i++) {
        char name[32];
        char *p;
        int gen_name_ok;

        strcpy (name, g_LiveDev[i].dev_name);
        gen_name_ok = 0;
        for (p = name; *p != '\0'; p++) {
            if (*p >= '0' && *p <= '9') {
                int num;
                sscanf (p, "%d", &num);
                sprintf (p, "%d", num + 1);
                gen_name_ok = 1;
                break;
            }
        }
        if (!gen_name_ok) {
            continue;
        }
        for (j = 0; j < g_nCompressCount; j++) {
            if (strcmp (name, g_CompressDev[j].dev_name) == 0) {
                g_CompressDev[j].has_buddy = 1;
                g_CompressDev[j].buddy_index = i;
                g_LiveDev[i].has_buddy = 1;
                g_LiveDev[j].buddy_index = j;
                break;
            }
        }
    }
}

void open_save_file (FILE *fp[], int n, struct DeviceInfo *pInfo, const char *ext)
{
    int chan;
    time_t now;
    struct tm tmNow;
    char name[256];
    int m;

    time (&now);
    localtime_r (&now, &tmNow);

    for (chan = 0; chan < n; chan++) {
        sprintf (name, "%s-ch%d-tm%04d-%02d-%02d-%02d:%02d:%02d%s",
                 pInfo->dev_name, chan,
                 tmNow.tm_year + 1900, tmNow.tm_mon + 1, tmNow.tm_mday,
                 tmNow.tm_hour, tmNow.tm_min, tmNow.tm_sec, ext);
        for (m = 0; name[m] != '-'; m++) {
            if (name[m] == '/') {
                name[m] = '_';
            }
        }
        fp[chan] = fopen (name, "wb");
    }
}

void close_save_file (FILE *fp[], int n)
{
    int chan;

    for (chan = 0; chan < n; chan++) {
        if (fp[chan]) {
            fclose (fp[chan]);
            fp[chan] = NULL;
        }
    }
}

void open_all_devices (int argc, char *argv[])
{
    int i, j;

    for (i = 1; i < argc; i++) {
        struct v4l2_capability cap;
        struct v4l2_requestbuffers req;
        struct DeviceInfo *pInfo;
        int fd = open (argv[i], O_RDWR|O_SYNC);
        int is_live;
        int rc;

        if (fd < 0) {
            fprintf (stderr, "Open device %s failed\n", argv[i]);
            continue;
        }

        memset (&cap, 0, sizeof (cap));
        if (ioctl (fd, VIDIOC_QUERYCAP, &cap) != 0) {
            close (fd);
            fprintf (stderr, "%s: VIDIOC_QUERYCAP failed\n", argv[i]);
            continue;
        }
        if (strcmp ((char *)cap.card, "SPCT6100") != 0) {
            close (fd);
            fprintf (stderr, "%s: cap.card %s\n", argv[i], cap.card);
            continue;
        }
        if (strcmp ((char *)cap.driver, "SPCT6100-V4L2-L") == 0) {
            is_live = 1;
            pInfo = &(g_LiveDev[g_nLiveCount]);
        } else if (strcmp ((char *)cap.driver, "SPCT6100-V4L2-C") == 0) {
            is_live = 0;
            pInfo = &(g_CompressDev[g_nCompressCount]);
        } else {
            close (fd);
            fprintf (stderr, "%s: cap.driver %s\n", argv[i], cap.driver);
            continue;
        }

        memset (&req, 0, sizeof (req));
        req.count = 4;
        req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_MMAP;

        if (ioctl (fd, VIDIOC_REQBUFS, &req) != 0) {
            close (fd);
            fprintf (stderr, "VIDIOC_REQBUFS failed\n");
            continue;
        }
        pInfo->n_buffers = req.count;
        pInfo->buffers = calloc (req.count, sizeof (struct buffer));

        rc = -1;
        for (j = 0; j < pInfo->n_buffers; j++) {
            struct v4l2_buffer buf;

            memset (&buf, 0, sizeof (buf));
            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_MMAP;
            buf.index = j;
            rc = ioctl (fd, VIDIOC_QUERYBUF, &buf);
            if (rc != 0) {
                fprintf (stderr, "VIDIOC_QUERYBUF failed\n");
                break;
            }
            pInfo->buffers[j].length = buf.length;
            pInfo->buffers[j].start = mmap (NULL /* start anywhere */,
                                            buf.length,
                                            PROT_READ | PROT_WRITE,
                                            MAP_SHARED /* recommended */,
                                            fd, buf.m.offset);
            if (pInfo->buffers[j].start == MAP_FAILED) {
                fprintf (stderr, "mmap failed\n");
                rc = -1;
                break;
            }
            rc = ioctl (fd, VIDIOC_QBUF, &buf);
            if (rc != 0) {
                fprintf (stderr, "VIDIOC_QBUF failed\n");
                break;
            }
        }
        if (rc != 0) {
            free (pInfo->buffers);
            pInfo->buffers = NULL;
            close (fd);
            continue;
        }

        if (is_live) {
            pInfo->dev_type = LiveDev;
            pInfo->dev_id = g_nLiveCount++;
        } else {
            pInfo->dev_type = CompressDev;
            pInfo->dev_id = g_nCompressCount++;
        }
        strcpy (pInfo->dev_name, argv[i]);
        pInfo->fd = fd;
    }

    if (g_nLiveCount > 0) {
        find_buddy ();
    }
}

void close_all_devices (void)
{
    int i, j;
    struct DeviceInfo *pInfo;

    vo_clean ();
    for (i = 0; i < g_nLiveCount; i++) {
        pInfo = &g_LiveDev[i];
        for (j = 0; j < pInfo->n_buffers; j++) {
            munmap (pInfo->buffers[j].start, pInfo->buffers[j].length);
        }
        free (pInfo->buffers);
        close (pInfo->fd);
    }
    g_nLiveCount = 0;
    for (i = 0; i < g_nCompressCount; i++) {
        pInfo = &g_CompressDev[i];
        for (j = 0; j < pInfo->n_buffers; j++) {
            munmap (pInfo->buffers[j].start, pInfo->buffers[j].length);
        }
        free (pInfo->buffers);
        close (pInfo->fd);
    }
    g_nCompressCount = 0;
}

typedef int (*VO_DRAW_RECT) (int channel, void *data, int src_w, int src_h,
                             unsigned int format, int x, int y,
                             unsigned int w, unsigned int h, unsigned int argb);
static void *LiveEntry (void *arg)
{
    enum v4l2_buf_type type;
    struct DeviceInfo *pInfo = (struct DeviceInfo *)arg;
    int dev_id = pInfo->dev_id;
    int fd = pInfo->fd;
    FILE *video_fp[4];
    int chan;
    struct pollfd pfd;
    char szFmt[16];
    int src_w[4], src_h[4];
    unsigned int fmt[4];
    int length[4];
    VO_DRAW_RECT vo_draw_rect[4];

    if (pInfo->video_save) {
        open_save_file (video_fp, 4, pInfo, ".yuv");
    } else {
        memset (video_fp, 0, sizeof (video_fp));
    }

    for (chan = 0; chan < 4; chan++) {
        s_input (fd, chan);
        g_fmt (fd, &src_w[chan], &src_h[chan], szFmt);
        if (strcmp (szFmt, "YUV420P") == 0) {
            fmt[chan] = SPCT_IYUV_FORMAT;
            length[chan] = src_w[chan] * src_h[chan] * 3 / 2;
            vo_draw_rect[chan] = vo_draw_rect_yuv420p;
        } else if (strcmp (szFmt, "UYVY422") == 0) {
            fmt[chan] = SPCT_UYVY_FORMAT;
            length[chan] = src_w[chan] * src_h[chan] * 2;
            vo_draw_rect[chan] = vo_draw_rect_uyvy;
        } else {
            fprintf (stderr, "%s fmt %s unhandled\n", pInfo->dev_name, szFmt);
            goto error_out;
        }
    }

    memset (&pfd, 0, sizeof (pfd));
    pfd.fd = fd;
    pfd.events = POLLIN;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl (fd, VIDIOC_STREAMON, &type) != 0) {
        fprintf (stderr, "%s STREAMON failed\n", pInfo->dev_name);
        goto error_out;
    }

    gettimeofday (&(pInfo->start_time), NULL);
    while (pInfo->thread_run) {
        struct v4l2_buffer buf;
        int result;

        result = poll (&pfd, 1, 2000);
        if (result < 0) {
            fprintf (stderr, "error: %s poll return %d, break\n",
                     pInfo->dev_name, result);
            break;
        } else if (result == 0) {
            fprintf (stderr, "%s poll time out, break\n",
                     pInfo->dev_name);
            break;
        }

        if (!(pfd.revents & POLLIN)) {
            fprintf (stderr, "error: %s pfd.revents = 0x%08x, break\n",
                     pInfo->dev_name, pfd.revents);
            break;
        }

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        if (ioctl (fd, VIDIOC_DQBUF, &buf) != 0) {
            fprintf (stderr, "%s DQBUF failed\n", pInfo->dev_name);
            break;
        }

        for (chan = 0; chan < 4; chan++) {
            int index, id;

            index = ((buf.index >> (chan*8))) & 0xFF;
            if (index == 0xFF) {
                continue;
            }

            if (pInfo->video_save_count[chan] < pInfo->video_save) {
                if (video_fp[chan] != NULL) {
                    fwrite (pInfo->buffers[index].start, 1, length[chan],
                            video_fp[chan]);
                    pInfo->video_save_count[chan]++;
                    if (pInfo->video_save_count[chan] == pInfo->video_save) {
                        fprintf (stderr, "%s channel %d save %u frames OK.\n",
                                 pInfo->dev_name, chan, pInfo->video_save);
                    }
                } else {
                    fprintf (stderr, "%s channel %d live save file not opened.\n",
                             pInfo->dev_name, chan);
                    pInfo->video_save_count[chan] = pInfo->video_save;
                }
            }
            pInfo->frames[chan]++;
            pInfo->bytes[chan] += length[chan];

            id = (4 * dev_id) + chan;

            if (pInfo->has_buddy) {
                int row, col;
                int max_row = (src_h[chan] + 31) / 32;
                int max_col = src_w[chan] / 32;

                for (row = 0; row < max_row; row++) {
                    for (col = 0; col < max_col; col++) {
                        if (pInfo->motion_area[chan][row][col]) {
                            vo_draw_rect[chan] (id,
                                                pInfo->buffers[index].start,
                                                src_w[chan], src_h[chan],
                                                fmt[chan],
                                                (32 * col) + 2,
                                                (32 * row) + 2,
                                                28, 28,
                                                0xff000000);
                            pInfo->motion_area[chan][row][col] = 0;
                        }
                    }
                }
            }

            vo_show_picture (id, pInfo->buffers[index].start,
                             src_w[chan], src_h[chan], fmt[chan], &g_Rect[id]);
        }

        if (ioctl (fd, VIDIOC_QBUF, &buf) != 0) {
            fprintf (stderr, "%s QBUF failed\n", pInfo->dev_name);
            break;
        }
    }
    gettimeofday (&(pInfo->end_time), NULL);

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl (fd, VIDIOC_STREAMOFF, &type) != 0) {
        fprintf (stderr, "%s STREAMOFF failed\n", pInfo->dev_name);
    }

error_out:
    close_save_file (video_fp, 4);
    pInfo->thread_run = 0;
    return NULL;
}

static void parse_motion_packet (struct DeviceInfo *pInfo, int channel,
                                 void *motion_buffer, int w, int h)
{
    int row, col;
    int row_num = (w + 31) / 32;
    int col_num = (h + 31) / 32;
    unsigned int *motion_data = motion_buffer;

    for (row = 0; row < row_num; row++) {
        for (col = 0; col < col_num; col++) {
            if ((motion_data[row] >> col) & 0x1) {
                pInfo->motion_area[channel][row][col] = 1;
            } else {
                pInfo->motion_area[channel][row][col] = 0;
            }
        }
    }
}

static void *CompressEntry (void *arg)
{
    enum v4l2_buf_type type;
    struct DeviceInfo *pInfo = (struct DeviceInfo *)arg;
    int fd = pInfo->fd;
    FILE *video_fp[4];
    FILE *audio_fp[4];
    int sps_got[4];
    int chan;
    struct pollfd pfd;
    char szFmt[16];
    int src_w[4], src_h[4];
    unsigned int fmt[4];
    int video_lost[4];

    if (pInfo->video_save) {
        open_save_file (video_fp, 4, pInfo, ".264");
    } else {
        memset (video_fp, 0, sizeof (video_fp));
    }

    if (pInfo->audio_save) {
        open_save_file (audio_fp, 4, pInfo, ".aud");
    } else {
        memset (audio_fp, 0, sizeof (audio_fp));
    }

    for (chan = 0; chan < 4; chan++) {
        s_input (fd, chan);
        g_fmt (fd, &src_w[chan], &src_h[chan], szFmt);
        if (strcmp (szFmt, "YUV420P") == 0) {
            fmt[chan] = SPCT_IYUV_FORMAT;
        } else if (strcmp (szFmt, "UYVY422") == 0) {
            fmt[chan] = SPCT_UYVY_FORMAT;
        } else {
            fprintf (stderr, "%s fmt %s unhandled\n", pInfo->dev_name, szFmt);
            goto error_out;
        }
        video_lost[chan] = -1;
    }

    memset (&pfd, 0, sizeof (pfd));
    pfd.fd = fd;
    pfd.events = POLLIN;

    memset (sps_got, 0, sizeof (sps_got));
#if 0
    for (chan = 0; chan < 4; chan++) {
        s_ctrl_REQIFRAME (fd, chan);
    }
#endif

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl (fd, VIDIOC_STREAMON, &type) != 0) {
        fprintf (stderr, "%s STREAMON failed\n", pInfo->dev_name);
        goto error_out;
    }

    gettimeofday (&(pInfo->start_time), NULL);

    while (pInfo->thread_run) {
        struct v4l2_buffer buf;
        int result;

        result = poll (&pfd, 1, 2000);
        if (result < 0) {
            fprintf (stderr, "error: %s poll return %d, break\n",
                     pInfo->dev_name, result);
            break;
        } else if (result == 0) {
            fprintf (stderr, "%s poll time out, break\n",
                     pInfo->dev_name);
            break;
        }

        if (!(pfd.revents & POLLIN)) {
            fprintf (stderr, "error: %s pfd.revents = 0x%08x, break\n",
                     pInfo->dev_name, pfd.revents);
            break;
        }

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        if (ioctl (fd, VIDIOC_DQBUF, &buf) != 0) {
            fprintf (stderr, "%s DQBUF failed\n", pInfo->dev_name);
            break;
        }

        for (chan = 0; chan < 4; chan++) {
            unsigned int index;
            int subpkt_index;
            unsigned char *subpkt_addr;

            index = ((buf.index >> (chan*8))) & 0xFF;
            if (index == 0xFF) {
                continue;
            }

            subpkt_addr = pInfo->buffers[index].start;
            for (subpkt_index = 0; subpkt_index < 4; subpkt_index++) {
                if (PACKET_WITH_MOTION (subpkt_addr)) {
                    if (pInfo->has_buddy) {
                        parse_motion_packet (&g_LiveDev[pInfo->buddy_index], chan,
                            subpkt_addr + 88, src_w[chan], src_h[chan]);
                    }
                }

                if (PACKET_WITH_VIDEO (subpkt_addr)) {
                    if (pInfo->video_save_count[chan] < pInfo->video_save) {
                        if (video_fp[chan] != NULL) {
                            if (PACKET_IS_ISPSFRAME (subpkt_addr)) {
                                sps_got[chan] = 1;
                            }
                            if (sps_got[chan]) {
                                fwrite (subpkt_addr + AV_VOFFSET (subpkt_addr),
                                        1, AV_VLEN (subpkt_addr), video_fp[chan]);
                                pInfo->video_save_count[chan]++;
                                if (pInfo->video_save_count[chan] == pInfo->video_save) {
                                   fprintf (stderr, "%s channel %d save %u video frames OK.\n",
                                          pInfo->dev_name, chan, pInfo->video_save);
                               }
                            }
                        } else {
                            fprintf (stderr, "%s channel %d video save file not opened.\n",
                                pInfo->dev_name, chan);
                            pInfo->video_save_count[chan] = pInfo->video_save;
                        }
                    }
                    pInfo->frames[chan]++;
                    pInfo->bytes[chan] += AV_VLEN (subpkt_addr);
                }

                if (PACKET_WITH_AUDIO (subpkt_addr)) {
                    if (pInfo->audio_save_count[chan] < pInfo->audio_save) {
                        if (audio_fp[chan] != NULL) {
                            fwrite (subpkt_addr + AV_AOFFSET (subpkt_addr),
                                    1, AV_ALEN (subpkt_addr), audio_fp[chan]);
                            pInfo->audio_save_count[chan]++;
                            if (pInfo->audio_save_count[chan] == pInfo->audio_save) {
                               fprintf (stderr, "%s channel %d save %u audio frames OK.\n",
                                      pInfo->dev_name, chan, pInfo->audio_save);
                           }
                        } else {
                            fprintf (stderr, "%s channel %d audio save file not opened.\n",
                                pInfo->dev_name, chan);
                            pInfo->audio_save_count[chan] = pInfo->audio_save;
                        }
                    }
                    pInfo->bytes[chan] += AV_ALEN (subpkt_addr);
                }

                if (PACKET_WITH_VIDEOLOST (subpkt_addr)) {
                    if (video_lost[chan] != AV_VLOSTV (subpkt_addr)) {
                        fprintf (stderr, "%s channel %d video lost = %d\n",
                            pInfo->dev_name, chan, AV_VLOSTV (subpkt_addr));
                        video_lost[chan] = AV_VLOSTV (subpkt_addr);
                    }
                }

                subpkt_addr += AV_TOTALSIZE (subpkt_addr);
            }
        }

        if (ioctl (fd, VIDIOC_QBUF, &buf) != 0) {
            fprintf (stderr, "%s QBUF failed\n", pInfo->dev_name);
            break;
        }
    }
    gettimeofday (&(pInfo->end_time), NULL);

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl (fd, VIDIOC_STREAMOFF, &type) != 0) {
        fprintf (stderr, "%s STREAMOFF failed\n", pInfo->dev_name);
    }

error_out:
    close_save_file (video_fp, 4);
    close_save_file (audio_fp, 4);
    pInfo->thread_run = 0;
    return NULL;
}

void start_live_thread (int id)
{
    pthread_attr_t attr;
    struct DeviceInfo *pInfo = &g_LiveDev[id];

    if (pInfo->dev_type != LiveDev || pInfo->thread_run) {
        return;
    }
    if (!g_SDLInited) {
        init_sdl ();
        g_SDLInited = 1;
    }

    memset (pInfo->frames, 0, sizeof (pInfo->frames));
    memset (pInfo->bytes, 0, sizeof (pInfo->bytes));
    memset (pInfo->video_save_count, 0, sizeof (pInfo->video_save_count));
    memset (pInfo->motion_area, 0, sizeof (pInfo->motion_area));
    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_JOINABLE);
    pInfo->thread_run = 1;
    pthread_create (&(pInfo->thread_id), &attr, LiveEntry, pInfo);
}

void start_compress_thread (int id)
{
    pthread_attr_t attr;
    struct DeviceInfo *pInfo = &g_CompressDev[id];

    if (pInfo->dev_type != CompressDev || pInfo->thread_run) {
        return;
    }
    memset (pInfo->frames, 0, sizeof (pInfo->frames));
    memset (pInfo->bytes, 0, sizeof (pInfo->bytes));
    memset (pInfo->audio_save_count, 0, sizeof (pInfo->audio_save_count));
    memset (pInfo->video_save_count, 0, sizeof (pInfo->video_save_count));
    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_JOINABLE);
    pInfo->thread_run = 1;
    pthread_create (&(pInfo->thread_id), &attr, CompressEntry, pInfo);
}

void start_all_live_threads (void)
{
    int i;
    for (i = 0; i < g_nLiveCount; i++) {
        start_live_thread (i);
    }
}

void start_all_compress_threads (void)
{
    int i;
    for (i = 0; i < g_nCompressCount; i++) {
        start_compress_thread (i);
    }
}

void stop_live_thread (int id)
{
    struct DeviceInfo *pInfo = &g_LiveDev[id];

    if (pInfo->thread_run) {
        pInfo->thread_run = 0;
        pthread_join (pInfo->thread_id, NULL);
    }
}

void stop_compress_thread (int id)
{
    struct DeviceInfo *pInfo = &g_CompressDev[id];

    if (pInfo->thread_run) {
        pInfo->thread_run = 0;
        pthread_join (pInfo->thread_id, NULL);
    }
}

void stop_all_live_threads (void)
{
    int i;
    for (i = 0; i < g_nLiveCount; i++) {
        stop_live_thread (i);
    }
}

void stop_all_compress_threads (void)
{
    int i;
    for (i = 0; i < g_nCompressCount; i++) {
        stop_compress_thread (i);
    }
}

void set_live_save (int id, int val)
{
    g_LiveDev[id].video_save = val;
}

void set_all_live_save (int val)
{
    int i;
    for (i = 0; i < g_nLiveCount; i++) {
        set_live_save (i, val);
    }
}

void set_compress_video_save (int id, int val)
{
    g_CompressDev[id].video_save = val;
}

void set_all_compress_video_save (int val)
{
    int i;
    for (i = 0; i < g_nCompressCount; i++) {
        set_compress_video_save (i, val);
    }
}

void set_compress_audio_save (int id, int val)
{
    g_CompressDev[id].audio_save = val;
}

void set_all_compress_audio_save (int val)
{
    int i;
    for (i = 0; i < g_nCompressCount; i++) {
        set_compress_audio_save (i, val);
    }
}

void resize_window (int width, int height)
{
    if (g_SDLInited && g_nLiveCount > 0) {
        int w, h, i;

        g_nWidth = width;
        g_nHeight = height;

        w = (g_nWidth - LINE_SIZE * (g_nCol - 1)) / g_nCol;
        h = (g_nHeight - LINE_SIZE * (g_nRow - 1)) / g_nRow;
        for (i = 0; i < 4*g_nLiveCount; i++) {
            g_Rect[i].x = (i % g_nCol) * (w + LINE_SIZE);
            g_Rect[i].y = (i / g_nCol) * (h + LINE_SIZE);
            g_Rect[i].w = w;
            g_Rect[i].h = h;
        }
        vo_resize (width, height);
    }
}
