#include <SDL/SDL.h>
#include <semaphore.h>
#include "vo_sdl.h"

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

static int g_bInitialized = 0;
static Uint32 g_uFlags;
static SDL_Surface *g_Screen = NULL;
static sem_t g_SemScreen;
static SDL_Overlay **g_Overlay = NULL;
static int g_nChannelNum = 0;

int
vo_init (int width, int height, int channel_num, const char *title)
{
  if (g_bInitialized) {
    return 0;
  }

  if (width <= 0 || height <= 0 || channel_num <= 0) {
    return -1;
  }

  if (SDL_Init (SDL_INIT_VIDEO) < 0) {
    return -1;
  }

  sem_init (&g_SemScreen, 0, 1);
  g_uFlags = SDL_HWSURFACE | SDL_RESIZABLE;
  g_Screen = SDL_SetVideoMode (width, height, 0, g_uFlags);
  if (channel_num == -1) {
	g_nChannelNum = 16;
  } else {
    g_nChannelNum = channel_num;
  }
  g_Overlay = (SDL_Overlay **)calloc (g_nChannelNum, sizeof (SDL_Overlay *));
  SDL_WM_SetCaption (title, "SPCT6100");

  g_bInitialized = 1;
  return 0;
}

int
vo_clean ()
{
  if (g_bInitialized) {
    int i;
    for (i = 0; i < g_nChannelNum; i++) {
      if (g_Overlay[i]) {
        SDL_FreeYUVOverlay (g_Overlay[i]);
      }
    }
    free (g_Overlay);
    g_Overlay = NULL;
    g_nChannelNum = 0;

    SDL_FreeSurface (g_Screen);
    g_Screen = NULL;

    SDL_Quit ();

    sem_destroy (&g_SemScreen);
    g_bInitialized = 0;
  }
  return 0;
}

static SDL_Overlay *
get_overlay (int channel, int src_w, int src_h, unsigned int format)
{
  SDL_Overlay *overlay;

  if (channel < 0 || channel >= g_nChannelNum) {
    return NULL;
  }

  overlay = g_Overlay[channel];
  if (overlay == NULL) {
    sem_wait (&g_SemScreen);
    overlay = SDL_CreateYUVOverlay (src_w, src_h, format, g_Screen);
    sem_post (&g_SemScreen);
    g_Overlay[channel] = overlay;

  } else if (src_w != overlay->w || src_h != overlay->h
             || format != overlay->format) {
    sem_wait (&g_SemScreen);
    SDL_FreeYUVOverlay (overlay);
    overlay = SDL_CreateYUVOverlay (src_w, src_h, format, g_Screen);
    sem_post (&g_SemScreen);
    g_Overlay[channel] = overlay;

  }
  return overlay;
}

int
vo_show_picture (int channel, void *data, int src_w, int src_h,
                 unsigned int format, Display_Rect *rect)
{
  SDL_Rect sdl_rect;
  SDL_Overlay *overlay;
  int size;
  int rc = 0;

  overlay = get_overlay (channel, src_w, src_h, format);
  if (overlay == NULL) {
    return -1;
  }

  size = src_w * src_h;

  switch (format) {
  case SDL_YV12_OVERLAY:
  case SDL_IYUV_OVERLAY:
    if (overlay->hw_overlay) {
      memcpy (overlay->pixels[0], data, size);
      memcpy (overlay->pixels[1], data + size, size/4);
      memcpy (overlay->pixels[1], data + (size*5/4), size/4);
    } else {
      overlay->pixels[0] = data;
      overlay->pixels[1] = data + size;
      overlay->pixels[2] = data + (size*5/4);
    }
    break;

  case SDL_YUY2_OVERLAY:
  case SDL_UYVY_OVERLAY:
  case SDL_YVYU_OVERLAY:
    if (overlay->hw_overlay) {
      memcpy (overlay->pixels[0], data, size*2);
    } else {
      overlay->pixels[0] = data;
    }
    break;

  default:
    rc = -1;
    break;
  }
  if (rc == 0) {
    sdl_rect.x = rect->x;
    sdl_rect.y = rect->y;
    sdl_rect.w = rect->w;
    sdl_rect.h = rect->h;

    sem_wait (&g_SemScreen);
    SDL_DisplayYUVOverlay (overlay, &sdl_rect);
    sem_post (&g_SemScreen);
  }
  return rc;
}

int
vo_resize (int width, int height)
{
  if (!g_bInitialized) {
    return -1;
  }
  if (g_Screen->w != width || g_Screen->h != height) {
    sem_wait (&g_SemScreen);
    g_Screen = SDL_SetVideoMode (width, height, 0, g_uFlags);
    sem_post (&g_SemScreen);
  }
  return 0;
}

int
vo_draw_rect_uyvy (int channel, void *data, int src_w, int src_h,
                   unsigned int format, int x, int y,
                   unsigned int w, unsigned int h, unsigned int argb)
{
  SDL_Overlay *overlay;
  unsigned int A, R, G, B, Y, U, V;
  int left, top, right, bottom, i;
  unsigned char *line, *extra_line;
  unsigned int pitch;

  overlay = get_overlay (channel, src_w, src_h, format);
  if (overlay == NULL) {
    return -1;
  }

  pitch = overlay->pitches[0];
  if (x >= src_w || y >= src_h || w == 0 || h == 0) {
    return 0;
  }
  right = x + w - 1;
  bottom = y + h - 1;

  if (x < 0) {
    x = 0;
  }
  if (y < 0) {
    y = 0;
  }
  if (right >= src_w) {
    right = src_w - 1;
  }
  if (bottom >= src_h) {
    bottom = src_h - 1;
  }
  left = x & (~0x1);
  top = y & (~0x1);
  right &= ~0x1;
  bottom &= ~0x1;

  A = (argb >> 24) & 0xff;
  R = (argb >> 16) & 0xff;
  G = (argb >> 8) & 0xff;
  B = (argb >> 0) & 0xff;
  Y = (0.299*R + 0.587*G + 0.114*B) * A;
  U = (-0.169*R - 0.331*G + 0.499*B + 128) * A;
  V = (0.499*R - 0.418*G - 0.0813*B + 128) * A;

  line = (unsigned char *)data + (pitch * top) +  (2 * left);
  extra_line = line + pitch;
  for (i = left; i <= right; i += 2) {
    line[0] = (U + line[0] * (255 - A)) / 255;
    line[1] = (Y + line[1] * (255 - A)) / 255;
    line[2] = (V + line[2] * (255 - A)) / 255;
    line[3] = (Y + line[3] * (255 - A)) / 255;
    line += 4;

    extra_line[0] = (U + extra_line[0] * (255 - A)) / 255;
    extra_line[1] = (Y + extra_line[1] * (255 - A)) / 255;
    extra_line[2] = (V + extra_line[2] * (255 - A)) / 255;
    extra_line[3] = (Y + extra_line[3] * (255 - A)) / 255;
    extra_line += 4;
  }

  line = (unsigned char *)data + (pitch * (bottom - 1)) +  (2 * left);
  extra_line = line + pitch;
  for (i = left; i <= right; i += 2) {
    line[0] = (U + line[0] * (255 - A)) / 255;
    line[1] = (Y + line[1] * (255 - A)) / 255;
    line[2] = (V + line[2] * (255 - A)) / 255;
    line[3] = (Y + line[3] * (255 - A)) / 255;
    line += 4;

    extra_line[0] = (U + extra_line[0] * (255 - A)) / 255;
    extra_line[1] = (Y + extra_line[1] * (255 - A)) / 255;
    extra_line[2] = (V + extra_line[2] * (255 - A)) / 255;
    extra_line[3] = (Y + extra_line[3] * (255 - A)) / 255;
    extra_line += 4;
  }

  line = (unsigned char *)data + (pitch * (top + 2)) + (2 * left);
  for (i = top + 2; i <= bottom - 2; i++) {
    line[0] = (U + line[0] * (255 - A)) / 255;
    line[1] = (Y + line[1] * (255 - A)) / 255;
    line[2] = (V + line[2] * (255 - A)) / 255;
    line[3] = (Y + line[3] * (255 - A)) / 255;
    line += pitch;
  }

  line = (unsigned char *)data + (pitch * (top + 2)) + (2 * (right - 1));
  for (i = top + 2; i <= bottom - 2; i++) {
    line[0] = (U + line[0] * (255 - A)) / 255;
    line[1] = (Y + line[1] * (255 - A)) / 255;
    line[2] = (V + line[2] * (255 - A)) / 255;
    line[3] = (Y + line[3] * (255 - A)) / 255;
    line += pitch;
  }

  return 0;
}

int
vo_draw_rect_yuv420p (int channel, void *data, int src_w, int src_h,
                      unsigned int format, int x, int y,
                      unsigned int w, unsigned int h, unsigned int argb)
{
  SDL_Overlay *overlay;
  unsigned int A, R, G, B, Y, U, V;
  int left, top, right, bottom, i;
  unsigned char *Ydata, *Udata, *Vdata;
  unsigned char *Yline, *Uline, *Vline;
  unsigned int Ypitch, Upitch, Vpitch;

  overlay = get_overlay (channel, src_w, src_h, format);
  if (overlay == NULL) {
    return -1;
  }

  Ypitch = overlay->pitches[0];
  Upitch = overlay->pitches[1];
  Vpitch = overlay->pitches[2];
  if (x >= src_w || y >= src_h || w == 0 || h == 0) {
    return 0;
  }
  right = x + w - 1;
  bottom = y + h - 1;

  if (x < 0) {
    x = 0;
  }
  if (y < 0) {
    y = 0;
  }
  if (right >= src_w) {
    right = src_w - 1;
  }
  if (bottom >= src_h) {
    bottom = src_h - 1;
  }
  left = x & (~0x1);
  top = y & (~0x1);
  right &= ~0x1;
  bottom &= ~0x1;

  A = (argb >> 24) & 0xff;
  R = (argb >> 16) & 0xff;
  G = (argb >> 8) & 0xff;
  B = (argb >> 0) & 0xff;
  Y = (0.299*R + 0.587*G + 0.114*B) * A;
  U = (-0.169*R - 0.331*G + 0.499*B + 128) * A;
  V = (0.499*R - 0.418*G - 0.0813*B + 128) * A;

  Ydata = (unsigned char *)data;
  Udata = (unsigned char *)data + (src_w * src_h);
  Vdata = (unsigned char *)data + (src_w * src_h) * 3 / 2;

  Yline = Ydata + (Ypitch * top) +  left;
  Uline = Udata + ((Upitch * top) + left) / 2;
  Vline = Vdata + ((Vpitch * top) + left) / 2;
  for (i = left; i < right; i += 2) {
    Yline[0] = (Y + Yline[0] * (255 - A)) / 255;
    Yline[1] = (Y + Yline[1] * (255 - A)) / 255;
    Uline[0] = (U + Uline[0] * (255 - A)) / 255;
    Vline[0] = (V + Vline[0] * (255 - A)) / 255;
    Yline += 2;
    Uline++;
    Vline++;
  }

  Yline = Ydata + (Ypitch * bottom) +  left;
  Uline = Udata + ((Upitch * bottom) + left) / 2;
  Vline = Vdata + ((Vpitch * bottom) + left) / 2;
  for (i = left; i < right; i += 2) {
    Yline[0] = (Y + Yline[0] * (255 - A)) / 255;
    Yline[1] = (Y + Yline[1] * (255 - A)) / 255;
    Uline[0] = (U + Uline[0] * (255 - A)) / 255;
    Vline[0] = (V + Vline[0] * (255 - A)) / 255;
    Yline += 2;
    Uline++;
    Vline++;
  }

  Yline = Ydata + (Ypitch * top) +  left;
  Uline = Udata + ((Upitch * top) + left) / 2;
  Vline = Vdata + ((Vpitch * top) + left) / 2;
  for (i = top; i < bottom; i += 2) {
    Yline[0] = (Y + Yline[0] * (255 - A)) / 255;
    Yline[Ypitch] = (Y + Yline[Ypitch] * (255 - A)) / 255;
    Uline[0] = (U + Uline[0] * (255 - A)) / 255;
    Vline[0] = (V + Vline[0] * (255 - A)) / 255;
    Yline += 2 * Ypitch;
    Uline += Upitch;
    Vline += Vpitch;
  }

  Yline = Ydata + (Ypitch * top) +  right;
  Uline = Udata + ((Upitch * top) + right) / 2;
  Vline = Vdata + ((Vpitch * top) + right) / 2;
  for (i = top; i < bottom; i += 2) {
    Yline[0] = (Y + Yline[0] * (255 - A)) / 255;
    Yline[Ypitch] = (Y + Yline[Ypitch] * (255 - A)) / 255;
    Uline[0] = (U + Uline[0] * (255 - A)) / 255;
    Vline[0] = (V + Vline[0] * (255 - A)) / 255;
    Yline += 2 * Ypitch;
    Uline += Upitch;
    Vline += Vpitch;
  }
  return 0;
}

#if 0
#include <SDL/SDL_ttf.h>

/* ===================== vo_osd functions ================ */
int
vo_osd_init ()
{
  if (TTF_Init() < 0) {
    return -1;
  }
  return 0;
}

void
vo_osd_clean ()
{
  TTF_Quit ();
}

FontHandle
vo_osd_open_font (const char *font, int ptsize, int renderstyle)
{
  TTF_Font *ttf_font;

  ttf_font = TTF_OpenFont(font, ptsize);
  TTF_SetFontStyle (ttf_font, renderstyle);
  return (FontHandle)ttf_font;
}

void
vo_osd_close_font (FontHandle hFont)
{
  TTF_CloseFont ((TTF_Font *)hFont);
}

TextHandle
vo_osd_create_text (FontHandle hFont, const char *str, unsigned int fcolor, unsigned int bcolor)
{
  SDL_Surface *text;
  SDL_Color fore_col, back_col;

  fore_col.r = fcolor & 0xff;
  fore_col.g = (fcolor >> 8) & 0xff;
  fore_col.b = (fcolor >> 16) & 0xff;
  fore_col.unused = 0;

  back_col.r = bcolor & 0xff;
  back_col.g = (bcolor >> 8) & 0xff;
  back_col.b = (bcolor >> 16) & 0xff;
  back_col.unused = 0;

  text = TTF_RenderText_Shaded((TTF_Font *)hFont, str, fore_col, back_col);
  if (text != NULL) {
    SDL_SetColorKey (text, SDL_SRCCOLORKEY | SDL_RLEACCEL, 0);
  }
  return (TextHandle)text;
}

int
vo_osd_show_text (TextHandle hText, Display_Rect *rect, unsigned int bcolor)
{
  SDL_Surface *text;
  SDL_Rect full_rect;
  SDL_Rect blit_rect;

  if (!g_bInitialized) {
    return -1;
  }

  text = (SDL_Surface *)hText;
  full_rect.x = rect->x;
  full_rect.y = rect->y;
  full_rect.w = rect->w;
  full_rect.h = rect->h;

  blit_rect.x = rect->x;
  blit_rect.y = rect->y;
  blit_rect.w = MIN (rect->w, text->w);
  blit_rect.h = MIN (rect->h, text->h);

  sem_wait (&g_SemScreen);
  SDL_FillRect(g_Screen, &full_rect,
               SDL_MapRGB(g_Screen->format,
                          bcolor&0xff, (bcolor>>8)&0xff, (bcolor>>16)&0xff));

  SDL_BlitSurface (text, NULL, g_Screen, &blit_rect);
  SDL_UpdateRects (g_Screen, 1, &full_rect);
  sem_post (&g_SemScreen);
  return 0;
}

void
vo_osd_free_text (TextHandle hText)
{
  SDL_FreeSurface ((SDL_Surface *)hText);
}
#endif
