#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "lib.h"

static void show_usage (int argc, char *argv[])
{
    printf ("Usage: %s <device_name> <path>\n", argv[0]);
    printf ("  device_name                /dev/video[0~7]\n");
    printf ("  path                       firmware bin path\n");
}

static int read_dword (int fd, unsigned int addr, unsigned int *pudata32)
{
    return g_ctrl_UPLOAD (fd, addr, (char *)pudata32, 4);
}

static int write_dword (int fd, unsigned int addr, unsigned int udata32)
{
    return s_ctrl_DOWNLOAD (fd, addr, (char *)&udata32, 4);
}

static int write_file (int fd, unsigned int addr, const char *path, const char *filename)
{
#define BUFFER_SIZE 4096
    FILE *fp;
    char buffer1[BUFFER_SIZE];
    char buffer2[BUFFER_SIZE];
    int len;
    int rc = 0;

    sprintf (buffer1, "%s/%s", path, filename);
    fp = fopen (buffer1, "rb");
    if (fp == NULL) {
        printf ("open file %s failed\n", buffer1);
        return -1;
    }

    while (1) {
        len = fread (buffer1, 1, BUFFER_SIZE, fp);
        if (len <= 0) {
            break;
        }
        if (s_ctrl_DOWNLOAD (fd, addr, buffer1, len) != 0) {
            printf ("download to 0x%08x failed\n", addr);
            rc = -1;
            break;
        }
        if (g_ctrl_UPLOAD (fd, addr, buffer2, len) != 0) {
            printf ("upload from 0x%08x failed\n", addr);
            rc = -1;
            break;
        }
        if (memcmp (buffer1, buffer2, len) != 0) {
            printf ("memory 0x%08x, len %d check failed\n", addr, len);
            rc = -1;
            break;
        }
        addr += len;
    }

    fclose (fp);
    return rc;
}

static int run_code (int fd, unsigned int addr)
{
#define DSP_RUN         0x1
#define DSP_EXCEPTION   0x2
#define DSP_DEBUGGING   0x4
    unsigned int dsp_status;

    if (s_ctrl_RUNCODE (fd, addr)) {
        printf ("s_ctrl_RUNCODE 0x%08x failed\n", addr);
        return -1;
    }

    while (1) {
        dsp_status = -1;
        if (g_ctrl_DSPSTATUS (fd, (int *)&dsp_status) != 0) {
            printf ("g_ctrl_DSPSTATUS fail\n");
            return -1;
        }
        if (dsp_status == 0x0) {
            //printf ("Program Trapped.\n\n");
            break;
        } else if (dsp_status == DSP_RUN) {
            usleep (10000);
        } else if (dsp_status == DSP_EXCEPTION) {
            printf ("DSP Exception!!\n\n");
            return -1;
        } else if (dsp_status == DSP_DEBUGGING) {
            printf ("DSP in Debugging mode!!\n\n");
            return -1;
        } else {
            printf ("unknow DSP status: 0x%x\n", dsp_status);
            return -1;
        }
    }

    return 0;
} 

int main (int argc, char *argv[])
{
    const char *device_name;
    const char *path;
    int fd;
    unsigned int udata32;
    int i;

    if (argc != 3) {
        show_usage (argc, argv);
        return -1;
    }

    path = argv[2];
    device_name = argv[1];
    fd = open (device_name, O_RDWR | O_SYNC);
    if (fd < 0) {
        printf ("Cannot open device %s\n", device_name);
        show_usage (argc, argv);
        return -1;
    }

    if (write_dword (fd, 0x1e162004, 1) != 0) {
        printf ("write_dword failed at line %d\n", __LINE__);
        goto err_out;
    }

    usleep (50*1000);

    if (write_file (fd, 0x1E000000, path, "ddr_config_128_sync_16bit.data.bin") != 0) {
        printf ("write_file failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (write_file (fd, 0x1C000000, path, "ddr_config_128_sync_16bit.text.bin") != 0) {
        printf ("write_file failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (run_code (fd, 0x1C000000) != 0) {
        printf ("run_code failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (read_dword (fd, 0x1E161020, &udata32) != 0) {
        printf ("read_dword failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (udata32 == 0x010A010A) {
        printf ("DDR Init Failed\n");
        goto err_out;
    }

    if (write_dword (fd, 0x1E162050, 0xC8) != 0) {
        printf ("write_dword failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (write_dword (fd, 0x1E162054, 0x12C) != 0) {
        printf ("write_dword failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (write_dword (fd, 0x1E162058, 0x1F4) != 0) {
        printf ("write_dword failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (write_dword (fd, 0x1E150000, 0xDDA9961E) != 0) {
        printf ("write_dword failed at line %d\n", __LINE__);
        goto err_out;
    }

    for( i = 0; i < 4*0x4000; i +=4 ) {
        if (write_dword (fd, 0x1E000000 + i, 0) != 0) {
            printf ("write_dword failed at line %d\n", __LINE__);
            goto err_out;
        }
    }

    if (write_file (fd, 0x01900000, path, "dsp_stub_main.text.bin") != 0) {
        printf ("write_file failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (write_file (fd, 0x03E80000, path, "ST6100_TABLE.bin") != 0) {
        printf ("write_file failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (write_file (fd, 0x1E003E00, path, "TimerISR.text.bin") != 0) {
        printf ("write_file failed at line %d\n", __LINE__);
        goto err_out;
    }

    if (s_ctrl_INITENV (fd) != 0) {
        printf ("s_ctrl_INITENV failed\n");
        goto err_out;
    }

    printf ("%s %s succeeded!\n", argv[0], argv[1]);

err_out:
    close (fd);
    return 0;
}
