#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "spct6100-main.h"
#include "spct6100-debug.h"

int spct6100_debug = 0;

struct spct6100_proc_method {
  void (*func)(struct seq_file *s, spct6100_dev *cur_dev);
  int debug_level;
  struct spct6100_proc_method *next;
};
static struct spct6100_proc_method *g_proc_method_list = NULL;

static void *spct6100_seq_start(struct seq_file *s, loff_t *pos)
{
  if (*pos >= g_device_num)
    return NULL;
  seq_printf(s, "SPCT6100 Video Capture System Status\n\n");
  return g_spct6100_dev + *pos;
}

static void *spct6100_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
  (*pos)++;
  if (*pos >= g_device_num)
    return NULL;
  return g_spct6100_dev + *pos;
}

static void spct6100_seq_stop(struct seq_file *s, void *v)
{
  /* Now nothing to be done */
}

static int spct6100_seq_show(struct seq_file *s, void *v)
{
  spct6100_dev *cur_dev = (spct6100_dev *)v;
  struct spct6100_proc_method *tnode
                     = g_proc_method_list;
  while (tnode != NULL) {
    if (spct6100_debug >= tnode->debug_level) {
      tnode->func(s, cur_dev);
      seq_printf(s, "\n");
    }
    tnode = tnode->next;
  }
  return 0;
}

static struct seq_operations spct6100_seq_ops = {
  .start = spct6100_seq_start,
  .next  = spct6100_seq_next,
  .stop  = spct6100_seq_stop,
  .show  = spct6100_seq_show,
};

#define INDENT_0 ""
#define INDENT_1 "  "
#define INDENT_2 "    "
#define INDENT_3 "      "
static void gen_proc_title(struct seq_file *s, spct6100_dev *cur_dev)
{
  seq_printf
    (s, "==============================================================\n");
  seq_printf
    (s, "Card %d Status:\n",
     cur_dev->index);
  seq_printf
    (s, "==============================================================\n");
}

static void gen_proc_frame_info(struct seq_file *s, spct6100_dev *cur_dev)
{
  int ch_id = 0;
  struct timeval cur_time;

  do_gettimeofday(&cur_time);
  seq_printf(s, INDENT_1"Frame Received Info\n");
  seq_printf(s, "\n");
  seq_printf(s,
             INDENT_2"=====================================================================\n"
             INDENT_2"  Channel Type        | F   Received |  Used Time (s) |    F/Seconds\n"
             INDENT_2"=====================================================================\n");
  for (ch_id = 0; ch_id < 4; ch_id++) {
    seq_printf(s,
             INDENT_2"  Live     Channel %u  |%13u |%15lu |%13lu\n",
             ch_id,
             cur_dev->l_irq[ch_id],
             cur_dev->l_irq_stime[ch_id].tv_sec == 0?0:cur_time.tv_sec - cur_dev->l_irq_stime[ch_id].tv_sec,
             cur_dev->l_irq[ch_id] / (cur_time.tv_sec - cur_dev->l_irq_stime[ch_id].tv_sec));
    seq_printf(s,
             INDENT_2"                      |              |                |\n");
  }
  seq_printf(s,
             INDENT_2"  Live     Total      |%13u |%15lu |%13lu\n",
             cur_dev->l_irq[4],
             cur_dev->l_irq_stime[4].tv_sec == 0?0:cur_time.tv_sec - cur_dev->l_irq_stime[4].tv_sec,
             cur_dev->l_irq[4] / (cur_time.tv_sec - cur_dev->l_irq_stime[4].tv_sec));
  seq_printf(s,
             INDENT_2"=====================================================================\n");

  for (ch_id = 0; ch_id < 4; ch_id++) {
    seq_printf(s,
             INDENT_2"  Compress Channel %u  |%13u |%15lu |%13lu\n",
             ch_id,
             cur_dev->c_irq[ch_id],
             cur_dev->c_irq_stime[ch_id].tv_sec == 0?0:cur_time.tv_sec - cur_dev->c_irq_stime[ch_id].tv_sec,
             cur_dev->c_irq[ch_id] / (cur_time.tv_sec - cur_dev->c_irq_stime[ch_id].tv_sec));
    seq_printf(s,
             INDENT_2"                      |              |                |\n");
  }
  seq_printf(s,
             INDENT_2"  Compress Total      |%13u |%15lu |%13lu\n",
             cur_dev->c_irq[4],
             cur_dev->c_irq_stime[4].tv_sec == 0?0:cur_time.tv_sec - cur_dev->c_irq_stime[4].tv_sec,
             cur_dev->c_irq[4] / (cur_time.tv_sec - cur_dev->c_irq_stime[4].tv_sec));
  seq_printf(s,
             INDENT_2"=====================================================================\n");


  if (lstream_toggle[cur_dev->index]) {
    for (ch_id = 0; ch_id < 5; ch_id++) {
      cur_dev->l_irq_lost_htime[ch_id].tv_sec += (cur_time.tv_sec - cur_dev->l_irq_lost_stime[ch_id].tv_sec);
      cur_dev->l_irq_lost_htime[ch_id].tv_usec += (cur_time.tv_usec - cur_dev->l_irq_lost_stime[ch_id].tv_usec);
      do_gettimeofday(&(cur_dev->l_irq_lost_stime[ch_id]));
    }
  }

  if (cstream_toggle[cur_dev->index]) {
    for (ch_id = 0; ch_id < 5; ch_id++) {
      cur_dev->c_irq_lost_htime[ch_id].tv_sec += (cur_time.tv_sec - cur_dev->c_irq_lost_stime[ch_id].tv_sec);
      cur_dev->c_irq_lost_htime[ch_id].tv_usec += (cur_time.tv_usec - cur_dev->c_irq_lost_stime[ch_id].tv_usec);
      do_gettimeofday(&(cur_dev->c_irq_lost_stime[ch_id]));
    }
  }

  seq_printf(s, "\n"INDENT_1"Frame Lost Info\n");
  seq_printf(s, "\n");
  seq_printf(s,
             INDENT_2"=====================================================================\n"
             INDENT_2"  Channel Type        | F       Lost |  Used Time (s) |    F/Seconds\n"
             INDENT_2"=====================================================================\n");
  for (ch_id = 0; ch_id < 4; ch_id++) {
    seq_printf(s,
             INDENT_2"  Live     Channel %u  |%13u |%15lu |%13lu\n",
             ch_id,
             cur_dev->l_lost_irq[ch_id],
             cur_dev->l_irq_lost_htime[ch_id].tv_sec,
             cur_dev->l_irq_lost_htime[ch_id].tv_sec == 0?0:cur_dev->l_lost_irq[ch_id] / cur_dev->l_irq_lost_htime[ch_id].tv_sec);
    seq_printf(s,
             INDENT_2"                      |              |                |\n");
  }
  seq_printf(s,
             INDENT_2"  Live     Total      |%13u |%15lu |%13lu\n",
             cur_dev->l_lost_irq[4],
             cur_dev->l_irq_lost_htime[4].tv_sec,
             cur_dev->l_irq_lost_htime[4].tv_sec == 0?0:cur_dev->l_lost_irq[4] / cur_dev->l_irq_lost_htime[4].tv_sec);

  seq_printf(s,
             INDENT_2"=====================================================================\n");

  for (ch_id = 0; ch_id < 4; ch_id++) {
    seq_printf(s,
             INDENT_2"  Compress Channel %u  |%13u |%15lu |%13lu\n",
             ch_id,
             cur_dev->c_lost_irq[ch_id],
             cur_dev->c_irq_lost_htime[ch_id].tv_sec,
             cur_dev->c_irq_lost_htime[ch_id].tv_sec == 0?0:cur_dev->c_lost_irq[ch_id] / cur_dev->c_irq_lost_htime[ch_id].tv_sec);
    seq_printf(s,
             INDENT_2"                      |              |                |\n");
  }
  seq_printf(s,
             INDENT_2"  Compress Total      |%13u |%15lu |%13lu\n",
             cur_dev->c_lost_irq[4],
             cur_dev->c_irq_lost_htime[4].tv_sec,
             cur_dev->c_irq_lost_htime[4].tv_sec == 0?0:cur_dev->c_lost_irq[4] / cur_dev->c_irq_lost_htime[4].tv_sec);
  seq_printf(s,
             INDENT_2"=====================================================================\n");
}

static void gen_proc_pci_info(struct seq_file *s, spct6100_dev *cur_dev)
{
  struct timeval cur_time;

  do_gettimeofday(&cur_time);
  seq_printf(s, INDENT_1"PCI Data Read Info\n");
  seq_printf(s, "\n");
  seq_printf(s,
             INDENT_2"=====================================================================\n"
             INDENT_2"  Data Read (Byte) |  Used Time (ms) |  Bytes/Second\n"
             INDENT_2"=====================================================================\n");
  seq_printf(s,
             INDENT_2"%18u |%16u |%14u\n",
             cur_dev->pci_read_len,
             cur_dev->pci_read_time,
             cur_dev->pci_read_len / (cur_dev->pci_read_time/1000));
  seq_printf(s,
             INDENT_2"=====================================================================\n");

  seq_printf(s, "\n");
  seq_printf(s,
             INDENT_2"=====================================================================\n"
             INDENT_2"  Data Write(Byte) |  Used Time (ms) |  Bytes/Second\n"
             INDENT_2"=====================================================================\n");
  seq_printf(s,
             INDENT_2"%18u |%16u |%14u\n",
             cur_dev->pci_write_len,
             cur_dev->pci_write_time,
             cur_dev->pci_write_len / (cur_dev->pci_write_time/1000));
  seq_printf(s,
             INDENT_2"=====================================================================\n");
}

static void gen_proc_property_info(struct seq_file *s, spct6100_dev *cur_dev)
{
  int ch_id = 0;
  int card_id = cur_dev->index;

  seq_printf(s, INDENT_1"Card Property At Current:\n");
  seq_printf(s, "\n");
  for (ch_id = 0; ch_id < 4; ch_id++) {
    seq_printf(s, INDENT_2"Channel %d Property:\n", ch_id);
    seq_printf(s, INDENT_3"Live path status: %s, "
                  "Compress path status: %s, "
                  "Brightness: %d, Contrast: %d, Saturation: %d, Hue: %d\n"INDENT_3
                  "Standard: %s, Compress resolution: %s, "
                  "Compress speed: %d, Compress quality: %d, "
                  "Live resolution: %s\n\n",
                  (qctl_current_values[card_id][ch_id][STATUS_INDEX] & 0x2)?"Working":"Closed",
                  (qctl_current_values[card_id][ch_id][STATUS_INDEX] & 0x1)?"Working":"Closed",
                  qctl_current_values[card_id][ch_id][BRIGHT_INDEX],
                  qctl_current_values[card_id][ch_id][CONTRAST_INDEX],
                  qctl_current_values[card_id][ch_id][SATURATION_INDEX],
                  qctl_current_values[card_id][ch_id][HUE_INDEX],
                  (qctl_current_values[card_id][ch_id][MODE_INDEX] == 1) ? "NTSC": "PAL",
                  (qctl_current_values[card_id][ch_id][CSHOWSIZE_INDEX] == 1)? "CIF": "4CIF",
                  qctl_current_values[card_id][ch_id][EENSPEED_INDEX],
                  qctl_current_values[card_id][ch_id][EQUALITY_INDEX],
                  (qctl_current_values[card_id][ch_id][LSHOWSIZE_INDEX]) == 3 ? "4CIF": "CIF");


  }
}

static int spct6100_proc_open(struct inode *inode, struct file *file)
{
  return seq_open(file, &spct6100_seq_ops);
}
static struct file_operations spct6100_proc_ops = {
  .owner   = THIS_MODULE,
  .open    = spct6100_proc_open,
  .read    = seq_read,
  .llseek  = seq_lseek,
  .release = seq_release,
};

#define SPCT6100_REGISTER_PROC_METHOD(x,y,z) \
  { \
    struct spct6100_proc_method *th = g_proc_method_list; \
    struct spct6100_proc_method *tn = NULL; \
    if (g_proc_method_list == NULL) { \
      g_proc_method_list = (struct spct6100_proc_method *) \
               kzalloc(sizeof(struct spct6100_proc_method), GFP_KERNEL); \
      g_proc_method_list->func = y; \
      g_proc_method_list->debug_level = z; \
      g_proc_method_list->next = NULL; \
    } else { \
      while(th->next != NULL) \
        th = th->next; \
      tn = (struct spct6100_proc_method *) \
                kzalloc(sizeof(struct spct6100_proc_method), GFP_KERNEL); \
      tn->func = y; \
      tn->debug_level = z; \
      tn->next = NULL; \
      th->next = tn; \
    } \
  }

#define DEBUG_LEVEL_1 1
#define DEBUG_LEVEL_2 2

void spct6100_create_proc(void)
{
  struct proc_dir_entry *entry;
  entry = create_proc_entry("spct6100-status", 0 , NULL);
  if (entry)
    entry->proc_fops = &spct6100_proc_ops;

  SPCT6100_REGISTER_PROC_METHOD("Print Card Title",   gen_proc_title,
                                DEBUG_LEVEL_1);
  SPCT6100_REGISTER_PROC_METHOD("Print Frame Info",   gen_proc_frame_info,
                                DEBUG_LEVEL_1);
  SPCT6100_REGISTER_PROC_METHOD("Print PCI Info",     gen_proc_pci_info,
                                DEBUG_LEVEL_1);
  SPCT6100_REGISTER_PROC_METHOD("Print Property Info",gen_proc_property_info,
                                DEBUG_LEVEL_1);
}

void spct6100_remove_proc(void)
{
  remove_proc_entry("spct6100-status", NULL);
}
