#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include "spct6100-main.h"
#include "spct6100-debug.h"
#include "spct6100-io.h"

/**
 * @brief Write 4 bytes into the specified address
 *
 *        The same address restrict with spct6100_readmem
 *
 * @param value Data to write
 *
 * @return Bytes have written
 */

static inline u32
spct6100_mmio_write32 (void __iomem * ioaddr, u32 offset, u32 value)
{
  u32 retval = -1;

  iowrite32 (value, ioaddr + offset);
  retval = ioread32 (ioaddr + offset);  // handle PCI write posting problem
  udelay (1);
  return retval;
}

/**
 * @brief Read 4 bytes from the specified address
 *
 *
 * @return Data readed
 */
static inline u32
spct6100_mmio_read32 (void __iomem * ioaddr, u32 offset)
{
  u32 value = -1;

  value = ioread32 (ioaddr + offset);
  return value;
}

/**
 * @brief Write base address 3 at PCI configuration space
 *
 * @param address Data to write
 */
void
spct6100_pacdmp_write_bar3 (spct6100_dev *cur_dev, u32 address)
{
  unsigned long tflag;
  spin_lock_irqsave (&(cur_dev->reg_lock), tflag);
  pci_write_config_dword (cur_dev->pci, PCI_BASE_ADDRESS_3, address);
  spin_unlock_irqrestore (&(cur_dev->reg_lock), tflag);
}

void
spct6100_pacdmp_read_bar3 (spct6100_dev *cur_dev, void *address)
{
  unsigned long tflag;
  spin_lock_irqsave (&(cur_dev->reg_lock), tflag);
  pci_read_config_dword(cur_dev->pci, PCI_BASE_ADDRESS_3, address);
  spin_unlock_irqrestore (&(cur_dev->reg_lock), tflag);
}

/** @brief Read data from the specified address to reading buffer
 *
 *       Caller should ensure:
 *       1. write the base register 3 of PCI configuration registers.
 *       2. address is 4-bytes align.
 *
 *       Whenever using spct6100_readmem, we should only pass
 *       the address as offset from base register 3.
 *
 *       This function can only access memory in
 *       the range 0x00000000 - 0x00FFFFFF plus base register 3.
 *
 *       All is the same to the function spct6100_writemem
 *
 * @param dev Global spct6100 device pointer
 * @param addr Start memory address we need to read
 * @param buf Reading buf
 * @param insize Length to read
 */
void
spct6100_readmem (spct6100_dev *dev, u32 addr, u8 *buf, s32 insize)
{
  struct timeval time_begin;
  struct timeval time_end;
  int not_record = 0;

  if (spct6100_debug) {
    if (!down_trylock(&(dev->status_read_sem))) {
      dev->pci_read_len += insize;
      do_gettimeofday(&time_begin);
    } else {
      not_record = 1;
    }
  }

  while (insize >= 4) {
    *((u32 *)buf) = spct6100_mmio_read32 (dev->mmio_vaddr, addr);
    addr += 4;
    buf += 4;
    insize -= 4;
  }

  if (insize > 0) {
      u32 tmp;
      tmp = spct6100_mmio_read32 (dev->mmio_vaddr, addr);
      memcpy (buf, &tmp, insize);
  }

  if (spct6100_debug) {
    if (!not_record) {
      do_gettimeofday(&time_end);
      dev->pci_read_time = dev->pci_read_time
                           + (time_end.tv_sec - time_begin.tv_sec)*1000
                           + time_end.tv_usec - time_begin.tv_usec;
      up(&(dev->status_read_sem));
    }
  }
}

/** @brief Write the data of writing buffer to the specified address
 *
 * @param dev Global spct6100 device pointer
 * @param addr Start memory address we need to write to
 * @param buf Writing buffer
 * @param insize Length to write
 *
 */
void
spct6100_writemem (spct6100_dev *dev, u32 addr, u8 *buf, s32 insize)
{
  struct timeval time_begin;
  struct timeval time_end;
  int not_record = 0;

  if (spct6100_debug) {
    if (!down_trylock(&(dev->status_write_sem))) {
      dev->pci_write_len += insize;
      do_gettimeofday(&time_begin);
    } else {
      not_record = 1;
    }
  }

  while (insize >= 4) {
    spct6100_mmio_write32 (dev->mmio_vaddr, addr, *((u32 *)buf));
    addr += 4;
    buf += 4;
    insize -= 4;
  }

  if (insize > 0) {
    u32 tmp;
    tmp = spct6100_mmio_read32 (dev->mmio_vaddr, addr);
    memcpy (&tmp, buf, insize);
    spct6100_mmio_write32 (dev->mmio_vaddr, addr, tmp);
  }

  if (spct6100_debug) {
    if (!not_record) {
      do_gettimeofday(&time_end);
      dev->pci_write_time = dev->pci_write_time
                          + (time_end.tv_sec - time_begin.tv_sec)*1000
                          + time_end.tv_usec - time_begin.tv_usec;
      up(&(dev->status_write_sem));
    }
  }
}
