
/// @mainpage SPCT6100 Linux V4L2 Driver
///
/// @section main_intro Introduction
/// Welcome to the SPCT6100 Linux V4L2 Driver
///
/// This documentation describes the @b internal source code that makes
/// up SPCT6100 linux V4L2 driver, not the @b design @b framework and @b external use of this driver.
/// There are no instructions here on how to use this driver, only the APIs that make up the driver.
/// We wish this driver policy free
///
/// For design information, please see the <a href="../DesignDoc/DesignDoc.html">design document</a>
///
/// For usage instructions, please see the <a href="#">user's guide</a>
///
/// @section main_caveat Caveat
/// This documentation is generated directly from the source code with doxygen.
/// Since the driver is constantly under active development, what you're about to
/// read is out of date! However, it may still be useful since certain portions
/// of the driver are very stable.
///
/// @section main_changelog Change Log
/// - Original content written 2008-11-05 by Guang-Yuan.Wang


/**
 * @file spct6100-main.c
 * @brief SPCT6100 Linux V4L2 Driver main file
 * @author Guang-Yuan.Wang
 * @date 2008-11-04
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
#include <linux/semaphore.h>
#else
#include <asm/semaphore.h>
#endif

#include <asm/atomic.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/vmalloc.h>
#include <linux/workqueue.h>
#include <asm/irq.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/moduleparam.h>
#include <linux/string.h>
#include <linux/jiffies.h>
#include <linux/kthread.h>

/* SPCT6100 Header File */
#include "spct6100-trace.h"
#include "spct6100-packet-define.h"
#include "spct6100-utils.h"
#include "spct6100-main.h"
#include "spct6100-io.h"
#include "spct6100-debug.h"
#include "spct6100-firmware.h"
#include "spctcommon/pacdmp-regs.h" /* for PACDMP_EVB_GPIO */

#define SPCT6100_PCI_VERSION "0.0.1"
#define MODULENAME "spct6100-pci"
#define PACDMP_PCI_VENDOR_ID    0x0216
#define PACDMP_PCI_DEVICE_ID    0x6100
#define PACDMP_EVB_DSP          0x1E000000


/*---------------------------------------------------------------------
 *  General usage macros, types and variables
 *---------------------------------------------------------------------*/
#define NR_DEVS                2

#define LIVE_PATH              1
#define COMPRESS_PATH          2

#define CLEAR(x)  memset(&(x), 0, sizeof (x))

//#define DEBUG_DATA_QUEUE

/*-----------------------------------------------------------------
 * For Fake Firmware
 *----------------------------------------------------------------*/

struct queue_context fip[NR_CARDS];
unsigned int spct6100_qtable_initialized[NR_CARDS];
unsigned int qtable_driver_initilized[NR_CARDS];
unsigned int command_status_init_flag;
unsigned int test_counter[NR_CARDS][NR_CH];
unsigned int global_cpacket[NR_CARDS];
unsigned int global_irq_mask[NR_CARDS];
struct timeval start_time[NR_CARDS];
struct timeval end_time[NR_CARDS];
struct spct6100_osdset g_osdset[NR_CARDS][NR_CH];
struct spct6100_osdset tmp_osdset;
struct spct6100_eqset g_eqset[NR_CARDS][NR_CH];
struct spct6100_eqset tmp_eqset;

struct spct6100_mosaic g_mosaic[NR_CARDS][NR_CH];
int user_mosaicset[NR_CARDS][NR_CH];
struct spct6100_mosaic tmp_mosaicset;

struct spct6100_gopset g_gopset[NR_CARDS][NR_CH];
struct spct6100_gopset tmp_gopset;

struct spct6100_cropset g_cropset[NR_CARDS][NR_CH];
int user_cropset[NR_CARDS][NR_CH];
struct spct6100_cropset tmp_cropset;

typedef struct DOWNLOAD
{
  unsigned int addr;
  char *buf;
  unsigned int len;
} DOWNLOAD;

struct size_table {
  int width;
  int height;
};

static struct size_table SIZE_TABLE_PAL[4] = {
  { 704, 576 },
  { 352, 288 },
  { 176, 144 },
  { 704, 288 },
};

static struct size_table SIZE_TABLE_NTSC[4] = {
  { 704, 480 },
  { 352, 240 },
  { 176, 120 },
  { 704, 240 },
};

/*-----------------------------------------------------------------------
 *  V4L2 related macros, global variables and  structure type
 *---------------------------------------------------------------------*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
struct v4l2_ext_control
{
  unsigned int id;
  unsigned reserved2[2];
  union {
    int value;
    long long value64;
    void *reserved;
  };
};

struct v4l2_ext_controls
{
  unsigned int ctrl_class;
  unsigned int count;
  unsigned int error_idx;
  unsigned int reserved[2];
  struct v4l2_ext_control *controls;
};

#define VIDIOC_S_EXT_CTRLS   _IOWR ('V', 71, struct v4l2_ext_controls)
#endif


#define V4L2_INPUT_TYPE_LIVE_PATH     (V4L2_INPUT_TYPE_TUNER + 1)
#define V4L2_INPUT_TYPE_COMPRESS_PATH (V4L2_INPUT_TYPE_TUNER + 1)

static struct video_device spct6100_v4l2_live[NR_CARDS];
static struct video_device spct6100_v4l2_compress[NR_CARDS];

enum spct6100_input {
  CHANNEL_0 = 0,
  CHANNEL_1,
  CHANNEL_2,
  CHANNEL_3,
  CHANNEL_ALL,
};

struct spct6100_fmt {
  char             *name;
  unsigned int     fourcc;
  int              depth;
};

static const struct spct6100_fmt SPCT_Format_422 = {
  .name    = "4:2:2, packed, UYVY",
  .fourcc  = V4L2_PIX_FMT_UYVY,
  .depth   = 16,
};

static const struct spct6100_fmt SPCT_Format_420 = {
  .name    = "4:2:0, packed, YUV420",
  .fourcc  = V4L2_PIX_FMT_YUV420,
  .depth   = 12,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18) \
    && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static struct v4l2_tvnorm spct6100_tvnorms[2] = {
  {
    "PAL-M",
    V4L2_STD_PAL_M,
    NULL,
  },
  {
    "NTSC-M",
    V4L2_STD_NTSC_M,
    NULL,
  },
};
#endif

static struct v4l2_queryctrl spct6100_qctrl[NUM_V4L2_CID] = {
  {
    .id                = V4L2_CID_REQIFRAME,
    .name              = "Request IFrame",
    .minimum           = 0,
    .maximum           = 4,
    .step              = 1,
    .default_value     = 4,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_INITENV,
    .name              = "Initialize environment",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_BRIGHTNESS,
    .name              = "Brightness",
    .minimum           = 0,
    .maximum           = 255,
    .step              = 1,
    .default_value     = 128,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_CONTRAST,
    .name              = "Contrast",
    .minimum           = 0,
    .maximum           = 255,
    .step              = 1,
    .default_value     = 128,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_SATURATION,
    .name              = "Saturation",
    .minimum           = 0,
    .maximum           = 255,
    .step              = 1,
    .default_value     = 128,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_HUE,
    .name              = "Hue",
    .minimum           = 0,
    .maximum           = 255,
    .step              = 1,
    .default_value     = 128,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_MODE,
    .name              = "Mode(NTSC/PAL)",
    .minimum           = 0,
    .maximum           = 1,
    .step              = 1,
    .default_value     = SPCT6100_DEFAULT_MODE,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_STATUS,
    .name              = "Status",
    .minimum           = 0,
    .maximum           = 3,
    .step              = 1,
    .default_value     = SPCT6100_DEFAULT_STATUS,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_FLAG,
    .name              = "Flag",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_EQUALITY,
    .name              = "CBR/VBR Setting",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_LSHOWSIZE,
    .name              = "Live Path ShowSize",
    .minimum           = 0,
    .maximum           = 3,
    .step              = 1,
    .default_value     = SPCT6100_DEFAULT_LSHOWSIZE,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_EENSPEED,
    .name              = "Encode Speed",
    .minimum           = 0,
    .maximum           = 30,
    .step              = 1,
    .default_value     = 30,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_LFRAMESIZE,
    .name              = "LivePahtFrameSize",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_LBUFBAR,
    .name              = "LBufBar",
    .minimum           = 0,
    .maximum           = 255,
    .step              = 1,
    .default_value     = 128,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_LPATHSCALE,
    .name              = "LPathScale",
    .minimum           = 0,
    .maximum           = 2,
    .step              = 1,
    .default_value     = 1,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_MOSAIC,
    .name              = "Mosic Area",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_MOTIONDETECT,
    .name              = "Motion Detection",
    .minimum           = 0,
    .maximum           = 1,
    .step              = 1,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_OSDTIMESYNC,
    .name              = "Sync OSD timer with host PC",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_DOWNLOAD,
    .name              = "Download data to PCI Card",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_CARDNUM,
    .name              = "Return current card amount",
    .minimum           = 1,
    .maximum           = 4,
    .step              = 1,
    .default_value     = 1,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_DSPSTATUS,
    .name              = "Return Dsp status",
    .minimum           = 1,
    .maximum           = 4,
    .step              = 1,
    .default_value     = 1,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_ALLMODE,
    .name              = "Set 4 Channel Mode",
    .minimum           = 0,
    .maximum           = 15,
    .step              = 1,
    .default_value     = 0,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_UPLOAD,
    .name              = "Upload data from PCI Card",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_DISABLEMOSAIC,
    .name              = "Disable Mosaic Area",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_RUNCODE,
    .name              = "Run Code at Specified Address",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_STOPBOARD,
    .name              = "Stop SPCT6100 CARD",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_TIMEOUT,
    .name              = "Control Process Timeout",
    .minimum           = 0,
    .maximum           = 0x7FFFFFFF,
    .step              = 1,
    .default_value     = 1000,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_MINTERVAL,
    .name              = "Motion interval",
    .minimum           = 0,
    .maximum           = 0x7FFFFFFF,
    .step              = 1,
    .default_value     = 0x2,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_MTHRESHOLD,
    .name              = "Motion Threshold",
    .minimum           = 0,
    .maximum           = 0x7FFFFFFF,
    .step              = 1,
    .default_value     = 0x4000,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_DROPFRAME,
    .name              = "Live Path Drop Frame",
    .minimum           = 0,
    .maximum           = 0xFF,
    .step              = 1,
    .default_value     = 0x7,
    .flags             = V4L2_CTRL_FLAG_DISABLED,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_DSPPING,
    .name              = "Ping DSP",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_WI2C,
    .name              = "Write I2C",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_RI2C,
    .name              = "Read I2C",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_GPIOD,
    .name              = "Set/Get GPIO Direction",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_WGPIO,
    .name              = "Write GPIO",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_VLOSTEN,
    .name              = "Video Lost Toggle",
    .minimum           = 0,
    .maximum           = 1,
    .step              = 1,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_OSDSET,
    .name              = "Set Screen OSD",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_GOPSET,
    .name              = "Set gopb and gopp",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_REQISEQ,
    .name              = "Request an I frame sequence",
    .minimum           = 1,
    .maximum           = 0x7FFFFFFF,
    .step              = 1,
    .default_value     = 1,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  }, {
    .id                = V4L2_CID_GETCLR,
    .name              = "Get color parameter",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_RGPIO,
    .name              = "Read gpio",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_WATCHEN,
    .name              = "Enable watchdog",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_WATCHSET,
    .name              = "watchdog setting",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  }, {
    .id                = V4L2_CID_CSHOWSIZE,
    .name              = "Compress Path ShowSize",
    .minimum           = 0,
    .maximum           = 3,
    .step              = 1,
    .default_value     = SPCT6100_DEFAULT_CSHOWSIZE,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  },{
    .id                = V4L2_CID_PREVIEW,
    .name              = "Set preview mode",
    .minimum           = 0,
    .maximum           = 1,
    .step              = 1,
    .default_value     = 1,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  },{
    .id                = V4L2_CID_AUDIO,
    .name              = "Set audio mode",
    .minimum           = 0,
    .maximum           = 1,
    .step              = 1,
    .default_value     = 1,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_INTEGER,
  },{
    .id                = V4L2_CID_TIMESET,
    .name              = "Set system time",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  },{
    .id                = V4L2_CID_CROPSET,
    .name              = "Set crop",
    .minimum           = 0,
    .maximum           = 0,
    .step              = 0,
    .default_value     = 0,
    .flags             = 0,
    .type              = V4L2_CTRL_TYPE_BUTTON,
  },
};

unsigned int qctl_current_values[NR_CARDS][NR_CH][NUM_V4L2_CID];
static int video_nr = -1;

#define COMPRESS_BUF_NUM   4

#define C(x)  (unsigned char)(x)
#define CP(x) (unsigned char *)(x)
#define I(x)  (unsigned int)(x)
#define IP(x) (unsigned int *)(x)
/* ------------------------------------------------------------------*/

struct data_control {
  int c_index[NR_CH];
  int c_reference;
  int l_index[NR_CH];
  int l_reference;
};

static struct data_control g_DataControl[NR_CARDS];

#ifdef DEBUG_FPS
static struct timeval debug_ltss[NR_CARDS];
static struct timeval debug_ctss[NR_CARDS];
static struct timeval debug_ltse[NR_CARDS];
static struct timeval debug_ctse[NR_CARDS];
static int debug_lframe[NR_CARDS][NR_CH], debug_cframe[NR_CARDS][NR_CH];
#endif

static int g_cfile_id[NR_CARDS];
static int g_lfile_id[NR_CARDS];
static struct timeval c_timestamp;
int lstream_toggle[NR_CARDS];
int cstream_toggle[NR_CARDS];
unsigned int prev_flag[NR_CARDS], prev_chan_id[NR_CARDS];

struct spct6100_wait_list_node {
  wait_queue_head_t *p;
  int data_flag;
  int pid;
  struct spct6100_file_private *cur_file;
  struct spct6100_wait_list_node *next;
};

#define PSTATUS_OPEN        0
#define PSTATUS_STREAMON    1
#define PSTATUS_STREAMOFF   2
#define PSTATUS_RELEASE     3
struct spct6100_file_private {
  spct6100_dev *cur_dev;
  int *totaluser;
  int type;
  int index;
  int pid;
  int input;
  int expires;
  int status;
  int l_index[4];
  struct timer_list trace;
#ifdef ENABLE_ASYNC_IO
  struct fasync_struct *async_node;
#endif
  struct spct6100_wait_list_node *qnode;
  wait_queue_head_t data_queue;
};

static struct spct6100_wait_list_node *g_ldata_wh[NR_CARDS]
                                        = {NULL, NULL, NULL, NULL};
static struct spct6100_wait_list_node *g_cdata_wh[NR_CARDS]
                                        = {NULL, NULL, NULL, NULL};

#define FP(x) ((struct spct6100_file_private *)(x))

spct6100_dev g_spct6100_dev[NR_CARDS];
int g_device_num = 0;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
static int
spct6100_open_compress(struct file *file);

static int
spct6100_release_compress(struct file *file);

static int
spct6100_open_live(struct file *file);

static int
spct6100_release_live(struct file *file);
#else
static int
spct6100_open_compress(struct inode *inode, struct file *file);

static int
spct6100_release_compress(struct inode *inode, struct file *file);

static int
spct6100_open_live(struct inode *inode, struct file *file);

static int
spct6100_release_live(struct inode *inode, struct file *file);
#endif


static ssize_t
spct6100_read_compress(struct file *file, char *buf, size_t count,
                                                 loff_t * offset);
static unsigned int
spct6100_poll_compress(struct file *file, poll_table *wait);


static ssize_t
spct6100_read_live(struct file *file, char *buf, size_t count,
                                             loff_t * offset);
static unsigned int
spct6100_poll_live(struct file *file, poll_table *wait);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
static int
spct6100_video_ioctl2(struct file *file, unsigned int cmd, unsigned long arg);
#endif

static int
spct6100_vidioc_g_ctrl(struct file *file, void *priv, struct v4l2_control *ctrl);

static int
spct6100_vidioc_querybuf_live (struct file *file, void *priv, struct v4l2_buffer *p);

static int
spct6100_vidioc_querybuf_compress (struct file *file, void *priv, struct v4l2_buffer *p);

static int
spct6100_vidioc_streamoff_live (struct file *file, void *priv,
                                         enum v4l2_buf_type i);
static int
spct6100_vidioc_streamoff_compress(struct file *file,
                                   void *priv,
                                   enum v4l2_buf_type i);

/*------------------------------------------------------------------------
 *  Common fops declaration
 *-----------------------------------------------------------------------*/
#ifdef ENABLE_ASYNC_IO
static int
spct6100_fasync_common(int fd, struct file *filp, int mode);
#endif
static int
spct6100_mmap_common(struct file *file, struct vm_area_struct *vma);
static ssize_t
spct6100_write_common(struct file *file, const char *buf, size_t count,
                                                      loff_t * offset);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
static int
spct6100_ioctl_common(struct inode *inode, struct file *filp,
                      unsigned int cmd, unsigned long arg);
static int
spct6100_do_ioctl(struct inode *inode, struct file *file,
                  unsigned int cmd, void *arg);
#endif
/*---------------------------------------------------------------------
 *   Other function declarations
 *--------------------------------------------------------------------*/
static int
spct6100_init(void);

static void
spct6100_cleanup(void);

static int
spct6100_download(struct v4l2_control *ctrl, spct6100_dev *cur_dev);

static int
spct6100_upload(struct v4l2_control *ctrl, spct6100_dev *cur_dev);

static int __devinit
spct6100_dvr_probe(struct pci_dev *pdev, const struct pci_device_id *ent);

static int
spct6100_vidioc_s_ctrl(struct file *file, void *priv,
                       struct v4l2_control *ctrl);
/*------------------------------------------------------------------------
 *   For PCI
 *-----------------------------------------------------------------------*/
static void __devexit spct6100_pci_remove_one    (struct pci_dev *pdev);
static struct pci_device_id spct6100_dvr_tbl[] = {

  {PCI_DEVICE (PACDMP_PCI_VENDOR_ID, PACDMP_PCI_DEVICE_ID)},
  {0,},
};

static struct pci_driver spct6100_dvr_pci_driver = {
  .name = MODULENAME,
  .id_table = spct6100_dvr_tbl,
  .probe = spct6100_dvr_probe,
  .remove = __devexit_p (spct6100_pci_remove_one),
};


/*-------------------------------------------------------------------------
 *  Live path file operation function pointer initialization
 *------------------------------------------------------------------------*/
static
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
struct file_operations
#else
struct v4l2_file_operations
#endif
spct6100_fops_live = {
  owner  :  THIS_MODULE,

  open   :  spct6100_open_live,
  read   :  spct6100_read_live,
  poll   :  spct6100_poll_live,
  release:  spct6100_release_live,

  mmap   :  spct6100_mmap_common,
  write  :  spct6100_write_common,

#ifdef ENABLE_ASYNC_IO
  fasync :  spct6100_fasync_common,
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
  ioctl  :  spct6100_ioctl_common,
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
  ioctl  :  video_ioctl2,
#else
  ioctl  :  spct6100_video_ioctl2,
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
  llseek :  NULL,
#endif
};

/*-------------------------------------------------------------------------
 *  Compress path file operation function pointer initialization
 *------------------------------------------------------------------------*/
static
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
struct file_operations
#else
struct v4l2_file_operations
#endif
spct6100_fops_compress = {
  owner  :  THIS_MODULE,

  open   :  spct6100_open_compress,
  poll   :  spct6100_poll_compress,
  read   :  spct6100_read_compress,
  release:  spct6100_release_compress,

  mmap   :  spct6100_mmap_common,
  write  :  spct6100_write_common,

#ifdef ENABLE_ASYNC_IO
  fasync :  spct6100_fasync_common,
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
  ioctl  :  spct6100_ioctl_common,
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
  ioctl  :  video_ioctl2,
#else
  ioctl  :  spct6100_video_ioctl2,
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
  llseek :  NULL,
#endif
};

/*****
 * Module param
 */
#ifdef __arm__
int spct6100_preview_pci = 0;
#else
int spct6100_preview_pci = 0;
#endif

int spct6100_cpacket_num = 4;

int spct6100_gpio_mode = 0;

#ifdef __arm__
int spct6100_irq_shared = 0;
#endif

static int spct6100_driver_download = 0;

#ifdef SPCT6100_RESERVED_ADDR
unsigned long spct6100_reserved_addr = SPCT6100_RESERVED_ADDR;
#else
unsigned long spct6100_reserved_addr = 0;
#endif

static int spct6100_process_expire = 1000;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
//ioctl wrapper
static int
spct6100_video_ioctl2(struct file *file, unsigned int cmd, unsigned long arg)
{
  int ret;
  char sbuf[128];

  switch (cmd) {
    case VIDIOC_G_CTRL:
      if (copy_from_user(sbuf, (void __user *)arg, _IOC_SIZE(cmd)))
        return -EFAULT;

      ret = spct6100_vidioc_g_ctrl(file, file->private_data,
                                   (struct v4l2_control *)sbuf);
      if (ret < 0)
        return ret;

      if (copy_to_user((void __user *)arg, sbuf, _IOC_SIZE(cmd)))
        return -EFAULT;
      return 0;

    case VIDIOC_QUERYBUF:
      {
        struct spct6100_file_private *cur_file =
                 (struct spct6100_file_private *)file->private_data;

        if (copy_from_user(sbuf, (void __user *)arg, _IOC_SIZE(cmd)))
          return -EFAULT;

        if (cur_file->type == LIVE_PATH) {
          ret = spct6100_vidioc_querybuf_live(file, file->private_data,
                                              (struct v4l2_buffer *)sbuf);
        } else if (cur_file->type == COMPRESS_PATH) {
          ret = spct6100_vidioc_querybuf_compress(file, file->private_data,
                                                  (struct v4l2_buffer *)sbuf);
        } else {
          ret = -EINVAL;
        }

        if (ret < 0)
          return ret;

        if (copy_to_user((void __user *)arg, sbuf, _IOC_SIZE(cmd)))
          return -EFAULT;
        return 0;
      }
  }

  return video_ioctl2(file, cmd, arg);
}
#endif

/** @brief Virtual live data device release function
 *
 *
 * @return 0 on success, error code on failure
 */
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
spct6100_release_live(struct file *file)
#else
spct6100_release_live(struct inode *inode, struct file *file)
#endif
{
  struct spct6100_file_private *cur_file =
          (struct spct6100_file_private *)file->private_data;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  struct spct6100_wait_list_node *th = g_ldata_wh[card_id];

  if (cur_file->status == PSTATUS_STREAMON) {
    spct6100_vidioc_streamoff_live(file, cur_file,
                                   V4L2_BUF_TYPE_VIDEO_CAPTURE);
  }

  if (cur_dev->live_users != 0) {
    cur_dev->live_users--;
    if (cur_dev->live_users == 0) {
      while (g_ldata_wh[card_id] != NULL) {
        th = g_ldata_wh[card_id];
        g_ldata_wh[card_id] = g_ldata_wh[card_id]->next;
        kfree(th);
      }
    }
  }
#ifdef ENABLE_ASYNC_IO
  spct6100_fasync_common(-1, file, 0);
#endif
  cur_file->status = PSTATUS_RELEASE;
  DEBUG ("SPCT6100-LIVE %d: released, users: %d\n",
         iminor(file->f_dentry->d_inode), cur_dev->live_users);

  return 0;
}

/** @brief Virtual compress data device release function.
 *
 *
 * @return The same as spct6100_release_live
 */
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
spct6100_release_compress(struct file *file)
#else
spct6100_release_compress(struct inode *inode, struct file *file)
#endif
{
  struct spct6100_file_private *cur_file =
    (struct spct6100_file_private *)file->private_data;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  unsigned int card_id = cur_dev->index;
  struct spct6100_wait_list_node *th = g_cdata_wh[card_id];

  if (cur_file->status == PSTATUS_STREAMON) {
    spct6100_vidioc_streamoff_compress(file, cur_file,
                                       V4L2_BUF_TYPE_VIDEO_CAPTURE);
  }

  if (down_interruptible(&(cur_dev->sem))) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    return -ERESTARTSYS;
  }
  if (cur_dev->compress_users != 0) {
    cur_dev->compress_users--;
    if (cur_dev->compress_users == 0) {
      while (g_cdata_wh[card_id] != NULL) {
        th = g_cdata_wh[card_id];
        g_cdata_wh[card_id] = g_cdata_wh[card_id]->next;
        kfree(th);
      }
    }
  }
  up(&(cur_dev->sem));

  cur_file->status = PSTATUS_RELEASE;
  DEBUG ("SPCT6100-COMPRESS %d: released, users: %d\n",
         iminor(file->f_dentry->d_inode), cur_dev->compress_users);
  return 0;
}

/** @brief Virtual compress data device read function.
 *
 *         We do not support read method yet, we suggest using
 *         buffer streaming methond instead.
 *
 *         Maybe we will support read method later, so every device
 *         has a individual read function instead of one common read
 *         function like write
 *
 * @return Always return -1 now
 */
static ssize_t
spct6100_read_compress(struct file *file, char *buf,
                      size_t count, loff_t * offset)
{
    return -1;
}

/** @brief Virtual live data device read function.
 *
 *
 * @return The same as spct6100_read_compress
 */
static ssize_t
spct6100_read_live(struct file *file, char *buf,
                  size_t count, loff_t * offset)
{
  return -1;
}

/** @brief The common write function.
 *
 *         We do not support write method also. User should never
 *         try to write to hardware PAC Card.
 *
 *
 * @return Always return -1 now
 */
static ssize_t
spct6100_write_common (struct file *file, const char *buf,
                            size_t count, loff_t * offset)
{
  return -1;
}

/** @brief Asynchronous IO hook
 *
 *       We need to implement this hook to support asynchronous io operation
 *
 * @param mode Passed down by kernel from user application, driver should not
 *             modify it
 *
 */
#ifdef ENABLE_ASYNC_IO
static int
spct6100_fasync_common (int fd, struct file *file, int mode)
{
  struct spct6100_file_private *cur_file =
  (struct spct6100_file_private *)file->private_data;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  return fasync_helper(fd, file, mode, &(cur_file->async_node));
}
#endif

/** @brief VM open hook
 *
 *         Do nothing now, can be reimplement later
 *
 */
static void spct6100_vma_open (struct vm_area_struct *vma)
{
}

/** @brief VM close hook
 *
 *         Do nothing now, can be reimplement later
 */
static void spct6100_vma_close (struct vm_area_struct *vma)
{
}

static struct vm_operations_struct spct6100_mmap_common_vm_ops = {
  .open  = spct6100_vma_open,
  .close = spct6100_vma_close,
 };

/** @brief Memory mapping hook
 *
 *         Map device physical memory address into user space
 *         virtual address, so that user application can access
 *         device physical memory directly
 *
 *         This can improve the data transport performance largely
 *
 *         We have not do strictly check on the vm_pgoff field of
 *         arg vma now, this checking should be added
 *
 * @param vma See the book <<Linux Device Drivers 3rd Edition>> for details
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_mmap_common (struct file *file, struct vm_area_struct *vma)
{
  struct spct6100_file_private *cur_file;
  spct6100_dev *cur_dev;
  unsigned long vm_pgoff;

  cur_file = (struct spct6100_file_private *)file->private_data;
  cur_dev = cur_file->cur_dev;

  vma->vm_flags |= VM_RESERVED;
#ifdef __arm__
  vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif

  vm_pgoff = vma->vm_pgoff;
  if (vm_pgoff == 0) {
    vm_pgoff = ((unsigned long)cur_dev->master_buf_phy) >> PAGE_SHIFT;
  }

  if (remap_pfn_range(vma,
                      vma->vm_start,
                      vm_pgoff,
                      vma->vm_end - vma->vm_start,
                      vma->vm_page_prot))
  {
    ERROR ("mmap error occured\n");
    return -EAGAIN;
  }
  vma->vm_ops = &spct6100_mmap_common_vm_ops;
  spct6100_vma_open(vma);
  return 0;
}

/** @brief IO multiplexing hook for virtual live deivce
 *
 *         The condition of that if statement means we have live
 *         path data, POLLIN and POLLRDNORM should be set in status
 *         mask and return to kernel, kernel will wake up the
 *         corresponding user application
 *
 * @param wait See the book <<Linux Device Drivers 3rd Edition>>
 *             for details, it is transparent to driver
 *
 * @return The status mask of the device
 */
static unsigned int
spct6100_poll_live (struct file *file, poll_table *wait)
{
  struct spct6100_file_private *cur_file = (struct spct6100_file_private *)
                                  file->private_data;
  //spct6100_dev *cur_dev = cur_file->cur_dev;
  //int card_id = cur_dev->index;

  unsigned int mask = 0;

  poll_wait(file, &(cur_file->data_queue), wait);

  if (cur_file->qnode->data_flag) {
    mask |= POLLIN | POLLRDNORM;
  }
  return mask;
}

/** @brief IO multiplexing hook for virtual compress device
 *
 *         The value of data_flag will show whether
 *         there are compress path data ready, then we can set the
 *         status mask respectively
 *
 * @return The status mask of the device
 */
static unsigned int
spct6100_poll_compress (struct file *file, poll_table *wait)
{
  struct spct6100_file_private *cur_file =
  (struct spct6100_file_private *) file->private_data;

  unsigned int mask = 0;

  poll_wait(file, &(cur_file->data_queue), wait);

  if (cur_file->qnode->data_flag) {
    mask |= POLLIN | POLLRDNORM;
  }
  return mask;
}

#define LOCAL_BUFFER_SIZE 4096
static int
spct6100_download(struct v4l2_control *ctrl, spct6100_dev *cur_dev)
{
  DOWNLOAD load;
  unsigned char *bufp;
  unsigned long aligned_addr;

  if (__copy_from_user (&load,
                        (void __user *)(unsigned long)(ctrl->value),
                        sizeof (DOWNLOAD))) {
    ERROR("%s: line %d error\n", __FUNCTION__, __LINE__);
    return -EFAULT;
  }

  if (load.len == 0)
    return 0;

  bufp = (unsigned char *) vmalloc (LOCAL_BUFFER_SIZE);
  if (!bufp) {
    ERROR("%s: line %d error\n", __FUNCTION__, __LINE__);
    return -EFAULT;
  }

  if (down_interruptible (&(cur_dev->sem))) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    vfree (bufp);
    return -ERESTARTSYS;
  }

  aligned_addr = load.addr & (~0x03);
  spct6100_pacdmp_write_bar3 (cur_dev, aligned_addr);
  load.addr -= aligned_addr;

  if (load.addr != 0) {
    int pad_size = load.addr;
    int head_size = 4 - pad_size;

    if (head_size > load.len) {
      head_size = load.len;
    }
    spct6100_readmem (cur_dev, 0, bufp, 4);
    __copy_from_user (bufp + pad_size, load.buf, head_size);
    spct6100_writemem (cur_dev, 0, bufp, 4);
    load.addr += head_size;
    load.buf += head_size;
    load.len -= head_size;
  }

  while (load.len >= LOCAL_BUFFER_SIZE) {
    __copy_from_user (bufp,
                      load.buf,
                      LOCAL_BUFFER_SIZE);
    spct6100_writemem (cur_dev,
                       load.addr,
                       bufp,
                       LOCAL_BUFFER_SIZE);
    load.addr += LOCAL_BUFFER_SIZE;
    load.buf += LOCAL_BUFFER_SIZE;
    load.len -= LOCAL_BUFFER_SIZE;
  }

  if (load.len > 0) {
    __copy_from_user (bufp, load.buf, load.len);
    spct6100_writemem (cur_dev, load.addr, bufp, load.len);
  }

  spct6100_pacdmp_write_bar3 (cur_dev, START_ADDRESS);

  up (&(cur_dev->sem));
  vfree (bufp);
  return 0;
}

static int
spct6100_upload(struct v4l2_control *ctrl, spct6100_dev *cur_dev)
{
  DOWNLOAD load;
  unsigned char *bufp;
  unsigned long aligned_addr;

  if (__copy_from_user (&load,
                        (void __user *)(unsigned long)(ctrl->value),
                        sizeof (DOWNLOAD)))
    return -EFAULT;

  if (load.len == 0)
    return 0;

  bufp = (unsigned char *) vmalloc (LOCAL_BUFFER_SIZE);
  if (!bufp) {
    return -EFAULT;
  }

  if (down_interruptible (&(cur_dev->sem))) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    vfree (bufp);
    return -ERESTARTSYS;
  }

  aligned_addr = load.addr & (~0x03);
  spct6100_pacdmp_write_bar3 (cur_dev, aligned_addr);
  load.addr -= aligned_addr;

  if (load.addr != 0) {
    int pad_size = load.addr;
    int head_size = 4 - pad_size;

    if (head_size > load.len) {
      head_size = load.len;
    }
    spct6100_readmem (cur_dev, 0, bufp, 4);
    __copy_to_user (load.buf, bufp + pad_size, head_size);
    load.addr += head_size;
    load.buf += head_size;
    load.len -= head_size;
  }

  while (load.len >= LOCAL_BUFFER_SIZE) {
    spct6100_readmem (cur_dev, load.addr, bufp, LOCAL_BUFFER_SIZE);
    __copy_to_user (load.buf, bufp, LOCAL_BUFFER_SIZE);
    load.addr += LOCAL_BUFFER_SIZE;
    load.buf += LOCAL_BUFFER_SIZE;
    load.len -= LOCAL_BUFFER_SIZE;
  }

  if (load.len > 0) {
    spct6100_readmem (cur_dev, load.addr, bufp, load.len);
    __copy_to_user (load.buf, bufp, load.len);
  }

  spct6100_pacdmp_write_bar3 (cur_dev, START_ADDRESS);

  up (&(cur_dev->sem));
  vfree (bufp);
  return 0;
}

unsigned long long mosaic_area_geam[4][4][6];
void
set_mosaic_area(void *mosaic, void *file, void *dev, int ch_id)
{
  struct spct6100_mosaic *imosaic = (struct spct6100_mosaic *)mosaic;
  struct spct6100_file_private *cur_file = (struct spct6100_file_private *)file;
  spct6100_dev *cur_dev = (spct6100_dev *)dev;
  int card_id = cur_file->index;
  int total_num = imosaic->num;
  int lx,ly,rx,ry;
  int index;
  SPCT_PAYLOAD payload, payload_alter;
  int scale = 2;

  if (qctl_current_values[card_id][ch_id][CSHOWSIZE_INDEX] == 0) {
    scale = 1;
  }

  memset((void *)&payload, 0, sizeof(SPCT_PAYLOAD));
  memset((void *)&payload_alter, 0, sizeof(SPCT_PAYLOAD));
  payload.MASK_SET.TotalMaskNum = total_num;
  for (index = 0; index < total_num; index++) {
    lx = imosaic->x[index];
    ly = imosaic->y[index];
    rx = lx + imosaic->w[index];
    ry = ly + imosaic->h[index];
    lx = lx*scale / 16 * 16;
    ly = ly*scale / 20 * 20 + 1;
    rx = (rx*scale + 15) / 16 * 16;
    ry = (ry*scale + 19) / 20 * 20 + 1;
    *(unsigned int *)((void *)&(payload.MASK_SET.UpL_Mask0) + index*8)
        = (lx & 0xFFFF) | (ly & 0xFFFF) << 16;
    *(unsigned int *)((void *)&(payload.MASK_SET.DownR_Mask0) + index*8)
        = (rx & 0xFFFF) | (ry & 0xFFFF) << 16;

    if (rx == lx || ry == ly)
      payload_alter.MASK_ENAB.MaskEnable[index]=0;
    else
      payload_alter.MASK_ENAB.MaskEnable[index]=1;
  }

  insert_cmd_node(cur_dev, WIN32_PID_MASK_SET,
      sizeof(payload.MASK_SET), ch_id, (unsigned char *)&payload, NULL);

  insert_cmd_node(cur_dev, WIN32_PID_MASK_ENAB,
      sizeof(payload.MASK_ENAB), ch_id, (unsigned char *)&payload_alter, NULL);
}

#ifdef DEBUG_DATA_QUEUE
void
display_data_queue (spct6100_dev *cur_dev, const char *funcname, int line)
{
  unsigned int address;
  unsigned int wptr;
  int i;
  int card_id = cur_dev->index;

  DEBUG ("[%d]%s: line %d\n", card_id, funcname, line);
  address = fip[card_id].dq_base;
  for (i = 0; i < fip[card_id].dq_num; i++) {
    DEBUG ("[%d]data queue %d:", card_id, i);
    DEBUG ("OWNER=0x%02x,", DPACKET_OWNER(address));
    DEBUG ("SERIAL=0x%02x,", DPACKET_SERIAL(address));
    DEBUG ("ID=0x%02x,", DPACKET_ID(address));
    DEBUG ("FLAG=0x%02x,", DPACKET_FLAG(address));
    DEBUG ("COMMON=0x%02x,", DPACKET_COMMON(address));
    DEBUG ("RESERVED=0x%02x,", DPACKET_RESERVED(address));
    DEBUG ("PAYLOADSIZE=0x%02x\n", DPACKET_PAYLOADSIZE(address));
    address += fip[card_id].dq_len;
  }
  DEBUG ("[%d]driver read_ptr=%d\n", card_id,
         (fip[card_id].cur_dq_pos - fip[card_id].dq_base)/fip[card_id].dq_len);

  spct6100_pacdmp_write_bar3(cur_dev, 0x01C80000);
  spct6100_readmem(cur_dev, 0x14, (u8 *)&wptr, 4);
  DEBUG ("[%d]firmware write_ptr=%d\n", card_id, wptr);
  spct6100_pacdmp_write_bar3(cur_dev, PACDMP_EVB_DSP);
}
#endif

/**
 * @brief Interrupt handler for this device
 *
 *        This handler should do the following task:
 *
 *          - handle the live path interrupt
 *            -# clear the interrupt status
 *            -# set the g_DataControl field to new value
 *            -# set the register DPEBUFUSG to new value
 *            -# wake up the corresponding queue if we have live path frame
 *          - handle the compress path interrupt
 *            -# clear the interrupt status
 *            -# wake up the corresponding queue if there are compress
 *               path frame and we have got I frame with SPS header
 *            -# Otherwise clear the FLIP register to drop this frame
 *          - handle the OSD hour interrupt
 *          - handle the OSD daily interrupt
 *
 * @return On success IRQ_HANDLED should be returned
 */
#ifdef __arm__
irqreturn_t
spct6100_interrupt (int irq, void *dev_id, struct pt_regs *regs)
#else
irqreturn_t
spct6100_interrupt (int irq, void *dev_id)
#endif
{
  unsigned int uChanID, packetid, payloadsize;
  u32 handled = 0, index = 0, should_wakeup = 0, should_wakeup_c=0;
  u32 bufptr, bufusg, garbage;
  u32 data32 = 0, vic1st = 0, vic2st = 0;
#ifndef __arm__
  u32 tmp32;
  u32 serialnum = 0;
#endif
  u16 tmp16;
  spct6100_dev *cur_dev;
  int card_id;
  struct spct6100_wait_list_node *lth, *cth;

  /// for working mode 3
  unsigned int old_bar3 = 0;
  unsigned int compress_hack = 0;

  unsigned int begin_address = 0;

#ifdef __arm__
  if (spct6100_irq_shared) {
    int found = 0;
    for (card_id = 0; card_id < g_device_num; card_id++) {
      cur_dev = &g_spct6100_dev[card_id];
      if (cur_dev->status != BOARD_RUNNING) {
        continue;
      }
      spct6100_pacdmp_read_bar3(cur_dev, &old_bar3);
      spct6100_pacdmp_write_bar3(cur_dev, PACDMP_EVB_DSP);

      spct6100_get_reg(cur_dev, REG_STAT_VIC1, (unsigned char *)&vic1st, -1);
      spct6100_get_reg(cur_dev, REG_STAT_VIC2, (unsigned char *)&vic2st, -1);
      if (vic1st || (vic2st & 0x4000)) {
        found = 1;
        break;
      }
      spct6100_pacdmp_write_bar3(cur_dev, old_bar3);
    }
    if (!found) {
      return IRQ_NONE;
    }
    global_irq_mask[card_id]++;
  } else {
    cur_dev = (spct6100_dev *)dev_id;
    card_id = cur_dev->index;

    global_irq_mask[card_id]++;
    spct6100_pacdmp_read_bar3(cur_dev, &old_bar3);
    spct6100_pacdmp_write_bar3(cur_dev, PACDMP_EVB_DSP);

    spct6100_get_reg(cur_dev, REG_STAT_VIC1, (unsigned char *)&vic1st, -1);
    spct6100_get_reg(cur_dev, REG_STAT_VIC2, (unsigned char *)&vic2st, -1);
  }
#else
  cur_dev = (spct6100_dev *)dev_id;
  card_id = cur_dev->index;

  global_irq_mask[card_id]++;
  spct6100_pacdmp_read_bar3(cur_dev, &old_bar3);
  spct6100_pacdmp_write_bar3(cur_dev, PACDMP_EVB_DSP);

  spct6100_get_reg(cur_dev, REG_STAT_VIC1, (unsigned char *)&vic1st, -1);
  spct6100_get_reg(cur_dev, REG_STAT_VIC2, (unsigned char *)&vic2st, -1);
#endif

  lth = g_ldata_wh[card_id];
  cth = g_cdata_wh[card_id];

  if (spct6100_preview_pci && vic1st) {
    for (index = 0; index < NR_CH; index++) {
      if ((vic1st >> (index * 5)) & 0x1) {

        spct6100_set_reg(cur_dev, REG_CLEAR_VIC1,
            (unsigned char *)&data32, index);

        if (g_DataControl[card_id].l_reference == 0) {
          spct6100_get_reg(cur_dev, REG_BUFPTR,
              (unsigned char *)&bufptr, index);
          bufusg = (bufptr == 0x1)?0x4:(bufptr/2);
          garbage = (3*index) + (bufusg/2);
          spct6100_set_reg(cur_dev, REG_BUFUSG,
              (unsigned char *)&bufusg, index);

          g_DataControl[card_id].l_index[index] = garbage;
          should_wakeup = 1;
        }
      }
    }
  }

  if (vic2st & 0x4000) {
    spct6100_set_reg(cur_dev, REG_CLEAR_VIC2,(unsigned char *)&data32, 14);
    handled = IRQ_HANDLED;
    if (global_irq_mask[card_id] > 1) {
      INFO ("irq nested in the same card, %d, %d\n",
            card_id, global_irq_mask[card_id]);
      goto SPCT6100_IRQ_END;
    }

#ifdef DEBUG_DATA_QUEUE
    display_data_queue (cur_dev, __FUNCTION__, __LINE__);
#endif

    begin_address = fip[card_id].cur_dq_pos;
    for (;;) {

      tmp16 = DPACKET_OWNER(fip[card_id].cur_dq_pos);
      if (tmp16 != 0x86) {
        break;
      }

      uChanID = DPACKET_COMMON(fip[card_id].cur_dq_pos);
      packetid = DPACKET_ID(fip[card_id].cur_dq_pos);
      payloadsize = DPACKET_PAYLOADSIZE(fip[card_id].cur_dq_pos);
      switch (packetid) {

        case WIN32_PID_V_LOST_STATUS:
          DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
          ERROR ("Wrong vlost\n");
          break;

        case WIN32_PID_A_RAW:
          DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
          ERROR ("Wrong audio\n");
          break;

        case WIN32_PID_MDECT_RESULT:
          DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
          ERROR ("Wrong md\n");
          break;

        case WIN32_PID_V_CPRS:
//Using av_xx macro instead of spct6100_get_reg to avoid
//lock with userspace process
#undef AV_BASE
#define AV_BASE (fip[card_id].cur_dq_pos+4*1024)

          if (uChanID > 3) {
            uChanID = AV_RESERV1(AV_BASE);
            if (uChanID > 3) {
              ERROR ("Card: %d, Compress Channel ID error %d\n", card_id, uChanID);
              uChanID = 0;
            } else {
              DEBUG ("Card: %d, fix Compress Channel ID %d\n", card_id, uChanID);
            }
          }

          if (!cstream_toggle[card_id]) {
            DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
            break;
          }

          do_gettimeofday (&c_timestamp);

          if (g_DataControl[card_id].c_reference) {
            compress_hack = 1;
            break;
          }

          if (g_DataControl[card_id].c_index[uChanID] != -1 ) {
            compress_hack = 1;
            break;
          }

          g_DataControl[card_id].c_index[uChanID] =
            (((fip[card_id].cur_dq_pos - fip[card_id].dq_base) / fip[card_id].dq_len)
             & 0xFFFF );

          should_wakeup_c = 1;
          break;

        case WIN32_PID_V_RAW:

          if (uChanID > 3) {
            INFO ("Card: %d, Live Channel ID error %d\n", card_id, uChanID);
            uChanID = 0;
          }

          if (!lstream_toggle[card_id]) {
            DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
            break;
          }


          if (g_DataControl[card_id].l_reference) {
            compress_hack = 1;
            break;
          }

          if (g_DataControl[card_id].l_index[uChanID] != -1) {
            compress_hack = 1;
            break;
          }


          if (uChanID != *(int *)(fip[card_id].cur_dq_pos+4*1024)) {
            INFO ("Tricky Live Chan ID: %d, %d\n",uChanID,
                  *(int *)(fip[card_id].cur_dq_pos+4*1024));
          }

          g_DataControl[card_id].l_index[uChanID] =
            (fip[card_id].cur_dq_pos
             - fip[card_id].dq_base) / fip[card_id].dq_len;

          should_wakeup = 1;

          break;

        case WIN32_PID_ACK_FROM_DSP:
        case WIN32_PID_ACK:
#ifndef __arm__
          tmp32 = DPACKET_PAYLOADSIZE(fip[card_id].cur_dq_pos);
          serialnum = DPACKET_SERIAL(fip[card_id].cur_dq_pos);
          INFO ("Got an ACK, payloadsize: %d, %d\n", tmp32,serialnum);
          INFO ("Contents at 4K: %d, %d, %d, %d, %d, %d, %d, %d\n",
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 0),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 1),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 2),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 3),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 4),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 5),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 6),
              *(unsigned char *)(fip[card_id].cur_dq_pos+4*1024 + 7));
          INFO ("Contents at Payload: %d, %d, %d, %d, %d, %d, %d, %d\n",
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 0),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 1),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 2),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 3),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 4),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 5),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 6),
              *(unsigned char *)(fip[card_id].cur_dq_pos+sizeof(struct pheader) + 7));
#endif
          DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
          break;

        default:
          DPACKET_OWNER(fip[card_id].cur_dq_pos) = 0x61;
          break;
      }

      if (compress_hack){
        break;
      } else {
        fip[card_id].cur_dq_pos = fip[card_id].cur_dq_pos + fip[card_id].dq_len;
        if (fip[card_id].cur_dq_pos == fip[card_id].dq_end) {
          fip[card_id].cur_dq_pos = fip[card_id].dq_base;
        }

        if (fip[card_id].cur_dq_pos == begin_address)
          break;

      }
    }

  }

  if (should_wakeup_c) {
    while (cth != NULL) {
      if (cth->cur_file->status == PSTATUS_STREAMON) {
        cth->data_flag = 1;
        g_DataControl[card_id].c_reference++;
        wake_up_interruptible(cth->p);
      }
      cth = cth->next;
    }

    if (g_DataControl[card_id].c_reference == 0) {
      for (index = 0; index < NR_CH; index++) {
        if ( g_DataControl[card_id].c_index[index] == -1)
          continue;
        DPACKET_OWNER(fip[card_id].dq_base+
            fip[card_id].dq_len*g_DataControl[card_id].c_index[index]) = 0x61;
        g_DataControl[card_id].c_index[index] = -1;
      }
    }
  }

  if (should_wakeup) {
    while (lth != NULL) {
      if (lth->cur_file->status == PSTATUS_STREAMON){
        lth->data_flag = 1;
        g_DataControl[card_id].l_reference++;

        if (spct6100_preview_pci) {
          for (index = 0; index < NR_CH; index++) {
            lth->cur_file->l_index[index] = g_DataControl[card_id].l_index[index];
          }
        }
        wake_up_interruptible(lth->p);
      }
      lth = lth->next;
    }
  }

SPCT6100_IRQ_END:
  spct6100_pacdmp_write_bar3(cur_dev, old_bar3);
  global_irq_mask[card_id]--;

  if (handled)
    return IRQ_HANDLED;
  else
    return IRQ_NONE;
}

/**
 * @brief PCI release hook
 *
 *        Do nothing now
 *
 */

static void __devexit
spct6100_pci_remove_one (struct pci_dev *pdev)
{
  /* Nothing */
}

void trace_process(unsigned long arg)
{
  unsigned long j;
  struct file *file = (struct file *)arg;
  struct spct6100_file_private *cur_file =
               (struct spct6100_file_private *)(file->private_data);
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  if (cur_file->qnode->data_flag == 1) {
    cur_file->qnode->data_flag = 0;
    g_DataControl[card_id].c_reference--;

    if (cur_file->status == PSTATUS_STREAMON) {
      spct6100_vidioc_streamoff_compress(file, cur_file,
                                         V4L2_BUF_TYPE_VIDEO_CAPTURE);
    }
  } else {
    j = jiffies;
    cur_file->trace.expires = j + (cur_file->expires / 1000) * HZ;
    add_timer(&(cur_file->trace));
  }
}

/**
 * @brief Virtual live device open hook
 *
 *          - Set the private_data field of file
 *          - Add the counter for user
 *          - Print some debug info
 *          - Initialize some video format related varibale
 * @return 0 on success
 */
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
spct6100_open_live (struct file *file)
#else
spct6100_open_live (struct inode *inode, struct file *file)
#endif
{
  int index;
  struct spct6100_file_private *cur_file;
  spct6100_dev *cur_dev = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
  u32 minornum = iminor(file->f_dentry->d_inode);
#else
  u32 minornum = iminor(inode);
#endif
  int card_id;
  unsigned int temp1;
  unsigned int temp2;
  unsigned int temp3;
  unsigned int temp4;
  unsigned char cpacket[CPACKET_LEN];

  for (index = 0; index < g_device_num; index++) {
    if (g_spct6100_dev[index].v4l2_live->minor == minornum) {
      cur_dev = &g_spct6100_dev[index];
      break;
    }
  }
  if (cur_dev == NULL) {
    return -ENODEV;
  }
  card_id = cur_dev->index;
  cur_dev->live_users++;
  cur_file = kzalloc(sizeof(struct spct6100_file_private), GFP_KERNEL);
  cur_file->cur_dev = cur_dev;
  cur_file->type = LIVE_PATH;
  cur_file->input = CHANNEL_ALL;
  cur_file->pid = g_lfile_id[card_id]++;
  cur_file->totaluser = &(cur_dev->live_users);
  cur_file->status = PSTATUS_OPEN;
  init_waitqueue_head(&(cur_file->data_queue));
  file->private_data = cur_file;

  if (qtable_driver_initilized[card_id] == 1) {
    if (!spct6100_qtable_initialized[card_id]) {
      fip[card_id].dq_base = (unsigned long)(cur_dev->master_buf_logic)
                             + SPCT6100_PRIVATE_SPACE;
      fip[card_id].dq_num  = DPACKET_NUM;
      fip[card_id].cq_len  = CPACKET_LEN;
      spct6100_get_reg(cur_dev, REG_DQ_PKTLEN,
          CP(&(fip[card_id].dq_len)), -1);
      spct6100_get_reg(cur_dev, REG_CQ_BASE,
          CP(&(fip[card_id].cq_base)), -1);
      spct6100_get_reg(cur_dev, REG_CQ_PKTNUM,
          CP(&(fip[card_id].cq_num)), -1);
      spct6100_get_reg(cur_dev, REG_CQ_PKTLEN, CP(&temp1), -1);
      spct6100_get_reg(cur_dev, REG_DQ_BASE, CP(&temp2), -1);
      spct6100_get_reg(cur_dev, REG_DQ_PKTNUM, CP(&temp3), -1);
      spct6100_get_reg(cur_dev, REG_FIRMWARE_STATUS, CP(&temp4), -1);

      INFO ("cq_base: 0x%08x, cq_num: 0x%08x, cq_len: 0x%08x, "
            "dq_base: 0x%08x, dq_num: 0x%08x, dq_len: 0x%08x, "
            "firmware running: %d\n",
            fip[card_id].cq_base,
            fip[card_id].cq_num,
            temp1,
            temp2,
            temp3,
            fip[card_id].dq_len,
            temp4);

      if (fip[card_id].cq_base == 0xFFFFFFFF) {
        ERROR ("Command queue initial failed!\n");
        return -ENODEV;
      }

      if (fip[card_id].cq_num == 0xFFFFFFFF) {
        ERROR ("Command packet number tricky!\n");
        return -ENODEV;
      }

      fip[card_id].cur_cq_pos = fip[card_id].cq_base;
      fip[card_id].cur_dq_pos = fip[card_id].dq_base;
      fip[card_id].cq_end = fip[card_id].cq_base
                            + CPACKET_LEN * fip[card_id].cq_num;
      fip[card_id].dq_end = fip[card_id].dq_base
                            + DPACKET_NUM * fip[card_id].dq_len;

      //spct6100_pacdmp_write_bar3(cur_dev, 0x03F80034);
      //spct6100_readmem(cur_dev, 0, CP(&temp1), 4);

      spct6100_pacdmp_write_bar3(cur_dev, fip[card_id].cq_base);
      for (index = 0; index < fip[card_id].cq_num; index++) {
        //spct6100_readmem(cur_dev, index*fip[card_id].cq_len, cpacket, fip[card_id].cq_len);
        memset(cpacket, 0, CPACKET_LEN);
        CPACKET_OWNER(cpacket) = 0x86;
        spct6100_writemem(cur_dev, index*fip[card_id].cq_len, (u8 *)cpacket,
            fip[card_id].cq_len);
      }

      //init_pid_status(cur_dev);
      command_status_init_flag = 1;
      spct6100_pacdmp_write_bar3(cur_dev, 0x1E000000);

      spct6100_qtable_initialized[card_id] = 1;
    }
  }

  DEBUG ("SPCT6100-LIVE %d: opened, users: %d\n",
         iminor(file->f_dentry->d_inode), cur_dev->live_users);
  return 0;
}

/**
 * @brief Virtual compress device open hook
 *
 *        Almost the same with that of live device, except:
 *
 *          - Clear the FLIP register to drop those outdated
 *                     compress path frame in the DMA buffer
 *
 * @return 0 on success
 */
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
spct6100_open_compress (struct file *file)
#else
spct6100_open_compress (struct inode *inode, struct file *file)
#endif
{
  int index;
  struct spct6100_file_private *cur_file;
  spct6100_dev *cur_dev = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
  u32 minornum = iminor(file->f_dentry->d_inode);
#else
  u32 minornum = iminor(inode) ;
#endif
  int card_id;
  unsigned int temp1;
  unsigned int temp2;
  unsigned int temp3;
  unsigned int temp4;
  unsigned char cpacket[CPACKET_LEN];

  for (index = 0; index < g_device_num; index++) {
    if (g_spct6100_dev[index].v4l2_compress->minor == minornum) {
      cur_dev = &g_spct6100_dev[index];
      break;
    }
  }
  if (cur_dev == NULL) {
    return -ENODEV;
  }

  card_id = cur_dev->index;
  cur_dev->compress_users++;
  cur_file = kzalloc(sizeof(struct spct6100_file_private), GFP_KERNEL);
  cur_file->cur_dev = cur_dev;
  cur_file->type = COMPRESS_PATH;
  cur_file->input = CHANNEL_ALL;
  cur_file->expires = spct6100_process_expire;
  cur_file->pid = g_cfile_id[card_id]++;
  cur_file->status = PSTATUS_OPEN;
  cur_file->totaluser = &(cur_dev->compress_users);
  init_waitqueue_head(&(cur_file->data_queue));
  file->private_data = cur_file;

  if (qtable_driver_initilized[card_id] == 1) {
    if (!spct6100_qtable_initialized[card_id]) {
      fip[card_id].dq_base = (unsigned long)(cur_dev->master_buf_logic)
                             + SPCT6100_PRIVATE_SPACE;
      fip[card_id].dq_num  = DPACKET_NUM;
      fip[card_id].cq_len  = CPACKET_LEN;
      spct6100_get_reg(cur_dev, REG_DQ_PKTLEN,
          CP(&(fip[card_id].dq_len)), -1);
      spct6100_get_reg(cur_dev, REG_CQ_BASE,
          CP(&(fip[card_id].cq_base)), -1);
      spct6100_get_reg(cur_dev, REG_CQ_PKTNUM,
          CP(&(fip[card_id].cq_num)), -1);
      spct6100_get_reg(cur_dev, REG_CQ_PKTLEN, CP(&temp1), -1);
      spct6100_get_reg(cur_dev, REG_DQ_BASE, CP(&temp2), -1);
      spct6100_get_reg(cur_dev, REG_DQ_PKTNUM, CP(&temp3), -1);
      spct6100_get_reg(cur_dev, REG_FIRMWARE_STATUS, CP(&temp4), -1);

      INFO ("cq_base: 0x%08x, cq_num: 0x%08x, cq_len: 0x%08x, "
            "dq_base: 0x%08x, dq_num: 0x%08x, dq_len: 0x%08x, "
            "firmware running: %d\n",
            fip[card_id].cq_base,
            fip[card_id].cq_num,
            temp1,
            temp2,
            temp3,
            fip[card_id].dq_len,
            temp4);

      if (fip[card_id].cq_base == 0xFFFFFFFF) {
        ERROR ("Command queue initial failed!\n");
        return -ENODEV;
      }

      if (fip[card_id].cq_num == 0xFFFFFFFF) {
        ERROR ("Command packet number tricky!\n");
        return -ENODEV;
      }

      fip[card_id].cur_cq_pos = fip[card_id].cq_base;
      fip[card_id].cur_dq_pos = fip[card_id].dq_base;
      fip[card_id].cq_end = fip[card_id].cq_base
        + CPACKET_LEN * fip[card_id].cq_num;
      fip[card_id].dq_end = fip[card_id].dq_base
        + DPACKET_NUM * fip[card_id].dq_len;

      //spct6100_pacdmp_write_bar3(cur_dev, 0x03F80034);
      //spct6100_readmem(cur_dev, 0, CP(&temp1), 4);

      spct6100_pacdmp_write_bar3(cur_dev, fip[card_id].cq_base);
      for (index = 0; index < fip[card_id].cq_num; index++) {
        //spct6100_readmem(cur_dev, index*fip[card_id].cq_len, cpacket, fip[card_id].cq_len);
        memset(cpacket, 0, CPACKET_LEN);
        CPACKET_OWNER(cpacket) = 0x86;
        spct6100_writemem(cur_dev, index*fip[card_id].cq_len, (u8 *)cpacket,
            fip[card_id].cq_len);
      }

      //init_pid_status(cur_dev);
      command_status_init_flag = 1;
      spct6100_pacdmp_write_bar3(cur_dev, 0x1E000000);

      spct6100_qtable_initialized[card_id] = 1;
    }
  }

  DEBUG ("SPCT6100-COMPRESS %d: opened, users: %d\n",
         iminor(file->f_dentry->d_inode), cur_dev->compress_users);
  return 0;
}

/**
 * @brief V4L2 release hook
 *
 *        Initialize the .release field of struct video_device to
 *        this dummy function to remove the warning
 *
 */
void spct6100_dummy_video_device_release(struct video_device *arg)
{
}

/**
 * @brief V4L2 enum_fmt hook (live)
 *
 *        Fill the fields of the passed in v4l2_fmtdesc structure
 *
 * @param f The .index field has been set by the user, driver should set
 *          the others using the index.
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_enum_fmt_cap(struct file *file, void *priv,
                             struct v4l2_fmtdesc *f)
{
  struct spct6100_file_private *cur_file =
             (struct spct6100_file_private *)priv;
  if (f->index>0
      || f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    return -EINVAL;

  if (cur_file->type == LIVE_PATH) {
    if (spct6100_preview_pci) {
      strlcpy(f->description, SPCT_Format_422.name, sizeof(f->description));
      f->pixelformat = SPCT_Format_422.fourcc;
    } else {
      strlcpy(f->description, SPCT_Format_420.name, sizeof(f->description));
      f->pixelformat = SPCT_Format_420.fourcc;
    }
  } else {
    strlcpy(f->description, "H.264", sizeof(f->description));
    f->pixelformat = v4l2_fourcc('H','2','6','4');
    f->flags = V4L2_FMT_FLAG_COMPRESSED;

  }
  return 0;
}

/**
 * @brief V4L2 g_fmt hook
 *
 *        Return the current format setting
 *
 * @param f Drver should fill all those fields of this structure
 *          to the current working value
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_g_fmt_cap (struct file *file, void *priv,
                                   struct v4l2_format *f)
{
  struct spct6100_file_private *cur_file =
      (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int showsize, card_id = cur_dev->index, ch_id = cur_file->input;

  if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    return -EINVAL;

  if (cur_file->type == LIVE_PATH) {
    if (ch_id != CHANNEL_ALL)
      showsize = qctl_current_values[card_id][ch_id][LSHOWSIZE_INDEX];
    else
      showsize = qctl_current_values[card_id][0][LSHOWSIZE_INDEX];
  } else {

    if (ch_id != CHANNEL_ALL) {

      showsize = qctl_current_values[card_id]
                                    [ch_id]
                                    [CSHOWSIZE_INDEX];
    } else {

      showsize = qctl_current_values[card_id]
                                    [0]
                                    [CSHOWSIZE_INDEX];
    }
  }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
  if (cur_dev->current_norm == V4L2_STD_NTSC_M) {
#else
  if (spct6100_v4l2_live[card_id].current_norm == V4L2_STD_NTSC_M) {
#endif

    f->fmt.pix.width        = SIZE_TABLE_NTSC[showsize].width;
    f->fmt.pix.height       = SIZE_TABLE_NTSC[showsize].height;
  } else {

    f->fmt.pix.width        = SIZE_TABLE_PAL[showsize].width;
    f->fmt.pix.height       = SIZE_TABLE_PAL[showsize].height;
  }

  f->fmt.pix.field        = V4L2_FIELD_NONE;
  if (spct6100_preview_pci) {
    f->fmt.pix.pixelformat  = V4L2_PIX_FMT_UYVY;
    f->fmt.pix.sizeimage    = f->fmt.pix.height * f->fmt.pix.width * 2;
  } else {
    f->fmt.pix.pixelformat  = V4L2_PIX_FMT_YUV420;
    f->fmt.pix.sizeimage    = f->fmt.pix.height * f->fmt.pix.width * 3 / 2;
  }
  f->fmt.pix.bytesperline = f->fmt.pix.width;

  return (0);
}

/**
 * @brief V4L2 try_fmt hook
 *
 *        Validate the format passed in, but do not really
 *        change the current format of the device
 *        Driver should reset those illegal field of f to the limit value
 *
 * @param f Format want to validate
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_try_fmt_cap (struct file *file, void *priv,
                                     struct v4l2_format *f)
{
  unsigned int maxw, maxh;
  unsigned int minw, minh;
  struct spct6100_file_private *cur_file =
  (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index, index;

  if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    return -EINVAL;

  maxw  = 704;
  maxh  = 576;

  minw  = 176;
  minh  = 120;

  f->fmt.pix.field = V4L2_FIELD_NONE;

  if (((spct6100_preview_pci && f->fmt.pix.pixelformat != V4L2_PIX_FMT_UYVY)
      || (!spct6100_preview_pci && f->fmt.pix.pixelformat != V4L2_PIX_FMT_YUV420))
     && cur_dev->stream_status == BOARD_STREAMON)
    return -EINVAL;

  if (f->fmt.pix.pixelformat == V4L2_PIX_FMT_UYVY)
    spct6100_preview_pci = 1;
  else
    spct6100_preview_pci = 0;

  for (index = 0; index < NR_CH; index++)
    qctl_current_values[card_id][index][PREVIEW_INDEX] = spct6100_preview_pci;

  if (f->fmt.pix.width % 176 != 0)
    f->fmt.pix.width = (f->fmt.pix.width / 176 + 1) *176;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
  if (cur_dev->current_norm == V4L2_STD_NTSC_M) {
#else
  if (spct6100_v4l2_live[card_id].current_norm == V4L2_STD_NTSC_M) {
#endif

    if (f->fmt.pix.height % 120 != 0)
      f->fmt.pix.height = (f->fmt.pix.height / 120 + 1) *120;
  } else {

    if (f->fmt.pix.height % 144 != 0)
      f->fmt.pix.height = (f->fmt.pix.height / 144 + 1) *144;
  }
  if (f->fmt.pix.height < minh)
    f->fmt.pix.height = minh;
  if (f->fmt.pix.height > maxh)
    f->fmt.pix.height = maxh;
  if (f->fmt.pix.width < minw)
    f->fmt.pix.width = minw;
  if (f->fmt.pix.width > maxw)
    f->fmt.pix.width = maxw;
  f->fmt.pix.width &= ~0x03;
  if (spct6100_preview_pci)
    f->fmt.pix.bytesperline =
      (f->fmt.pix.width * SPCT_Format_422.depth) >> 3;
  else
    f->fmt.pix.bytesperline =
      (f->fmt.pix.width * SPCT_Format_420.depth) >> 3;
  f->fmt.pix.sizeimage =
    f->fmt.pix.height * f->fmt.pix.bytesperline;
  return 0;
}

/**
 * @brief V4L2 s_fmt hook
 *
 *        Set the current format of the device to the specified one
 *
 * @param f New format want to set
 *
 * @return 0 on success, error code on failure
 */
static int spct6100_vidioc_s_fmt_cap (struct file *file, void *priv,
                                     struct v4l2_format *f)
{
  struct spct6100_file_private *cur_file =
                   (struct spct6100_file_private *)priv;

  struct v4l2_control local_chang_showsize = {
    .id = V4L2_CID_LSHOWSIZE,
    .value = 0,
  };
  int ret  = 0;

  if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    return -EINVAL;

  ret = spct6100_vidioc_try_fmt_cap(file, priv, f);
  if (ret < 0)
    return (ret);

  if (f->fmt.pix.width == 176
      && (f->fmt.pix.height == 120
          || f->fmt.pix.height == 144))
  {

    local_chang_showsize.value = 2;
  } else if (f->fmt.pix.width == 352
             && (f->fmt.pix.height == 240
                 || f->fmt.pix.height == 288)) {

    local_chang_showsize.value = 1;
  } else if (f->fmt.pix.width == 704
             && (f->fmt.pix.height == 240
                 || f->fmt.pix.height == 288))
  {
    local_chang_showsize.value = 3;
  } else if (f->fmt.pix.width == 704
             && (f->fmt.pix.height == 576
                 || f->fmt.pix.height == 480)) {
    local_chang_showsize.value = 0;
  }

  if (cur_file->type == COMPRESS_PATH)
    local_chang_showsize.id = V4L2_CID_CSHOWSIZE;

  return spct6100_vidioc_s_ctrl(file, priv, &local_chang_showsize);

#if 0
  dev->fmt[ch_id]           = &format;
  dev->width[ch_id]         = f->fmt.pix.width;
  dev->height[ch_id]        = f->fmt.pix.height;
  dev->field[ch_id]         = f->fmt.pix.field;
#endif
}

/**
 * @brief V4L2 querycap hook (live path)
 *
 *        Fill those fields of cap
 *
 * @return
 */
int
spct6100_vidioc_querycap(struct file *file, void *priv,
                         struct v4l2_capability *cap)
{
  struct spct6100_file_private *cur_file =
                   (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  strcpy(cap->card, "SPCT6100");
  sprintf(cap->bus_info, "PCI:%s", pci_name(cur_dev->pci));
  cap->version = KERNEL_VERSION(0,0,1);
  cap->capabilities = V4L2_CAP_VIDEO_CAPTURE
          | V4L2_CAP_STREAMING;
  if (cur_file->type == LIVE_PATH) {
    strcpy(cap->driver, "SPCT6100-V4L2-L");
  } else if (cur_file->type == COMPRESS_PATH) {
    strcpy(cap->driver, "SPCT6100-V4L2-C");
  } else {
    return -EINVAL;
  }
  return 0;
}

/**
 * @brief V4L2 reqbufs hook (for live path)
 *
 *        This function will not make buffer ready, it just do some check on:
 *
 *          - Buffer type
 *          - Buffer memory acess method
 *          - Amount of buffers
 *
 *        @warning We have only one live path buffer, mapped
 *                 directly to DMA ring buffer
 *
 * @return 0 on success, error code on failure
 */
int
spct6100_vidioc_reqbufs_live (struct file *file, void *priv,
                              struct v4l2_requestbuffers *p)
{
  int live_buf_num = 4;
  if (p->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {

      ERROR("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE\n");
      return -EINVAL;
    }
  if (p->memory != V4L2_MEMORY_MMAP) {
      ERROR("Only support V4L2_MEMORY_MAP now.\n");
      return -EINVAL;
    }

  if (spct6100_preview_pci)
    live_buf_num = 12;

  if (p->count != live_buf_num) {
    WARNING("The number of live Streaming IO buffer is not optimized.\n");
    p->count = live_buf_num;
  }

  return 0;
}

/**
 * @brief V4L2 querybuf hook (for live path)
 *
 *        This function should fill two important fields of p,
 *        user application will use these two fields to map
 *        the device physical memory buffer to user space
 *
 * @param p.m.offset driver should fill the physical address offset
 *                   of the buffer here
 * @param p.m.length driver should fill the length of the buffer here
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_querybuf_live (struct file *file, void *priv,
                                       struct v4l2_buffer *p)
{
  struct spct6100_file_private *cur_file =
              (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;

  if (p->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE\n");
    return -EINVAL;
  }

  if (p->memory != V4L2_MEMORY_MMAP) {
    ERROR("Only support V4L2_MEMORY_MMAP now\n");
    return -EINVAL;
  }

  p->m.offset = (int)cur_dev->master_buf_phy
                + SPCT6100_PRIVATE_SPACE
                + (p->index * fip[card_id].dq_len)
                + (4*1024);
  p->length = fip[card_id].dq_len - (4*1024);

  if (spct6100_preview_pci) {
    p->m.offset = (int)cur_dev->master_buf_phy
                  + SPCT6100_PRIVATE_SPACE
                  + 4 * 0xA0000
                  + p->index*SPCT6100_DVR_ELEM_SZ;
    p->length = SPCT6100_DVR_ELEM_SZ;
  }
  return 0;
}

/**
 * @brief V4L2 qbuf hook (for live path)
 *
 *        - The qbuf operation on buffer should make buffer
 *          able to receive data again
 *
 *        - After qbuf, hardware can transport data to the buffer
 *
 * @return 0 on success, error code on failure
 */
int
spct6100_vidioc_qbuf_live (struct file *file, void *priv,
                                   struct v4l2_buffer *p)
{
  struct spct6100_file_private *cur_file =
      (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  int live_buf_num = 4;
  int index, temp;

  if (p->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Live buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE\n");
    return -EINVAL;
  }

  if (p->memory != V4L2_MEMORY_MMAP) {
    ERROR("Only support V4L2_MEMORY_MMAP now\n");
    return -EINVAL;
  }

  if (!lstream_toggle[card_id] || cur_file->qnode == NULL) {
    return 0;
  }

  if (down_interruptible(&cur_dev->sem)) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    return -ERESTARTSYS;
  }

  if (g_DataControl[card_id].l_reference > 1) {
    cur_file->qnode->data_flag = 0;
    g_DataControl[card_id].l_reference--;
    up(&(cur_dev->sem));
    return 0;
  } else if (g_DataControl[card_id].l_reference == 0) {
    up(&(cur_dev->sem));
    return 0;
  }
  up(&(cur_dev->sem));

  if (spct6100_preview_pci)
    live_buf_num = 12;

  for (index = 0; index < 32; index=index+8) {
    temp = (p->index >> index) & 0x000000FF;
    if (temp >= live_buf_num) {
      continue;
    }
    if (lstream_toggle[card_id]) {
      if (!spct6100_preview_pci)
        DPACKET_OWNER(fip[card_id].dq_base+ fip[card_id].dq_len*temp) = 0x61;

      if (spct6100_preview_pci)
        cur_file->l_index[index/8] = -1;

      g_DataControl[card_id].l_index[index/8] = -1;
    }
  }

  cur_file->qnode->data_flag = 0;
  g_DataControl[card_id].l_reference = 0;
  return 0;
}

/**
 * @brief V4L2 dqbuf hook (for live path)
 *
 *        The index of the buffer which has data ready should be filled
 *        in the corresponding field fo the argument p
 *
 * @return 0 on success, error code on failure
 */

int
spct6100_vidioc_dqbuf_live (struct file *file, void *priv,
                                    struct v4l2_buffer *p)
{
  struct timeval time;
  unsigned int index = 0, showsize, ret = 0;
  struct spct6100_file_private *cur_file =
  (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index, ch_id = cur_file->input;
  int temp;

  if (p->memory != V4L2_MEMORY_MMAP) {
    ERROR("Only support V4L2_MEMORY_MMAP now\n");
    return -EINVAL;
  }

  ch_id = CHANNEL_ALL;

  if (cur_file->qnode->data_flag == 0) {
    if (file->f_flags & O_NONBLOCK)
      return -EAGAIN;
    else
      wait_event_interruptible(cur_file->data_queue,
                               (cur_file->qnode->data_flag == 1));
  }
  if (cur_file->qnode->data_flag) {
    do_gettimeofday(&time);
    if (p->type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {

      p->input = ch_id;
      if (ch_id == CHANNEL_ALL) {

        p->index = 0;
        p->flags = 0;
        for (index = 0; index < NR_CH; index++) {
          if (spct6100_preview_pci)
            temp = cur_file->l_index[index];
          else
            temp = g_DataControl[card_id].l_index[index];

          if (temp != -1) {
#ifdef DEBUG_FPS
            debug_lframe[card_id][index]++;
#endif
            p->index |= temp << (index*8);
          }
          else
            p->index |= 0xFF << (index*8);
        }
      } else {

        for (index = 0; index < NR_CH; index++) {
          if (index == ch_id)
            continue;
        }
        p->index = g_DataControl[card_id].l_index[ch_id];
      }

      p->timestamp = time;

      showsize = qctl_current_values[card_id][0][LSHOWSIZE_INDEX];
      if (
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
          cur_dev->current_norm == V4L2_STD_NTSC_M
#else
          spct6100_v4l2_live[card_id].current_norm == V4L2_STD_NTSC_M
#endif
         ) {
            p->length = SIZE_TABLE_NTSC[showsize].width *
              SIZE_TABLE_NTSC[showsize].height * 2;
      } else {
            p->length = SIZE_TABLE_PAL[showsize].width *
              SIZE_TABLE_PAL[showsize].height * 2;
      }

      if (!spct6100_preview_pci)
        p->length = p->length * 3 / 4;
    } else {

      ERROR("Live Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
      return -EINVAL;
    }
  }
  return ret;
}

/**
 * @brief V4L2 streamon hook (for live path)
 *
 *        After this operation:
 *
 *          - g_DataControl should be reset
 *          - Live path should work, hardware begin to fill data in the
 *            live path buffer
 *
 * @param i Live path buffer type, used for checking
 *
 * @return 0 on success, error code on failure
 */
int
spct6100_vidioc_streamon_live (struct file *file, void *priv,
                                        enum v4l2_buf_type i)
{
  unsigned int index;
  struct spct6100_wait_list_node *th;
  struct spct6100_file_private *cur_file =
               (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  int ch_id = 0;

  if (i != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Live Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
    return -EINVAL;
  }

  if (cur_file->qnode == NULL) {
    th = (struct spct6100_wait_list_node *)
      kzalloc(sizeof(struct spct6100_wait_list_node), GFP_KERNEL);
    cur_file->qnode = th;
    th->cur_file = cur_file;
    th->p = &(cur_file->data_queue);
    th->pid = cur_file->pid;
    th->next = g_ldata_wh[card_id];
    g_ldata_wh[card_id] = th;
  }

  cur_file->status = PSTATUS_STREAMON;
  if (*(cur_file->totaluser) > 1) {
    INFO ("Live user > 1\n");
    return 0;
  }

#ifdef DEBUG_FPS
  do_gettimeofday(&debug_ltss[card_id]);
#endif

  if (down_interruptible(&(cur_dev->sem))) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    return -ERESTARTSYS;
  }
  g_DataControl[card_id].l_reference = 0;
  for (index = 0; index < CHANNEL_ALL; index++) {
    g_DataControl[card_id].l_index[index] = -1;
  }

  if (cur_dev->stream_status == BOARD_STREAMOFF) {

    cur_dev->stream_status = BOARD_STREAMON;
    spct6100_init_cmd(cur_file, cur_dev);

    do_gettimeofday(&start_time[card_id]);
    spct6100_pacdmp_write_bar3(cur_dev, 0x1E000000);

    if (spct6100_debug) {
      for (ch_id = 0; ch_id < 5; ch_id++) {
        do_gettimeofday(&(cur_dev->l_irq_stime[ch_id]));
        cur_dev->c_irq_stime[ch_id] = cur_dev->l_irq_stime[ch_id];
      }
    }
  }
  lstream_toggle[card_id] = 1;
  up(&(cur_dev->sem));

  if (spct6100_debug) {
    for (ch_id = 0; ch_id < 5; ch_id++) {
      do_gettimeofday(&(cur_dev->l_irq_lost_stime[ch_id]));
    }
  }

  return 0;
}

/**
 * @brief V4L2 streamoff hook (for live path)
 *.
 *        After this operation live path should be closed, hardware
 *        still work but all
 *        live frames will be droped instead of storing in the buffer
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_streamoff_live (struct file *file, void *priv,
                                         enum v4l2_buf_type i)
{
  unsigned int index;
  struct spct6100_wait_list_node *th;
  struct timeval cur_time;
  struct spct6100_file_private *cur_file =
  (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  int ch_id = 0;

  if (i != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Live Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
    return -EINVAL;
  }

  if (cur_file->status != PSTATUS_STREAMON) {
    return 0;
  }

  th = g_ldata_wh[card_id];
  while (th != NULL) {
    if (th->pid == cur_file->pid) {
        if (spct6100_preview_pci)
          break;
        if (cur_file->qnode->data_flag == 1) {
          if (down_interruptible(&cur_dev->sem)) {
            ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
            return -ERESTARTSYS;
          }

          cur_file->qnode->data_flag = 0;
          g_DataControl[card_id].l_reference--;

          if (g_DataControl[card_id].l_reference == 0) {
            for (index = 0; index < NR_CH; index++) {
              if (g_DataControl[card_id].l_index[index] != -1) {
                DPACKET_OWNER(fip[card_id].dq_base
                              + fip[card_id].dq_len
                                *g_DataControl[card_id].l_index[index]) = 0x61;
                g_DataControl[card_id].l_index[index] = -1;
              }
            }
          }
          up(&(cur_dev->sem));
        }
        break;
    }
    th = th->next;
  }

#ifdef DEBUG_FPS
  do_gettimeofday(&debug_ltse[card_id]);
  printk("[%d] lts dur: %d sec\n", card_id, debug_ltse[card_id].tv_sec - debug_ltss[card_id].tv_sec);
  for (index = 0; index < NR_CH; index++)
    printk("  ch %d, frame: %d, fps: %d\n", index, debug_lframe[card_id][index], debug_lframe[card_id][index]/(debug_ltse[card_id].tv_sec - debug_ltss[card_id].tv_sec));
#endif

  cur_file->status = PSTATUS_STREAMOFF;
  if (*(cur_file->totaluser) > 1)
    return 0;

INFO ("L Call streamoff %d!!!!\n", card_id);

  if (down_interruptible(&cur_dev->sem)) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    return -ERESTARTSYS;
  }
  for (index = 0; index < CHANNEL_ALL; index++) {
    if (g_DataControl[card_id].l_index[index] != -1) {
      if (!spct6100_preview_pci)
        DPACKET_OWNER(fip[card_id].dq_base+ fip[card_id].dq_len*g_DataControl[card_id].l_index[index]) = 0x61;
      g_DataControl[card_id].l_index[index] = -1;
    }
  }
  lstream_toggle[card_id] = 0;
  g_DataControl[card_id].l_reference = 0;
  up(&(cur_dev->sem));

  do_gettimeofday(&end_time[card_id]);
  if (spct6100_debug) {
    do_gettimeofday(&cur_time);
    for (ch_id = 0; ch_id < 5; ch_id++) {
      cur_dev->l_irq_lost_htime[ch_id].tv_sec
        += (cur_time.tv_sec - cur_dev->l_irq_lost_stime[ch_id].tv_sec);
      cur_dev->l_irq_lost_htime[ch_id].tv_usec
        += (cur_time.tv_usec - cur_dev->l_irq_lost_stime[ch_id].tv_usec);
    }
  }

  return 0;
}

/**
 * @brief V4L2 reqbufs hook (for compress path)
 *
 *        See that of live path for details
 *
 *        @warning We have two compress path buffer
 *
 * @return 0 on success, error code on failure
 */

int
spct6100_vidioc_reqbufs_compress (struct file *file, void *priv,
                                  struct v4l2_requestbuffers *p)
{
  if (p->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {

    ERROR("Compress buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE\n");
    return -EINVAL;
  }

  if (p->memory != V4L2_MEMORY_MMAP) {

    ERROR("Streaming IO only support mmap now.\n");
    return -EINVAL;
  }

#if 0
  if (p->count != COMPRESS_BUF_NUM) {

    WARNING("The number of Streaming IO buffer for compress path"
      " should be eight for most optimized performance.");
    p->count = COMPRESS_BUF_NUM;
  }
#endif

  if (p->count != 4) {
    WARNING("The number of compress Streaming IO buffer is not optimized.\n");
    p->count = 4;
  }

  return 0;
}

/**
 * @brief V4L2 querybuf hook (for compress path)
 *
 *        See that of live path for details
 *
 * @return 0 on success, error code on failure
 */
int
spct6100_vidioc_querybuf_compress (struct file *file, void *priv,
                                   struct v4l2_buffer *p)
{
  struct spct6100_file_private *cur_file =
              (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;

  if (p->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE\n");
    return -EINVAL;
  }

  if (p->memory != V4L2_MEMORY_MMAP) {
    ERROR("Only support V4L2_MEMORY_MMAP now\n");
    return -EINVAL;
  }

  p->m.offset = (p->index * fip[card_id].dq_len) + (4*1024)
                + (int)cur_dev->master_buf_phy + SPCT6100_PRIVATE_SPACE;
  p->length = fip[card_id].dq_len - (4*1024);
  return 0;
}

/**
 * @brief V4L2 qbuf hook (for compress path)
 *
 *        See that of live path for details
 *
 * @return 0 on success, error code on failure
 */

int
spct6100_vidioc_qbuf_compress (struct file *file, void *priv,
    struct v4l2_buffer *p)
{
  unsigned long j;
  struct spct6100_file_private *cur_file =
    (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  int index = 0;
  int temp = 0;

  if (cur_file->qnode == NULL)
    return 0;

  if (p->type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {

    if (cstream_toggle[card_id]) {
      j = jiffies + (cur_file->expires / 1000) * HZ;
      mod_timer(&(cur_file->trace), j);
      if (cur_file->qnode->data_flag) {
        if (down_interruptible (&cur_dev->sem)) {
          ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
          return -ERESTARTSYS;
        }

        if (g_DataControl[card_id].c_reference > 1) {
          cur_file->qnode->data_flag = 0;
          g_DataControl[card_id].c_reference--;
          up(&(cur_dev->sem));
          return 0;
        }

        for (index = 0; index < 32; index=index+8) {
          temp = (p->index >> index) & 0xFF;
          if (temp >= 4) {
            continue;
          }

          DPACKET_OWNER(fip[card_id].dq_base+ fip[card_id].dq_len*temp) = 0x61;
          g_DataControl[card_id].c_index[index/8] = -1;
        }

        cur_file->qnode->data_flag = 0;
        g_DataControl[card_id].c_reference--;
        up(&(cur_dev->sem));
      }
    }
  } else {

    ERROR("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
    return -EINVAL;
  }
  return 0;
}

/**
 * @brief V4L2 dqbuf hook (for compress path)
 *
 *        See that of live path for details
 *
 * @return 0 on success, error code on failure
 */

int
spct6100_vidioc_dqbuf_compress (struct file *file, void *priv,
                                        struct v4l2_buffer *p)
{
  unsigned int l_flag = 0;
  int index = 0, l_index;
  struct spct6100_file_private *cur_file = (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;

  if (cur_file->qnode->data_flag == 0) {
   if (file->f_flags & O_NONBLOCK)
     return -EAGAIN;
   else
     wait_event_interruptible(cur_file->data_queue,
                             (cur_file->qnode->data_flag==1));
  }

  if (cur_file->qnode->data_flag) {
    if (p->type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {

      p->index = 0;
      p->input = 0;
      p->length = 0;
      p->bytesused = 0;
      p->m.offset = 0;
      p->flags = 0;

        for (index = 0; index < NR_CH; index++) {
          l_index = g_DataControl[card_id].c_index[index];
          l_flag = 0;

          if (l_index != -1) {
#ifdef DEBUG_FPS
            int i;
            unsigned int base = fip[card_id].dq_base
                                + fip[card_id].dq_len * l_index
                                + 4*1024;
            for (i = 0; i < spct6100_cpacket_num; i++) {
              if (AV_VENCODEN(base) && AV_VLEN(base)) {
                debug_cframe[card_id][index]++;
              }
              base += AV_TOTAL(base);
            }
#endif

            p->index  |= l_index << (index * 8);
            p->input  |= index << (index*8);
            p->length |= spct6100_cpacket_num << (index*8);
          } else {
            p->index  |= 0xFF << (index*8);
            p->input  |= 0xFF << (index*8);
          }
        }
        p->timestamp = c_timestamp;
      } else {
        ERROR ("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
        return -EINVAL;
    }
  } else {
    INFO ("dqbuf_compress error, please contact SunplusCT\n");
  }
  return 0;
}

/**
 * @brief V4L2 streamon hook (for compress path)
 *
 *        See that of live path for details
 *
 * @return 0 on success, error code on failure
 */
int
spct6100_vidioc_streamon_compress (struct file *file, void *priv,
    enum v4l2_buf_type i)
{
  unsigned long j;
  struct spct6100_wait_list_node *th;
  unsigned int index;
  struct spct6100_file_private *cur_file =
    (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  int ch_id;

  if (i != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
    return -EINVAL;
  }

  if (cur_file->qnode == NULL) {
    th = (struct spct6100_wait_list_node *)
      kzalloc(sizeof(struct spct6100_wait_list_node), GFP_KERNEL);
    cur_file->qnode = th;
    th->cur_file = cur_file;
    th->p = &(cur_file->data_queue);
    th->pid = cur_file->pid;
    th->next = g_cdata_wh[card_id];
    g_cdata_wh[card_id] = th;
  }

  cur_file->status = PSTATUS_STREAMON;
  init_timer(&(cur_file->trace));
  cur_file->trace.data = (unsigned long)file;
  cur_file->trace.function = trace_process;
  j = jiffies;
  cur_file->trace.expires = j + (cur_file->expires / 1000) * HZ;
  add_timer(&(cur_file->trace));

  if (down_interruptible(&(cur_dev->sem))) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    return -ERESTARTSYS;
  }
  g_DataControl[card_id].c_reference = 0;
  for (index = 0; index < CHANNEL_ALL; index++) {
    g_DataControl[card_id].c_index[index] = -1;
  }

#ifdef DEBUG_FPS
  do_gettimeofday(&debug_ctss[card_id]);
#endif

  if (cur_dev->stream_status == BOARD_STREAMOFF) {

    cur_dev->stream_status = BOARD_STREAMON;
    spct6100_init_cmd(cur_file, cur_dev);

    do_gettimeofday(&start_time[card_id]);
    spct6100_pacdmp_write_bar3(cur_dev, 0x1E000000);
    if (spct6100_debug) {
      for (ch_id = 0; ch_id < 5; ch_id++) {
        do_gettimeofday(&(cur_dev->l_irq_stime[ch_id]));
        cur_dev->c_irq_stime[ch_id] = cur_dev->l_irq_stime[ch_id];
      }
    }
  }
  cstream_toggle[card_id] = 1;
  up(&(cur_dev->sem));

  if (spct6100_debug) {
    for (ch_id = 0; ch_id < 5; ch_id++) {
      do_gettimeofday(&(cur_dev->c_irq_lost_stime[ch_id]));
    }
  }

  return 0;
}

/**
 * @brief V4L2 streamoff hook (for compress path)
 *
 *        See that of live path for details
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_streamoff_compress(struct file *file,
                                   void *priv,
                                   enum v4l2_buf_type i)
{
  struct spct6100_wait_list_node *th;
  struct spct6100_file_private *cur_file =
    (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  struct timeval cur_time;
  int card_id = cur_dev->index;
  int index;

  if (i != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    ERROR("Buffer type should be V4L2_BUF_TYPE_VIDEO_CAPTURE");
    return -EINVAL;
  }

  if (cur_file->status != PSTATUS_STREAMON) {
    return 0;
  }

  del_timer(&(cur_file->trace));

  th = g_cdata_wh[card_id];

  while (th != NULL) {
    if (th->pid == cur_file->pid) {
      if (cur_file->qnode->data_flag == 1) {
        if (down_interruptible(&cur_dev->sem)) {
          ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
          return -ERESTARTSYS;
        }

        cur_file->qnode->data_flag = 0;
        g_DataControl[card_id].c_reference--;

        if (g_DataControl[card_id].c_reference == 0) {
          for (index = 0; index < NR_CH; index++) {
            if (g_DataControl[card_id].c_index[index] != -1) {
              DPACKET_OWNER(fip[card_id].dq_base
                  + fip[card_id].dq_len
                  *g_DataControl[card_id].c_index[index]) = 0x61;
              g_DataControl[card_id].c_index[index] = -1;
            }
          }
        }
        up(&(cur_dev->sem));
      }
      break;
    }
    th = th->next;
  }

  cur_file->status = PSTATUS_STREAMOFF;
  if (*(cur_file->totaluser) > 1)
    return 0;

#ifdef DEBUG_FPS
  do_gettimeofday(&debug_ctse[card_id]);
  printk("[%d] cts dur: %d sec\n", card_id, debug_ctse[card_id].tv_sec - debug_ctss[card_id].tv_sec);
  for (index = 0; index < NR_CH; index++)
    printk("  ch %d, frame: %d, fps: %d\n", index, debug_cframe[card_id][index], debug_cframe[card_id][index]/(debug_ctse[card_id].tv_sec - debug_ctss[card_id].tv_sec));
#endif

  if (down_interruptible(&(cur_dev->sem))) {
    ERROR("%s: line %d down_interruptible failed\n", __FUNCTION__, __LINE__);
    return -ERESTARTSYS;
  }
  for (index = 0; index < CHANNEL_ALL; index++) {
    if (g_DataControl[card_id].c_index[index] != -1) {
      DPACKET_OWNER(fip[card_id].dq_base+ fip[card_id].dq_len*g_DataControl[card_id].c_index[index]) = 0x61;
      g_DataControl[card_id].c_index[index] = -1;
    }
  }
  cstream_toggle[card_id] = 0;
  g_DataControl[card_id].c_reference = 0;
  up(&(cur_dev->sem));

  do_gettimeofday(&cur_time);
  if (spct6100_debug) {
    do_gettimeofday(&cur_time);
    for (index = 0; index < 5; index++) {
      cur_dev->c_irq_lost_htime[index].tv_sec
        += (cur_time.tv_sec - cur_dev->c_irq_lost_stime[index].tv_sec);
      cur_dev->c_irq_lost_htime[index].tv_usec
        += (cur_time.tv_usec - cur_dev->c_irq_lost_stime[index].tv_usec);
    }
  }
  return 0;
}

/**
 * @brief V4L2 queryctrl hook
 *
 *        The control command extension of V4L2
 *
 *        If the command queryed by the user has been implemented by driver,
 *        description of that command should be copyed to the user
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_queryctrl(struct file *file, void *priv,
                          struct v4l2_queryctrl *qc)
{
  int i;

  if ((qc->id < V4L2_CID_BASE
       || qc->id >= V4L2_CID_LASTP1)
      && (qc->id < V4L2_CID_PRIVATE_BASE
          || qc->id >= V4L2_CID_PRIVATE_LASTP1)) {
    return -EINVAL;
  }

  if (qc->id >= V4L2_CID_BASE && qc->id < V4L2_CID_LASTP1) {
    switch (qc->id) {
    case V4L2_CID_BRIGHTNESS:
    case V4L2_CID_CONTRAST:
    case V4L2_CID_SATURATION:
    case V4L2_CID_HUE:
      break;
    default:
      qc->flags |= V4L2_CTRL_FLAG_DISABLED;
      return 0;
    }
  }

  for (i = 0; i < NUM_V4L2_CID; i++) {
    if (qc->id == spct6100_qctrl[i].id) {
      memcpy(qc, &(spct6100_qctrl[i]), sizeof(struct v4l2_queryctrl));
      return 0;
    }
  }
  return -EINVAL;
}

/**
 * @brief V4L2 g_ctrl hook
 *
 *        The field .value of ctrl should be set to the current value
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_g_ctrl(struct file *file, void *priv,
                           struct v4l2_control *ctrl)
{
  int i;
  struct spct6100_file_private *cur_file =
                  (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index, ch_id = cur_file->input;
  struct spct6100_gpio gpio;

  for (i = 0; i < NUM_V4L2_CID; i++) {

    if (ctrl->id != spct6100_qctrl[i].id) {
      continue;
    }

    if (spct6100_qctrl[i].flags & V4L2_CTRL_FLAG_DISABLED) {
      break;
    }

    switch (ctrl->id) {

      case V4L2_CID_CARDNUM:
        ctrl->value = g_device_num;
        break;

      case V4L2_CID_DSPSTATUS:
        spct6100_get_reg(cur_dev, REG_DSP_CBUSY, (unsigned char *)&ctrl->value, -1);
        break;

      case V4L2_CID_MODE:
        ctrl->value = qctl_current_values[card_id][ch_id][MODE_INDEX];
        break;

      case V4L2_CID_PREVIEW:
        ctrl->value = qctl_current_values[card_id][ch_id][PREVIEW_INDEX];
        break;

      case V4L2_CID_AUDIO:
        ctrl->value = qctl_current_values[card_id][ch_id][AUDIO_INDEX];
        break;

      case V4L2_CID_TIMESET:
        ctrl->value = qctl_current_values[card_id][ch_id][TIMESET_INDEX];
        break;

      case V4L2_CID_VLOSTEN:
        ctrl->value = qctl_current_values[card_id][ch_id][VLOSTEN_INDEX];
        break;

      case V4L2_CID_CROPSET:
        if (!user_cropset[card_id][ch_id])
          return -EFAULT;
        if (__copy_to_user((void __user *)(ctrl->value),
                           &(g_cropset[card_id][ch_id]),
                           sizeof(g_cropset[card_id][ch_id])))
          return -EFAULT;
        break;

      case V4L2_CID_GOPSET:
        if (__copy_to_user((void __user *)(ctrl->value),
                           &(g_gopset[card_id][ch_id]),
                           sizeof(g_gopset[card_id][ch_id])))
          return -EFAULT;
        break;

      case V4L2_CID_EQUALITY:
        if (__copy_to_user((void __user *)(ctrl->value),
                           &(g_eqset[card_id][ch_id]),
                           sizeof(g_eqset[card_id][ch_id])))
          return -EFAULT;
        break;

      case V4L2_CID_MOTIONDETECT:
        ctrl->value = qctl_current_values[card_id][ch_id][MOTION_INDEX];
        break;

      case V4L2_CID_MTHRESHOLD:
        ctrl->value = qctl_current_values[card_id][ch_id][MTHRESHOLD_INDEX];
        break;

      case V4L2_CID_MINTERVAL:
        ctrl->value = qctl_current_values[card_id][ch_id][MINTERVAL_INDEX];
        break;

      case V4L2_CID_OSDSET:
        if (__copy_to_user((void __user *)(ctrl->value), &(g_osdset[card_id][ch_id]), sizeof(g_osdset[card_id][ch_id])))
          return -EFAULT;
        break;

      case V4L2_CID_MOSAIC:
        if (__copy_to_user((void __user *)(ctrl->value), &(g_mosaic[card_id][ch_id]), sizeof(g_mosaic[card_id][ch_id])))
          return -EFAULT;
        break;

      case V4L2_CID_UPLOAD:
        return spct6100_upload(ctrl, cur_dev);

      case V4L2_CID_TIMEOUT:
        ctrl->value = cur_file->expires;
        break;

      case V4L2_CID_LSHOWSIZE:
        if (ch_id == CHANNEL_ALL)
          ctrl->value = qctl_current_values[card_id][0][LSHOWSIZE_INDEX];
        else
          ctrl->value = qctl_current_values[card_id][ch_id][LSHOWSIZE_INDEX];
        break;

      case V4L2_CID_CSHOWSIZE:
        if (ch_id == CHANNEL_ALL)
          ctrl->value = qctl_current_values[card_id][0][CSHOWSIZE_INDEX];
        else
          ctrl->value = qctl_current_values[card_id][ch_id][CSHOWSIZE_INDEX];
        break;

      case V4L2_CID_GPIOD:
#if 0
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
            return -EFAULT;

          if (gpio.pinid > 9 || gpio.pinid < 0)
            return -EFAULT;

          g_payload.W_GPIO.pin = gpio.pinid;

          insert_cmd_node(cur_dev, WIN32_PID_W_GPIO, sizeof(g_payload.W_GPIO),
                          -1, (unsigned char *)&g_payload, NULL);
          msleep (5);
        } else {
          return -EFAULT;
        }
#else
        if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
          return -EFAULT;

        if (gpio.pinid > 9 || gpio.pinid < 0)
          return -EFAULT;

        spct6100_readmem(cur_dev, (PACDMP_EVB_GPIO + 0x10) & 0xFFFFFF, (u8 *)&(gpio.value), 4);
        gpio.value = (gpio.value >> gpio.pinid) & 0x1;

        if (__copy_to_user((void __user *)ctrl->value, &gpio, sizeof(struct spct6100_gpio)))
          return -EFAULT;
#endif
        break;


      default:
        if (ch_id == CHANNEL_ALL) {

            ctrl->value = qctl_current_values[card_id][0][i];
        } else {

            ctrl->value = qctl_current_values[card_id][ch_id][i];
        }
        break;
    }
    return (0);
  }

  return -EINVAL;
}

static int
spct6100_vidioc_s_ext_ctrls(struct file *file, void *priv,
                            struct v4l2_ext_controls *controls)
{
#if 0//shengguo
  struct spct6100_file_private *cur_file =
                              (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int index = 0, card_id = cur_dev->index, ch_id = cur_file->input;
  long long motion_value;
  long long motion_area_value = 0;
  int motion_param_value = 0;
  int motion_x, motion_y, motion_width, motion_height;
  int motion_x1, motion_y1, motion_sensitive, motion_count;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  if (controls->ctrl_class != V4L2_CTRL_CLASS_USER) {
    WARNING("Class should be V4L2_CTRL_CLASS_USER for ext controls");
  }
#endif

  if (controls->count < 1) {
    ERROR("No controls included!");
    return -EINVAL;
  }

  switch (controls->controls[0].id) {

    case V4L2_CID_MOSAIC:

      if (controls->count >= 7) {
        ERROR("Every channel can have at most 6 mosaic area");
        return -EINVAL;
      }
      break;

    default:
      break;
  }
  return 0;
}
/**
 * @brief The V4L2 s_ctrl hook
 *
 *        This function will handle all spct6100 control command
 *        for virtual live and compres device
 *
 *        After this function, hardware should work on the
 *        new value user has set
 *
 * @return 0 on success, error code on failure
 */

SPCT_PAYLOAD_EXTERN g_payload;
unsigned char store[64];
static int
spct6100_vidioc_s_ctrl (struct file *file, void *priv,
                            struct v4l2_control *ctrl)
{
  int i, index, j, m, n;
  int l_year, l_month, l_day, l_hour, l_minu, l_sec;
  unsigned int udata32 = 0;
  unsigned int bytemask = 0;
  struct spct6100_file_private *cur_file =
                   (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index, ch_id = cur_file->input;
  SPCT_PAYLOAD payload;
  struct spct6100_gpio gpio;

  for (i = 0; i < NUM_V4L2_CID; i++) {
    if (ctrl->id != spct6100_qctrl[i].id)
      continue;

    if (spct6100_qctrl[i].flags & V4L2_CTRL_FLAG_DISABLED)
      break;

    if (spct6100_qctrl[i].flags & V4L2_CTRL_TYPE_INTEGER) {
      if (ctrl->value < spct6100_qctrl[i].minimum
          || ctrl->value > spct6100_qctrl[i].maximum) {
        return (-ERANGE);
      }
    }

    switch (ctrl->id) {

      case V4L2_CID_PREVIEW:
        udata32 = ctrl->value;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (ch_id != CHANNEL_ALL) {
            if (udata32 == qctl_current_values[card_id][ch_id][PREVIEW_INDEX])
              return 0;
          } else {
            if (udata32 == qctl_current_values[card_id][0][PREVIEW_INDEX])
              return 0;
          }
          return -EFAULT;
        }

        if (ch_id != CHANNEL_ALL) {
          qctl_current_values[card_id][ch_id][PREVIEW_INDEX] = ctrl->value;
        } else {
          for (index = 0; index < NR_CH; index++)
            qctl_current_values[card_id][index][PREVIEW_INDEX] = ctrl->value;
        }

        spct6100_preview_pci = ctrl->value;
        break;

      case V4L2_CID_AUDIO:
        udata32 = ctrl->value;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (ch_id != CHANNEL_ALL) {
            if (udata32 == qctl_current_values[card_id][ch_id][AUDIO_INDEX])
              return 0;
          } else {
            if (udata32 == qctl_current_values[card_id][0][AUDIO_INDEX])
              return 0;
          }
          return -EFAULT;
        }

        if (ch_id != CHANNEL_ALL) {
          qctl_current_values[card_id][ch_id][AUDIO_INDEX] = ctrl->value;
        } else {
          for (index = 0; index < NR_CH; index++)
            qctl_current_values[card_id][index][AUDIO_INDEX] = ctrl->value;
        }
        break;

      case V4L2_CID_TIMESET:
        udata32 = ctrl->value;

        for (index = 0; index < NR_CH; index++)
          qctl_current_values[card_id][index][TIMESET_INDEX] = ctrl->value;

        if (cur_dev->stream_status == BOARD_STREAMON) {
          spct6100_calculate_year_month_day(udata32, &l_year, &l_month, &l_day,
                                            &l_hour, &l_minu, &l_sec);
          l_year = ((l_year / 1000) << 12) + (((l_year / 100) % 10) << 8)
                   + (((l_year / 10) % 10) << 4) + (l_year % 10);
          l_month = ((l_month / 10) << 4) + (l_month % 10);
          l_day = ((l_day / 10) << 4) + (l_day % 10);
          payload.S_TIMER.ymd = (l_year << 16) + (l_month << 8) + (l_day);
          payload.S_TIMER.hms = (l_hour << 16) + (l_minu << 8) + (l_sec);
          payload.S_TIMER.hms =  (0x07000000) | (payload.S_TIMER.hms);

          insert_cmd_node(cur_dev, WIN32_PID_S_TIMER, sizeof(payload.S_TIMER), 0,
                          (unsigned char *)&payload, NULL);

          spct6100_calculate_year_month_day(udata32 + 24*3600, &l_year, &l_month, &l_day,
                                            &l_hour, &l_minu, &l_sec);
          l_year = ((l_year / 1000) << 12) + (((l_year / 100) % 10) << 8)
                    + (((l_year / 10) % 10) << 4) + (l_year % 10);
          l_month = ((l_month / 10) << 4) + (l_month % 10);
          l_day = ((l_day / 10) << 4) + (l_day % 10);
          payload.S_TIMER.ymd = (l_year << 16) + (l_month << 8) + (l_day);

          insert_cmd_node(cur_dev, WIN32_PID_S_TIMER, sizeof(payload.S_TIMER),
                          1, (unsigned char *)&payload, NULL);
        }
        break;

      case V4L2_CID_EENSPEED:
#if 1
        qctl_current_values[card_id][ch_id][EENSPEED_INDEX] = ctrl->value;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.E_V_FPS.fps = ctrl->value;
          insert_cmd_node(cur_dev, WIN32_PID_E_V_FPS_SET, sizeof(payload.E_V_FPS),
                          ch_id,  (unsigned char *)&payload, NULL);
        }
#endif
        break;

      case V4L2_CID_LSHOWSIZE:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          /* Can not set after BOARD_STREAMON. */
          if (ch_id != CHANNEL_ALL) {
            if (ctrl->value == qctl_current_values[card_id][ch_id][LSHOWSIZE_INDEX])
              return 0;
          } else {
            if (ctrl->value == qctl_current_values[card_id][0][LSHOWSIZE_INDEX])
              return 0;
          }
          return -EFAULT;
        } else {
          qctl_current_values[card_id][ch_id][LSHOWSIZE_INDEX] = ctrl->value;
        }
        break;

      case V4L2_CID_CSHOWSIZE:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          /* Can not set after BOARD_STREAMON. */
          if (ch_id != CHANNEL_ALL) {
            if (ctrl->value == qctl_current_values[card_id][ch_id][CSHOWSIZE_INDEX])
              return 0;
          } else {
            if (ctrl->value == qctl_current_values[card_id][0][CSHOWSIZE_INDEX])
              return 0;
          }
          return -EFAULT;
        } else {
          qctl_current_values[card_id][ch_id][CSHOWSIZE_INDEX] = ctrl->value;
        }
        break;

      case V4L2_CID_WATCHEN:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.WATDOG_ENAB.enable = ctrl->value;
          if (ch_id == CHANNEL_ALL) {
            for (index = 0; index < CHANNEL_ALL; index++)
              insert_cmd_node(cur_dev, WIN32_PID_WATDOG_ENAB, sizeof(payload.WATDOG_ENAB),
                              index, NULL, NULL);
          } else {
            insert_cmd_node(cur_dev, WIN32_PID_WATDOG_ENAB, sizeof(payload.WATDOG_ENAB),
                            ch_id, NULL, NULL);
          }
        } else {
          return -EFAULT;
        }
        break;

      case V4L2_CID_WATCHSET:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          struct spct6100_watchset watchset;
          if (__copy_from_user(&watchset, (void __user *)ctrl->value, sizeof(struct spct6100_watchset)))
            return -EFAULT;

          payload.WATDOG_SET.timeout = watchset.timeout;
          if (cur_dev->stream_status == BOARD_STREAMON) {
            if (ch_id == CHANNEL_ALL) {
              for (index = 0; index < CHANNEL_ALL; index++)
                insert_cmd_node(cur_dev, WIN32_PID_WATDOG_SET, sizeof(payload.WATDOG_SET),
                                index, NULL, NULL);
            } else {
              insert_cmd_node(cur_dev, WIN32_PID_WATDOG_SET, sizeof(payload.WATDOG_SET),
                              ch_id, NULL, NULL);
            }
          }
        } else {
          return -EFAULT;
        }
        break;

      case V4L2_CID_GETCLR:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (ch_id == CHANNEL_ALL) {
            for (index = 0; index < CHANNEL_ALL; index++)
              insert_cmd_node(cur_dev, WIN32_PID_V_GET_CLR, -1, index, NULL, NULL);
          } else
            insert_cmd_node(cur_dev, WIN32_PID_V_GET_CLR, -1, ch_id, NULL, NULL);
        } else {
          return -EFAULT;
        }
        break;

      case V4L2_CID_REQISEQ:
        payload.REQ_ISEQ.count = ctrl->value;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (ch_id == CHANNEL_ALL) {
            for (index = 0; index < CHANNEL_ALL; index++)
              insert_cmd_node(cur_dev, WIN32_PID_E_V_REQ_SEQ, 0x4, index,
                              (unsigned char *)&payload, NULL);
          } else
            insert_cmd_node(cur_dev, WIN32_PID_E_V_REQ_SEQ, 0x4, ch_id,
                            (unsigned char *)&payload, NULL);
        }
        break;

      case V4L2_CID_CROPSET:
        if (ch_id >= CHANNEL_ALL)
          return -EFAULT;

        if (__copy_from_user(&tmp_cropset, (void __user *)ctrl->value, sizeof(struct spct6100_cropset)))
          return -EFAULT;

        g_cropset[card_id][ch_id] = tmp_cropset;
        user_cropset[card_id][ch_id] = 1;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.V_CROP.crop =   ((tmp_cropset.start_y & 0x300) >> 2)
                                | ((tmp_cropset.height & 0x300) >> 4)
                                | ((tmp_cropset.start_x & 0x300) >> 6)
                                | ((tmp_cropset.width & 0x300) >> 8);
          payload.V_CROP.v_delay = tmp_cropset.start_y & 0xff;
          payload.V_CROP.v_active = tmp_cropset.height & 0xff;
          payload.V_CROP.h_delay = tmp_cropset.start_x & 0xff;
          payload.V_CROP.h_active = tmp_cropset.width & 0xff;
          insert_cmd_node(cur_dev, WIN32_PID_V_CROPPING, sizeof(payload.V_CROP), ch_id,
                          (unsigned char *)&payload, NULL);
        }
        break;

      case V4L2_CID_GOPSET:
        if (ch_id >= CHANNEL_ALL)
          return -EFAULT;

        if (__copy_from_user(&tmp_gopset, (void __user *)ctrl->value, sizeof(struct spct6100_gopset)))
          return -EFAULT;

        g_gopset[card_id][ch_id] = tmp_gopset;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.E_V_GOP_SET.GOP = ((tmp_gopset.gopb & 0xFFFF) << 16)
                                     | (tmp_gopset.gopp & 0xFFFF);
          insert_cmd_node(cur_dev, WIN32_PID_E_V_GOP_SET, 0x4, ch_id,
                          (unsigned char *)&payload, NULL);
        }
        break;

      case V4L2_CID_MOTIONDETECT:
        if (ch_id >= CHANNEL_ALL)
          return -EFAULT;

        qctl_current_values[card_id][ch_id][MOTION_INDEX] = ctrl->value;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.MD_SETTING.md_thold = qctl_current_values[card_id][ch_id][MTHRESHOLD_INDEX];
          payload.MD_SETTING.md_inter = qctl_current_values[card_id][ch_id][MINTERVAL_INDEX];

          insert_cmd_node(cur_dev, WIN32_PID_MDECT_SET, sizeof(payload.MD_SETTING),
                          ch_id, (unsigned char *)&payload, NULL);

          payload.MD_ENABLE.enable = ctrl->value;
          insert_cmd_node(cur_dev, WIN32_PID_MDECT_ENAB, sizeof(payload.MD_ENABLE),
                          ch_id,  (unsigned char *)&payload, NULL);
        }
        break;

      case V4L2_CID_OSDSET:
        {
          unsigned int pos_x, pos_y;

          if (ch_id >= CHANNEL_ALL)
            return -EFAULT;

          if (__copy_from_user(&tmp_osdset, (void __user *)ctrl->value, sizeof(struct spct6100_osdset)))
            return -EFAULT;

#if 0
          if (tmp_osdset.tran > 45)
            return -EFAULT;
#endif

          g_osdset[card_id][ch_id] = tmp_osdset;

          if (cur_dev->stream_status == BOARD_STREAMON) {

#if 1
            payload.OSD_SET.translucent = osd_color_table[tmp_osdset.acolor];
#else
            payload.OSD_SET.translucent = 0 << 14;
            payload.OSD_SET.translucent |= 3 << 12;
            payload.OSD_SET.translucent |= (tmp_osdset.acolor & 0x7) << 6;
            payload.OSD_SET.translucent |= 1 << 0;
#endif
            if (tmp_osdset.timer_enable) {
              payload.OSD_SET.Osd1position = 0x000a000a;// (osd1 position y, osd0 position x)
              payload.OSD_SET.year =   0x8C8D8E8F;//time1
              payload.OSD_SET.month =  0x2F8A8B2F;//time2
              payload.OSD_SET.secday = 0x850C8889;//time4
              payload.OSD_SET.minsec = 0x82833A84;//time5
              payload.OSD_SET.hr =     0x3080813A;
            }

            if (tmp_osdset.osd_enable) {
              pos_x = tmp_osdset.osd_pos_x & 0x0000FFFF;
              pos_y = tmp_osdset.osd_pos_y & 0x0000FFFF;
              payload.OSD_SET.Osd0position = (pos_y << 16) | pos_x;
              spct6100_osd_str(payload.OSD_SET.str, tmp_osdset.str);
            } else {
              payload.OSD_SET.Osd0position = 0x000a01a0;//pos_y << 16 | pos_x;
              spct6100_osd_str(payload.OSD_SET.str, "dummydummydummydummy");
            }

            insert_cmd_node(cur_dev, WIN32_PID_OSD_SET, sizeof(payload.OSD_SET), ch_id,
                            (unsigned char *)&payload, NULL);

            if (tmp_osdset.timer_enable)
              payload.OSD_ENAB.Osd1Enable = 1;
            else
              payload.OSD_ENAB.Osd1Enable = 0;

            if (tmp_osdset.osd_enable)
              payload.OSD_ENAB.Osd0Enable = 1;
            else
              payload.OSD_ENAB.Osd0Enable = 0;

            insert_cmd_node(cur_dev, WIN32_PID_OSD_ENAB, sizeof(payload.OSD_ENAB),
                            ch_id,  (unsigned char *)&payload, NULL);
          }
          break;
        }

      case V4L2_CID_VLOSTEN:
        qctl_current_values[card_id][ch_id][VLOSTEN_INDEX] = ctrl->value;

        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.V_LOST_ENAB.enable = ctrl->value;
          if (ch_id == CHANNEL_ALL) {
            for (index = 0; index < CHANNEL_ALL; index++)
                insert_cmd_node(cur_dev, WIN32_PID_V_LOST_ENAB,
                    sizeof(payload.V_LOST_ENAB), index, (unsigned char *)&payload, NULL);
          } else {
            insert_cmd_node(cur_dev, WIN32_PID_V_LOST_ENAB,
                sizeof(payload.V_LOST_ENAB), ch_id, (unsigned char *)&payload, NULL);
          }
        }
        break;

      case V4L2_CID_WGPIO:
#if 0
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
            return -EFAULT;

          if (gpio.pinid > 9 || gpio.pinid < 0)
            return -EFAULT;

          g_payload.W_GPIO.pin = gpio.pinid;

          if (__copy_from_user(&(g_payload.W_GPIO.value),
                               (void __user *)(gpio.value),
                               2))
            return -EFAULT;

          insert_cmd_node(cur_dev, WIN32_PID_W_GPIO, sizeof(g_payload.W_GPIO), -1,
                          (unsigned char *)&g_payload, NULL);
          msleep (5);
        } else {
          return -EFAULT;
        }
#else
        if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
          return -EFAULT;

        if (gpio.pinid > 9 || gpio.pinid < 0)
          return -EFAULT;

        spct6100_readmem(cur_dev, (PACDMP_EVB_GPIO + 0x08) & 0xFFFFFF, (u8 *)&udata32, 4);
        if (gpio.value) {
          udata32 |= 1 << gpio.pinid;    /* 1: In */
        } else {
          udata32 &= ~(1 << gpio.pinid); /* 0: Out */
        }
        spct6100_writemem(cur_dev, (PACDMP_EVB_GPIO + 0x08) & 0xFFFFFF, (u8 *)&udata32, 4);
#endif
        break;

      case V4L2_CID_RGPIO:
#if 0
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
            return -EFAULT;

          if (gpio.pinid > 9 || gpio.pinid < 0)
            return -EFAULT;

          g_payload.W_GPIO.pin = gpio.pinid;

          insert_cmd_node(cur_dev, WIN32_PID_W_GPIO, sizeof(g_payload.W_GPIO),
                          -1, (unsigned char *)&g_payload, NULL);
          msleep (5);
        } else {
          return -EFAULT;
        }
#else
        if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
          return -EFAULT;

        if (gpio.pinid > 9 || gpio.pinid < 0)
          return -EFAULT;

        spct6100_readmem(cur_dev, (PACDMP_EVB_GPIO + 0x08) & 0xFFFFFF, (u8 *)&(gpio.value), 4);
        gpio.value = (gpio.value >> gpio.pinid) & 0x1;

        if (__copy_to_user((void __user *)ctrl->value, &gpio, sizeof(struct spct6100_gpio)))
          return -EFAULT;
#endif
        break;

      case V4L2_CID_GPIOD:
#if 0
        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
            return -EFAULT;

          g_payload.SET_GPIO_D.pin = gpio.pinid;
          g_payload.SET_GPIO_D.dir = gpio.value;

          insert_cmd_node(cur_dev, WIN32_PID_SET_GPIO_D, sizeof(g_payload.SET_GPIO_D),
                          -1, (unsigned char *)&g_payload, NULL);
        } else {
          return -EFAULT;
        }
#else
        if (__copy_from_user(&gpio, (void __user *)ctrl->value, sizeof(struct spct6100_gpio)))
          return -EFAULT;

        spct6100_readmem(cur_dev, (PACDMP_EVB_GPIO + 0x00) & 0xFFFFFF, (u8 *)&udata32, 4);
        udata32 |= 1 << gpio.pinid;
        spct6100_writemem(cur_dev, (PACDMP_EVB_GPIO + 0x00) & 0xFFFFFF, (u8 *)&udata32, 4);

        spct6100_readmem(cur_dev, (PACDMP_EVB_GPIO + 0x10) & 0xFFFFFF, (u8 *)&udata32, 4);
        if (gpio.value) {
          udata32 |= 1 << gpio.pinid;    /* 1: In */
        } else {
          udata32 &= ~(1 << gpio.pinid); /* 0: Out */
        }
        spct6100_writemem(cur_dev, (PACDMP_EVB_GPIO + 0x10) & 0xFFFFFF, (u8 *)&udata32, 4);
#endif
        break;

      case V4L2_CID_WI2C:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          struct spct6100_eeprom local_wi2c;

          if (__copy_from_user(&local_wi2c, (void __user *)ctrl->value, sizeof(struct spct6100_eeprom)))
            return -EFAULT;

//          g_payload.W_EEP.deviceid = local_wi2c.deviceid;
 //         g_payload.W_EEP.dst = (unsigned short)local_wi2c.addr;
  //        g_payload.W_EEP.len = local_wi2c.length;

          if (__copy_from_user(g_payload.W_EEP.data, (void __user *)(local_wi2c.buf), local_wi2c.length))
            return -EFAULT;

          insert_cmd_node(cur_dev, WIN32_PID_W_EEP, local_wi2c.length, local_wi2c.deviceid,
                          (unsigned char *)&g_payload, &(local_wi2c.addr));
          msleep(100);
        } else {
          return -EFAULT;
        }
        break;

      case V4L2_CID_RI2C:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          struct spct6100_eeprom local_ri2c;


          if (__copy_from_user(&local_ri2c, (void __user *)ctrl->value, sizeof(struct spct6100_eeprom)))
            return -EFAULT;

          //g_payload.R_EEP.deviceid = local_ri2c.deviceid;
          g_payload.R_EEP.src = local_ri2c.addr;
          g_payload.R_EEP.len = local_ri2c.length;

          insert_cmd_node(cur_dev, WIN32_PID_R_EEP, 0x8, local_ri2c.deviceid, (unsigned char *)&g_payload, NULL);
          msleep(500);

          spct6100_pacdmp_write_bar3(cur_dev, FW_RETURN_ADDR);
          spct6100_readmem(cur_dev, 0, (u8 *)store, 64);
          for (index = 0; index < 64; index++) {
            INFO ("0x%08x,", store[index]);
            if (index % 8 == 7)
              INFO ("\n");
          }

          spct6100_pacdmp_write_bar3(cur_dev, 0x1E000000);

          if (__copy_to_user((void __user *)(local_ri2c.buf), store, local_ri2c.length))
            return -EFAULT;
        } else {
          return -EFAULT;
        }
        break;

      case V4L2_CID_REQIFRAME:
        payload.E_V_REQ_I.I_Counter = 0x1;

        if (cur_dev->stream_status == BOARD_STREAMON) {
          if (ctrl->value == 4) {
            for (index = 0; index < CHANNEL_ALL; index++)
               insert_cmd_node(cur_dev, pid_local_to_external(PID_REQIFRAME),
                               4, index, (unsigned char *)&payload, NULL);
          } else {
            insert_cmd_node(cur_dev, pid_local_to_external(PID_REQIFRAME),
                            4, ctrl->value, (unsigned char *)&payload, NULL);
          }
        }
        break;

      case V4L2_CID_INITENV:
        init_queue_table(cur_dev);
        qtable_driver_initilized[card_id] = 1;
        spct6100_enabledevice(cur_dev, 0x01900000);
        cur_dev->status = BOARD_RUNNING;

        INFO ("Enable device successed!\n");
        break;

      case V4L2_CID_RUNCODE:
        spct6100_enabledevice(cur_dev, ctrl->value);
        break;

      case V4L2_CID_STOPBOARD:
        return spct6100_stop_board(cur_dev);

      case V4L2_CID_TIMEOUT:
        cur_file->expires = ctrl->value;
        break;

      case V4L2_CID_OSDTIMESYNC:
        break;

      case V4L2_CID_DOWNLOAD:
        return spct6100_download(ctrl, cur_dev);
        break;

      case V4L2_CID_UPLOAD:
        return spct6100_upload(ctrl, cur_dev);
        break;

      case V4L2_CID_ALLMODE:
        for (index = 0; index < NR_CH; index++) {
          udata32 = (ctrl->value >> index) & 0x00000001;
          spct6100_set_reg(cur_dev, REG_UMODE,
                           (unsigned char *)&udata32, index);
          for (j = 0; j < NR_CH; j++)
            qctl_current_values[index][j][i] = ctrl->value;
        }
        break;

      case V4L2_CID_MOSAIC:
        {
          if (__copy_from_user(&tmp_mosaicset, (void __user *)ctrl->value, sizeof(struct spct6100_mosaic)))
            return -EFAULT;

          if (tmp_mosaicset.num > 6)
            return -EFAULT;

          g_mosaic[card_id][ch_id] = tmp_mosaicset;
          user_mosaicset[card_id][ch_id] = 1;

          if (cur_dev->stream_status == BOARD_STREAMON) {
            set_mosaic_area(&tmp_mosaicset, cur_file, cur_dev, ch_id);
          }
        }
        break;

#if 0 //shengguo
      case V4L2_CID_DISABLEMOSAIC:

        disable_mosaic_area(cur_file, cur_dev);
        break;
#endif

      case V4L2_CID_MINTERVAL:
      case V4L2_CID_MTHRESHOLD:
        if (ctrl->value < spct6100_qctrl[i].minimum
            || ctrl->value > spct6100_qctrl[i].maximum) {
          return (-ERANGE);
        }
        qctl_current_values[card_id][ch_id][i] = ctrl->value;

        if (cur_dev->stream_status == BOARD_STREAMON) {
          payload.MD_SETTING.md_thold = qctl_current_values[card_id][ch_id][MTHRESHOLD_INDEX];
          payload.MD_SETTING.md_inter = qctl_current_values[card_id][ch_id][MINTERVAL_INDEX];
          insert_cmd_node(cur_dev, WIN32_PID_MDECT_SET, 0x8, ch_id,
                          (unsigned char *)&payload, NULL);

          payload.MD_ENABLE.enable = qctl_current_values[card_id][ch_id][MOTION_INDEX];

          if (ch_id == CHANNEL_ALL) {
            for (index = 0; index < CHANNEL_ALL; index++)
              insert_cmd_node(cur_dev, WIN32_PID_MDECT_ENAB, 0x4, index,
                              (unsigned char *)&payload, NULL);
          } else {
            insert_cmd_node(cur_dev, WIN32_PID_MDECT_ENAB, 0x4, ch_id,
                            (unsigned char *)&payload, NULL);
          }
        }
        break;

      case V4L2_CID_DROPFRAME:
        qctl_current_values[card_id][ch_id][i] = ctrl->value;
        if (cur_dev->stream_status == BOARD_STREAMON) {
          spct6100_dropframe(cur_dev);
        }
        break;

      case V4L2_CID_DSPPING:
        if (cur_dev->stream_status == BOARD_STREAMON) {
          insert_cmd_node(cur_dev, pid_local_to_external(PID_PING), 0, 0, NULL, NULL);
        }
        break;

      default:
        /* FIXME: Every change will influence 4 Channels
         */
        m = card_id;
        n = ch_id;

        switch (ctrl->id) {
          case V4L2_CID_BRIGHTNESS:
          case V4L2_CID_CONTRAST:
          case V4L2_CID_SATURATION:
          case V4L2_CID_HUE:

            if (cur_dev->stream_status != BOARD_STREAMON) {
              break;
            }

            if (n != CHANNEL_ALL) {

              udata32 = ctrl->value;
              if (udata32 == qctl_current_values[m][n][BRIGHT_INDEX + ctrl->id - V4L2_CID_BRIGHTNESS])
                break;
              qctl_current_values[m][n][BRIGHT_INDEX + ctrl->id - V4L2_CID_BRIGHTNESS] = ctrl->value;
            } else {

              for (index = 0; index < NR_CH; index++)
                qctl_current_values[m][index][BRIGHT_INDEX + ctrl->id - V4L2_CID_BRIGHTNESS] = ctrl->value;
              n = n-1;
            }

            bytemask = 0x1 << (ctrl->id - V4L2_CID_BRIGHTNESS);
            spct6100_set_reg(cur_dev, REG_UCOLOR,
                (unsigned char *)&udata32,
                ch_id, (unsigned char *)&udata32, bytemask,
                qctl_current_values[m][n][BRIGHT_INDEX],
                qctl_current_values[m][n][HUE_INDEX],
                qctl_current_values[m][n][CONTRAST_INDEX],
                qctl_current_values[m][n][SATURATION_INDEX]);
            break;

          case V4L2_CID_EQUALITY:
            {
              unsigned int payloadsize = 0;
              if (__copy_from_user(&tmp_eqset, (void __user *)ctrl->value, sizeof(struct spct6100_eqset)))
                return -EFAULT;

              if (tmp_eqset.type == g_eqset[card_id][ch_id].type
                  && ((tmp_eqset.type == 0
                    && tmp_eqset.minbitrate == g_eqset[card_id][ch_id].minbitrate
                    && tmp_eqset.maxbitrate == g_eqset[card_id][ch_id].maxbitrate)
                  || (tmp_eqset.type == 2
                    && tmp_eqset.qpp == g_eqset[card_id][ch_id].qpp
                    && tmp_eqset.qpi == g_eqset[card_id][ch_id].qpi
                    && tmp_eqset.qpb == g_eqset[card_id][ch_id].qpb)))
                break;

              g_eqset[card_id][ch_id] = tmp_eqset;

              if (cur_dev->stream_status == BOARD_STREAMON) {
                payload.E_V_RATE_CTL_TYPE_SET.typesetting = tmp_eqset.type;
                insert_cmd_node(cur_dev, WIN32_PID_E_V_RATE_CTL_TYPE_SET,
                                sizeof(payload.E_V_RATE_CTL_TYPE_SET), ch_id,
                                (unsigned char *)&payload, NULL);
                if (tmp_eqset.type == 0) {
                  payload.E_V_RATE_CTL_BR_SET.minbitrate = tmp_eqset.minbitrate;
                  payload.E_V_RATE_CTL_BR_SET.maxbitrate = tmp_eqset.maxbitrate;
                  payloadsize = sizeof(payload.E_V_RATE_CTL_BR_SET);
                } else if (tmp_eqset.type == 2) {
                  payload.E_V_RATE_CTL_QP_SET.QP_I = tmp_eqset.qpi;
                  payload.E_V_RATE_CTL_QP_SET.QP_P = tmp_eqset.qpp;
                  payload.E_V_RATE_CTL_QP_SET.QP_B = tmp_eqset.qpb;
                  payloadsize = sizeof(payload.E_V_RATE_CTL_QP_SET);
                }
                insert_cmd_node(cur_dev,
                               (tmp_eqset.type == 2) ? WIN32_PID_E_V_RATE_CTL_QP_SET
                                                       : WIN32_PID_E_V_RATE_CTL_BR_SET,
                               payloadsize, ch_id, (unsigned char *)&payload, NULL);
              }
              break;

            }

          case V4L2_CID_MODE:
            {
              v4l2_std_id l_std;
              unsigned int udata32 = ctrl->value;

              if (cur_dev->stream_status == BOARD_STREAMON) {
                if (n != CHANNEL_ALL) {
                  if (udata32 == qctl_current_values[m][n][MODE_INDEX])
                    return 0;
                } else {
                  if (udata32 == qctl_current_values[m][0][MODE_INDEX])
                    return 0;
                }
                return -EFAULT;
              }

              if (ctrl->value == 0)
                l_std = V4L2_STD_NTSC_M;
              else if (ctrl->value == 1)
                l_std = V4L2_STD_PAL_M;
              else
                return -EFAULT;

              if (n != CHANNEL_ALL) {

                if (udata32 == qctl_current_values[m][n][MODE_INDEX])
                  break;
                qctl_current_values[m][n][MODE_INDEX] = ctrl->value;
              } else {

                for (index = 0; index < NR_CH; index++)
                  qctl_current_values[m][index][MODE_INDEX] = ctrl->value;
                n = n-1;
              }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)

              if (l_std != cur_dev->current_norm) {

                cur_dev->current_norm = l_std;
              }

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)

              if (l_std != spct6100_v4l2_live[card_id].current_norm) {

                spct6100_v4l2_live[card_id].current_norm = l_std;
                spct6100_v4l2_compress[card_id].current_norm = l_std;
              }
#else

              if (l_std != spct6100_v4l2_live[card_id].current_norm) {

                spct6100_v4l2_live[card_id].current_norm = l_std;
                spct6100_v4l2_compress[card_id].current_norm = l_std;
              }
#endif
              break;
            }

          case V4L2_CID_STATUS:
            udata32 = ctrl->value;
            if (cur_dev->stream_status == BOARD_STREAMON) {
              if (n != CHANNEL_ALL) {
                if (udata32 == qctl_current_values[m][n][STATUS_INDEX])
                  return 0;
              } else {
                if (udata32 == qctl_current_values[m][0][STATUS_INDEX])
                  return 0;
              }
              return -EFAULT;
            }

            if (n != CHANNEL_ALL) {
              if (udata32 == qctl_current_values[m][n][STATUS_INDEX])
                break;
              qctl_current_values[m][n][STATUS_INDEX] = ctrl->value;
            } else {
              for (index = 0; index < NR_CH; index++)
                qctl_current_values[m][index][STATUS_INDEX] = ctrl->value;
            }
            break;

          default:
            if (cur_dev->stream_status != BOARD_STREAMON) {
              return -EFAULT;
            }

            udata32 = ctrl->value;
#if 0
            DEBUG ("[%d]chan=%d, ctrl->id=(V4L2_CID_PRIVATE_BASE + %d), value=%d\n",
                   card_id, ch_id, ctrl->id - V4L2_CID_PRIVATE_BASE, udata32);
#endif
            for (index = 0; index < NR_CH; index++) {

              if ((ch_id == CHANNEL_ALL)
                  || (ch_id == index)) {
          enum spct6100_reg_index local_index[NUM_V4L2_CID] = {
                  0,
                  0,
                  0,
                  0,
                  0,
                  0,
                  REG_UMODE,
                  REG_USTATUS,
                  REG_UFLAGS,
                  REG_UQUALITY,
                  REG_USHOWSIZE,
                  REG_UENCODERSPEED,
                  REG_ULPFS,
                  REG_ULPBB,
                  REG_ULPS,
                  REG_UMOSAIC,
                };

                spct6100_set_reg(cur_dev,
                    local_index[ctrl->id - V4L2_CID_PRIVATE_BASE],
                    (unsigned char *)&udata32,
                    index,
                    (unsigned char *)&udata32);
              }
            }
            break;
        }
        break;
    }
    return (0);
  }
  return -EINVAL;
}

/**
 * @brief V4L2 enum_input hook
 *
 *        Fill some description field of inp
 *
 * @param inp Field .index is filled by user,
 *            driver should fill other fields
 *            according to this
 *
 * @return 0 on success, error code on failure
 */
static int spct6100_vidioc_enum_input (struct file *file, void *priv,
                                              struct v4l2_input *inp)
{
  struct spct6100_file_private *cur_file =
            (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int card_id = cur_dev->index;
  if (inp->index > CHANNEL_3)
    return -EINVAL;
  inp->type = V4L2_INPUT_TYPE_CAMERA;
  inp->std = V4L2_STD_NTSC_M
             | V4L2_STD_PAL_M;
  inp->tuner = 0;
  if (cur_file->type == LIVE_PATH)
    sprintf(inp->name, "CARD %d CHANNEL %d(L)", card_id, inp->index);
  else if (cur_file->type == COMPRESS_PATH)
    sprintf(inp->name, "CARD %d CHANNEL %d(C)", card_id, inp->index);
  else
    return -EINVAL;
  return (0);
}

/** @brief V4L2 g_input hook
 *
 * @param i Save the current index number here
 *
 * @return 0 on success, error code on failure
 */
static int spct6100_vidioc_g_input(struct file *file, void *priv,
                                                 unsigned int *i)
{
  struct spct6100_file_private *cur_file =
                     (struct spct6100_file_private *)priv;
  int ch_id = cur_file->input;
  *i = ch_id;
  return (0);
}

/** @brief V4L2 s_input hook
 *
 * @param i Index of new input
 *
 * @return 0 on success, error code on failure
 */
static int
spct6100_vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
  struct spct6100_file_private *cur_file =
      (struct spct6100_file_private *)priv;

  if (i > CHANNEL_ALL)
    return -EINVAL;

  cur_file->input = (enum spct6100_input)i;
  return (0);
}

/** @brief V4L2 s_std hook
 *
 * @param i The new std
 *
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
static int
spct6100_vidioc_s_std (struct file *file, void *priv, v4l2_std_id *i)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static int
spct6100_vidioc_s_std (struct file *file, void *priv, v4l2_std_id i)
#else
static int
spct6100_vidioc_s_std (struct file *file, void *priv, v4l2_std_id *i)
#endif
{
  struct spct6100_file_private *cur_file =
  (struct spct6100_file_private *)priv;
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int rc;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
  struct v4l2_control local_chang_mode = {
    .id = V4L2_CID_MODE,
    .value = (*i == V4L2_STD_NTSC_M)?0:1,
  };

  rc = spct6100_vidioc_s_ctrl(file, priv, &local_chang_mode);
  if (rc != 0)
    return rc;

  if (*i != cur_dev->current_norm) {

    cur_dev->current_norm = *i;
  }

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
  int card_id = cur_dev->index;
  struct v4l2_control local_chang_mode = {
    .id = V4L2_CID_MODE,
    .value = (spct6100_tvnorms[i].id == V4L2_STD_NTSC_M)?0:1,
  };

  rc = spct6100_vidioc_s_ctrl(file, priv, &local_chang_mode);
  if (rc != 0)
    return rc;

  if (spct6100_tvnorms[i].id != spct6100_v4l2_live[card_id].current_norm) {

    spct6100_v4l2_live[card_id].current_norm = spct6100_tvnorms[i].id;
    spct6100_v4l2_compress[card_id].current_norm = spct6100_tvnorms[i].id;
  }
#else
  int card_id = cur_dev->index;
  struct v4l2_control local_chang_mode = {
    .id = V4L2_CID_MODE,
    .value = (*i == V4L2_STD_NTSC_M)?0:1,
  };

  rc = spct6100_vidioc_s_ctrl(file, priv, &local_chang_mode);
  if (rc != 0)
    return rc;

  if (*i != spct6100_v4l2_live[card_id].current_norm) {

    spct6100_v4l2_live[card_id].current_norm = *i;
    spct6100_v4l2_compress[card_id].current_norm = *i;
  }
#endif

  return 0;
}

static int
spct6100_vidioc_g_tuner(struct file *file, void *priv, struct v4l2_tuner *a)
{
  return -EINVAL;
}

static int
spct6100_vidioc_s_tuner(struct file *file, void *priv, struct v4l2_tuner *a)
{
  return -EINVAL;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
static int spct6100_ioctl_common(struct inode *inode, struct file *filp,
                                 unsigned int cmd, unsigned long arg)
{
  return video_usercopy(inode, filp, cmd, arg, spct6100_do_ioctl);
}

static int
spct6100_do_ioctl(struct inode *inode, struct file *file,
                  unsigned int cmd, void *arg)
{
  struct spct6100_file_private *cur_file =
               (struct spct6100_file_private *)(file->private_data);
  spct6100_dev *cur_dev = cur_file->cur_dev;
  int ret=0;

  switch(cmd) {
    /* --- capabilities ------------------------------------------ */
    case VIDIOC_QUERYCAP:
      {
        struct v4l2_capability *cap = (struct v4l2_capability*)arg;
        memset(cap, 0, sizeof(*cap));
        ret = spct6100_vidioc_querycap(file, file->private_data, cap);
        break;
      }
      /* --- capture ioctls ---------------------------------------- */
    case VIDIOC_ENUM_FMT:
      {
        struct v4l2_fmtdesc *f = arg;
        ret = spct6100_vidioc_enum_fmt_cap(file, file->private_data, f);
        break;
      }
    case VIDIOC_G_FMT:
      {
        struct v4l2_format *f = (struct v4l2_format *)arg;

        ret = spct6100_vidioc_g_fmt_cap(file, file->private_data, f);
        break;
      }
    case VIDIOC_S_FMT:
      {
        struct v4l2_format *f = arg;

        ret = spct6100_vidioc_s_fmt_cap(file, file->private_data, f);
        break;
      }
    case VIDIOC_TRY_FMT:
      {
        struct v4l2_format *f = arg;
        ret = spct6100_vidioc_try_fmt_cap(file, file->private_data, f);
        break;
      }
    case VIDIOC_REQBUFS:
      {
        struct v4l2_requestbuffers *p = arg;
        if (cur_file->type == LIVE_PATH)
          ret = spct6100_vidioc_reqbufs_live(file, file->private_data, p);
        else if (cur_file->type == COMPRESS_PATH)
          ret = spct6100_vidioc_reqbufs_compress(file, file->private_data, p);
        else
          return -EINVAL;
        break;
      }
    case VIDIOC_QUERYBUF:
      {
        struct v4l2_buffer *p = arg;
#if 1
        if (cur_file->type == LIVE_PATH)
          ret=spct6100_vidioc_querybuf_live(file, file->private_data, p);
        else if (cur_file->type == COMPRESS_PATH)
          ret=spct6100_vidioc_querybuf_compress(file, file->private_data, p);
        else
          return -EINVAL;
#else
        if (cur_file->type == LIVE_PATH || cur_file->type == COMPRESS_PATH)
          ret=spct6100_vidioc_querybuf_live(file, file->private_data, p);
        else
          return -EINVAL;
#endif

        break;
      }
    case VIDIOC_QBUF:
      {
        struct v4l2_buffer *p = arg;
        if (cur_file->type == LIVE_PATH)
          ret = spct6100_vidioc_qbuf_live(file, file->private_data, p);
        else if (cur_file->type == COMPRESS_PATH)
          ret = spct6100_vidioc_qbuf_compress(file, file->private_data, p);
        else
          return -EINVAL;
        break;
      }
    case VIDIOC_DQBUF:
      {
        struct v4l2_buffer *p = arg;
        if (cur_file->type == LIVE_PATH)
          ret = spct6100_vidioc_dqbuf_live(file, file->private_data, p);
        else if (cur_file->type == COMPRESS_PATH)
          ret = spct6100_vidioc_dqbuf_compress(file, file->private_data, p);
        else
          return -EINVAL;
        break;
      }

    case VIDIOC_STREAMON:
      {
        enum v4l2_buf_type p = *(enum v4l2_buf_type *)arg;
        if (cur_file->type == LIVE_PATH)
          ret = spct6100_vidioc_streamon_live(file, file->private_data, p);
        else if (cur_file->type == COMPRESS_PATH)
          ret=spct6100_vidioc_streamon_compress(file, file->private_data, p);
        else
          return -EINVAL;
        break;
      }
    case VIDIOC_STREAMOFF:
      {
        enum v4l2_buf_type p = *(enum v4l2_buf_type *)arg;
        if (cur_file->type == LIVE_PATH)
          ret = spct6100_vidioc_streamoff_live(file, file->private_data, p);
        else if (cur_file->type == COMPRESS_PATH)
          ret=spct6100_vidioc_streamoff_compress(file, file->private_data, p);
        else
          return -EINVAL;
        break;
      }
      /* ---------- tv norms ---------- */
    case VIDIOC_ENUMSTD:
      {
        struct v4l2_standard *e = arg;

        if (e->index>1) {
          ret=-EINVAL;
          break;
        }
        if (e->index == 0)
          ret = v4l2_video_std_construct(e, V4L2_STD_NTSC_M, "NTSC-M");
        else
          ret = v4l2_video_std_construct(e, V4L2_STD_PAL_M, "PAL-M");

        break;
      }
    case VIDIOC_G_STD:
      {
        v4l2_std_id *id = arg;

        *id = cur_dev->current_norm;
        break;
      }
    case VIDIOC_S_STD:
      {
        v4l2_std_id *id = arg;
        ret = spct6100_vidioc_s_std(file, file->private_data, id);
        break;
      }
      /* ------ input switching ---------- */
    case VIDIOC_ENUMINPUT:
      { /* only one input in this sample driver */
        struct v4l2_input *inp = arg;
        ret = spct6100_vidioc_enum_input(file, file->private_data, inp);
        break;
      }
    case VIDIOC_G_INPUT:
      {
        unsigned int *i = arg;
        ret = spct6100_vidioc_g_input(file, file->private_data, i);
        break;
      }
    case VIDIOC_S_INPUT:
      {
        unsigned int *i = arg;

        ret = spct6100_vidioc_s_input(file, file->private_data, *i);
        break;
      }

      /* --- controls ---------------------------------------------- */
    case VIDIOC_QUERYCTRL:
      {
        struct v4l2_queryctrl *qc = arg;
        ret = spct6100_vidioc_queryctrl(file, file->private_data, qc);
        break;
      }

    case VIDIOC_G_CTRL:
      {
        struct v4l2_control *ctrl = arg;
        ret = spct6100_vidioc_g_ctrl(file, file->private_data, ctrl);
        break;
      }

    case VIDIOC_S_CTRL:
      {
        struct v4l2_control *ctrl = arg;
        ret = spct6100_vidioc_s_ctrl(file, file->private_data, ctrl);
        break;
      }

    case VIDIOC_S_EXT_CTRLS:
      {
        struct v4l2_ext_controls *ctrls = arg;
        void __user *user_ptr = NULL;
        void *mbuf = NULL;
        int count = ctrls->count;
        int ctrls_size =  sizeof(struct v4l2_ext_control) * count;
        user_ptr = (void __user *)ctrls->controls;
        mbuf = kmalloc(ctrls_size, GFP_KERNEL);
        copy_from_user(mbuf, user_ptr, ctrls_size);
        ctrls->controls = mbuf;
        ret = spct6100_vidioc_s_ext_ctrls(file, file->private_data, ctrls);
        break;
      }

    default:
      ret=v4l_compat_translate_ioctl(inode, file, cmd, arg,
          (v4l2_kioctl)spct6100_ioctl_common);
  }
  return ret;
}
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
struct v4l2_ioctl_ops spct6100_v4l2_live_ops = {
  .vidioc_enum_fmt_vid_cap  = spct6100_vidioc_enum_fmt_cap,
  .vidioc_g_fmt_vid_cap     = spct6100_vidioc_g_fmt_cap,
  .vidioc_try_fmt_vid_cap   = spct6100_vidioc_try_fmt_cap,
  .vidioc_s_fmt_vid_cap     = spct6100_vidioc_s_fmt_cap,
  .vidioc_querycap          = spct6100_vidioc_querycap,
  .vidioc_reqbufs           = spct6100_vidioc_reqbufs_live,
  .vidioc_querybuf          = spct6100_vidioc_querybuf_live,
  .vidioc_qbuf              = spct6100_vidioc_qbuf_live,
  .vidioc_dqbuf             = spct6100_vidioc_dqbuf_live,
  .vidioc_streamon          = spct6100_vidioc_streamon_live,
  .vidioc_streamoff         = spct6100_vidioc_streamoff_live,
  .vidioc_queryctrl         = spct6100_vidioc_queryctrl,
  .vidioc_g_ctrl            = spct6100_vidioc_g_ctrl,
  .vidioc_s_ctrl            = spct6100_vidioc_s_ctrl,
  .vidioc_s_ext_ctrls       = spct6100_vidioc_s_ext_ctrls,
  .vidioc_enum_input        = spct6100_vidioc_enum_input,
  .vidioc_g_input           = spct6100_vidioc_g_input,
  .vidioc_s_input           = spct6100_vidioc_s_input,
  .vidioc_s_std             = spct6100_vidioc_s_std,
  .vidioc_g_tuner           = spct6100_vidioc_g_tuner,
  .vidioc_s_tuner           = spct6100_vidioc_s_tuner,
};

struct v4l2_ioctl_ops spct6100_v4l2_comp_ops = {
  .vidioc_enum_fmt_vid_cap   = spct6100_vidioc_enum_fmt_cap,
  .vidioc_g_fmt_vid_cap      = spct6100_vidioc_g_fmt_cap,
  .vidioc_try_fmt_vid_cap    = spct6100_vidioc_try_fmt_cap,
  .vidioc_s_fmt_vid_cap      = spct6100_vidioc_s_fmt_cap,
  .vidioc_querycap           = spct6100_vidioc_querycap,
  .vidioc_reqbufs            = spct6100_vidioc_reqbufs_compress,
  .vidioc_querybuf       = spct6100_vidioc_querybuf_compress,
  //.vidioc_querybuf           = spct6100_vidioc_querybuf_live,
  .vidioc_qbuf               = spct6100_vidioc_qbuf_compress,
  .vidioc_dqbuf              = spct6100_vidioc_dqbuf_compress,
  .vidioc_streamon           = spct6100_vidioc_streamon_compress,
  .vidioc_streamoff          = spct6100_vidioc_streamoff_compress,
  .vidioc_queryctrl          = spct6100_vidioc_queryctrl,
  .vidioc_g_ctrl             = spct6100_vidioc_g_ctrl,
  .vidioc_s_ctrl             = spct6100_vidioc_s_ctrl,
  .vidioc_enum_input         = spct6100_vidioc_enum_input,
  .vidioc_g_input            = spct6100_vidioc_g_input,
  .vidioc_s_input            = spct6100_vidioc_s_input,
  .vidioc_s_std              = spct6100_vidioc_s_std,
};
#endif

static struct video_device spct6100_v4l2_live[NR_CARDS] = {
  {
    .name                  = "SPCT6100-V4L2-LIVE",
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
#define VID_TYPE_CAPTURE 1
    .vfl_type                  = VID_TYPE_CAPTURE,
#else
    .type                  = VID_TYPE_CAPTURE
      | VID_TYPE_MPEG_ENCODER,
#endif
    .fops                  = &spct6100_fops_live,
    .minor                 = -1,
    .release               = spct6100_dummy_video_device_release,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
    /* EMPTY, implemented in spct6100_ioctl_common. */
#else
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
    .vidioc_enum_fmt_cap   = spct6100_vidioc_enum_fmt_cap,
    .vidioc_g_fmt_cap      = spct6100_vidioc_g_fmt_cap,
    .vidioc_try_fmt_cap    = spct6100_vidioc_try_fmt_cap,
    .vidioc_s_fmt_cap      = spct6100_vidioc_s_fmt_cap,
    .vidioc_querycap       = spct6100_vidioc_querycap,
    .vidioc_reqbufs        = spct6100_vidioc_reqbufs_live,
    .vidioc_querybuf       = spct6100_vidioc_querybuf_live,
    .vidioc_qbuf           = spct6100_vidioc_qbuf_live,
    .vidioc_dqbuf          = spct6100_vidioc_dqbuf_live,
    .vidioc_streamon       = spct6100_vidioc_streamon_live,
    .vidioc_streamoff      = spct6100_vidioc_streamoff_live,
    .vidioc_queryctrl      = spct6100_vidioc_queryctrl,
    .vidioc_g_ctrl         = spct6100_vidioc_g_ctrl,
    .vidioc_s_ctrl         = spct6100_vidioc_s_ctrl,
    .vidioc_s_ext_ctrls    = spct6100_vidioc_s_ext_ctrls,
    .vidioc_enum_input     = spct6100_vidioc_enum_input,
    .vidioc_g_input        = spct6100_vidioc_g_input,
    .vidioc_s_input        = spct6100_vidioc_s_input,
    .vidioc_s_std          = spct6100_vidioc_s_std,
    .vidioc_g_tuner        = spct6100_vidioc_g_tuner,
    .vidioc_s_tuner        = spct6100_vidioc_s_tuner,
  #endif
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
    .tvnorms               = spct6100_tvnorms,
    .tvnormsize            = 2,
  #else
    .tvnorms               = V4L2_STD_PAL_M
                             | V4L2_STD_NTSC_M,
  #endif
    .current_norm          = V4L2_STD_PAL_M,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
    .ioctl_ops             = &spct6100_v4l2_live_ops,
#endif
  }
};

static struct video_device spct6100_v4l2_compress[NR_CARDS] = {
  {
    .name                  = "SPCT6100-V4L2-COMPRESS",
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
#define VID_TYPE_CAPTURE 1
    .vfl_type                  = VID_TYPE_CAPTURE,
#else
    .type                  = VID_TYPE_CAPTURE
      | VID_TYPE_MPEG_ENCODER,
#endif
    .fops                  = &spct6100_fops_compress,
    .minor                 = -1,
    .release               = spct6100_dummy_video_device_release,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
    /* EMPTY, implemented in spct6100_ioctl_common. */
#else
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
    .vidioc_enum_fmt_cap   = spct6100_vidioc_enum_fmt_cap,
    .vidioc_g_fmt_cap      = spct6100_vidioc_g_fmt_cap,
    .vidioc_try_fmt_cap    = spct6100_vidioc_try_fmt_cap,
    .vidioc_s_fmt_cap      = spct6100_vidioc_s_fmt_cap,
    .vidioc_querycap       = spct6100_vidioc_querycap,
    .vidioc_reqbufs        = spct6100_vidioc_reqbufs_compress,
    .vidioc_querybuf       = spct6100_vidioc_querybuf_compress,
    //.vidioc_querybuf       = spct6100_vidioc_querybuf_live,
    .vidioc_qbuf           = spct6100_vidioc_qbuf_compress,
    .vidioc_dqbuf          = spct6100_vidioc_dqbuf_compress,
    .vidioc_streamon       = spct6100_vidioc_streamon_compress,
    .vidioc_streamoff      = spct6100_vidioc_streamoff_compress,
    .vidioc_queryctrl      = spct6100_vidioc_queryctrl,
    .vidioc_g_ctrl         = spct6100_vidioc_g_ctrl,
    .vidioc_s_ctrl         = spct6100_vidioc_s_ctrl,
    .vidioc_enum_input     = spct6100_vidioc_enum_input,
    .vidioc_g_input        = spct6100_vidioc_g_input,
    .vidioc_s_input        = spct6100_vidioc_s_input,
    .vidioc_s_std          = spct6100_vidioc_s_std,
  #endif
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
    .tvnorms               = spct6100_tvnorms,
    .tvnormsize            = 2,
  #else
    .tvnorms               = V4L2_STD_PAL_M
                             | V4L2_STD_NTSC_M,
  #endif
    .current_norm          = V4L2_STD_PAL_M,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
    .ioctl_ops             = &spct6100_v4l2_comp_ops,
#endif
  }
};

/**
 * @brief PCI probe hook
 *
 *        The probe function should do the following task:
 *
 *          - Set the initial value of base register 3 to 0x1E000000
 *          - Ioremap the high reserved memory(for DMA buffer) into
 *           kernel logical space and get the kernel logical address,
 *            so kernel can access them
 *          - Set the initial value of base register 4 to the kernel
 *            logical address of the high reserved memory(for DMA buffer)
 *          - Check DMA capability
 *          - Register interrupt handler
 *
 * @return 0 on success, other error code on failure
 */
static int __devinit
spct6100_dvr_probe (struct pci_dev *pdev, const struct pci_device_id *ent)
{
#define CUR_DEV (g_spct6100_dev[g_device_num])
  int index;
  int rc, j, k;

  sprintf(CUR_DEV.name, "SPCT6100(LIVE/Compress)-%d", g_device_num);

  rc = spct6100_dvr_probe_common(pdev);
  if (rc != 0) {
    return rc;
  }

  if (video_register_device(CUR_DEV.v4l2_live,
                            VFL_TYPE_GRABBER, video_nr) != 0) {
    ERROR ("%s: Register live device fail\n", CUR_DEV.name);
    rc = -EIO;
    goto err_out;
  }

  if (video_register_device(CUR_DEV.v4l2_compress,
                            VFL_TYPE_GRABBER, video_nr) != 0) {
    ERROR ("%s: Register compress device fail\n", CUR_DEV.name);
    rc = -EIO;
    goto err_out;
  }

  j = CUR_DEV.index;

  for (k = 0; k < NR_CH; k++) {
    for (index = 0; index < NUM_V4L2_CID; index++)
      qctl_current_values[j][k][index] =
                 spct6100_qctrl[index].default_value;
  }

  for (k = 0; k < NR_CH; k++) {
    g_osdset[j][k].acolor = 1;
    g_osdset[j][k].timer_enable = 1;
    g_osdset[j][k].osd_enable = 1;
    g_osdset[j][k].osd_pos_x = 0x01a0;
    g_osdset[j][k].osd_pos_y = 0x000a;
    strcpy(g_osdset[j][k].str, "dummydummydummydummy");

    g_eqset[j][k].type = 2;
    g_eqset[j][k].qpi = 29;
    g_eqset[j][k].qpp = 30;
    g_eqset[j][k].qpb = 31;

    g_gopset[j][k].gopp = 30;
    g_gopset[j][k].gopb = 2;
  }

  if (spct6100_driver_download) {
    if (download_firmware(&CUR_DEV) != 0) {
      ERROR ("driver download firmware fail\n");
    } else {
      spct6100_pacdmp_write_bar3(&CUR_DEV, START_ADDRESS);
      spct6100_initenv(&CUR_DEV, 0, 0, 0);
    }
  }

  g_device_num++;

  return 0;

err_out:
  spct6100_dvr_free_resources(pdev);
  return rc;
}

/** @brief Driver init function
 *
 * @return
 */
static int __init
spct6100_init (void)
{
  int index;

  spct6100_qctrl[PREVIEW_INDEX].default_value = spct6100_preview_pci;

  for (index = 0; index < NR_CARDS; index++) {

    g_spct6100_dev[index].status = BOARD_STOP;
    g_spct6100_dev[index].stream_status = BOARD_STREAMOFF;
    g_spct6100_dev[index].index = -1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
    g_spct6100_dev[index].current_norm = V4L2_STD_PAL_M;
#endif
    sema_init(&(g_spct6100_dev[index].sem), 1);
    sema_init(&(g_spct6100_dev[index].status_read_sem), 1);
    sema_init(&(g_spct6100_dev[index].status_write_sem), 1);
    lstream_toggle[index] = 0;
    cstream_toggle[index] = 0;

    if (index != 0) {
      memcpy(spct6100_v4l2_live+index, spct6100_v4l2_live,
             sizeof(struct video_device));
      memcpy(spct6100_v4l2_compress+index, spct6100_v4l2_compress,
             sizeof(struct video_device));
    }
    g_spct6100_dev[index].v4l2_live = &spct6100_v4l2_live[index];
    g_spct6100_dev[index].v4l2_compress = &spct6100_v4l2_compress[index];
  }

  if (spct6100_debug) {
    spct6100_create_proc();
  }

  return pci_register_driver(&spct6100_dvr_pci_driver);
}

/** @brief Driver exit function
 *
 */
static void __exit
spct6100_cleanup (void)
{
  int temp = 0;
  int index = 0;
  unsigned char cpacket[CPACKET_LEN];
  u32 udata32;

  if (spct6100_debug)
    spct6100_remove_proc();

  for (temp = g_device_num - 1; temp >= 0; temp--) {

    struct pci_dev *pdev = g_spct6100_dev[temp].pci;
    spct6100_dev *cur_dev = &g_spct6100_dev[temp];
    int card_id = cur_dev->index;

    if (qtable_driver_initilized[card_id]) {
      if (spct6100_qtable_initialized[card_id] == 0) {
        fip[card_id].cq_len  = CPACKET_LEN;
        spct6100_get_reg(cur_dev, REG_CQ_BASE, CP(&(fip[card_id].cq_base)), -1);
        spct6100_get_reg(cur_dev, REG_CQ_PKTNUM, CP(&(fip[card_id].cq_num)), -1);
        fip[card_id].cur_cq_pos = fip[card_id].cq_base;
        fip[card_id].cq_end = fip[card_id].cq_base + CPACKET_LEN * fip[card_id].cq_num;

        spct6100_pacdmp_write_bar3(cur_dev, fip[card_id].cq_base);
        for (index = 0; index < fip[card_id].cq_num; index++) {
          //spct6100_readmem(cur_dev, index*fip[card_id].cq_len, cpacket, fip[card_id].cq_len);
          memset(cpacket, 0, CPACKET_LEN);
          CPACKET_OWNER(cpacket) = 0x86;
          spct6100_writemem(cur_dev, index*fip[card_id].cq_len, (u8 *)cpacket, fip[card_id].cq_len);
        }
        spct6100_pacdmp_write_bar3(cur_dev, 0x1E000000);
      }
      if (!spct6100_gpio_mode && cur_dev->stream_status == BOARD_STREAMON) {
        insert_cmd_node(cur_dev, pid_local_to_external(PID_STOP), 0, -1, NULL, NULL);
      }
      insert_cmd_node(cur_dev, pid_local_to_external(PID_END), 0, -1, NULL, NULL);
      msleep(500);
    }

    // Disable ALL VIC First
    udata32 = 0xFFFFFFFF;
    spct6100_set_reg(cur_dev, REG_INT_VIC1_ENABLE, (unsigned char *)&udata32, -1);
    spct6100_set_reg(cur_dev, REG_INT_VIC2_ENABLE, (unsigned char *)&udata32, -1);

    cur_dev->status = BOARD_STOP;
    del_timer(&(cur_dev->update_osd));

#ifdef __arm__
    if (spct6100_irq_shared) {
      if (temp == 0) {
        free_irq(pdev->irq, cur_dev);
      }
    } else {
      free_irq(pdev->irq, cur_dev);
    }
#else
    free_irq(pdev->irq, cur_dev);
#endif

    video_unregister_device(&spct6100_v4l2_live[card_id]);
    video_unregister_device(&spct6100_v4l2_compress[card_id]);

    if (spct6100_reserved_addr == 0) {
#ifdef __arm__
      iounmap (cur_dev->master_buf_logic);
#endif
      free_pages(cur_dev->master_free_pages, SPCT6100_DVR_ORDER);
    } else {
      iounmap (cur_dev->master_buf_logic);
    }

    pci_iounmap (pdev, cur_dev->mmio_vaddr);
    pci_disable_device(pdev);
    pci_release_region(pdev, 0);
    /* Ensure that the driver data is no longer valid. */
    pci_set_drvdata(pdev, NULL);
  }

  pci_unregister_driver(&spct6100_dvr_pci_driver);
}

/*=====================================================================*/
module_init (spct6100_init);
module_exit (spct6100_cleanup);
module_param(spct6100_preview_pci, int, S_IRUGO);
module_param(spct6100_cpacket_num, int, S_IRUGO);
module_param(spct6100_gpio_mode, int, S_IRUGO);
module_param(spct6100_reserved_addr, ulong, S_IRUGO);
module_param(spct6100_debug, int, S_IRUGO);
module_param(spct6100_process_expire, int, S_IRUGO);
module_param(spct6100_trace, int, S_IRUGO);
MODULE_DEVICE_TABLE(pci, spct6100_dvr_tbl);
MODULE_VERSION(SPCT6100_PCI_VERSION);
MODULE_AUTHOR ("Guang-Yuan.Wang/Cp.Lin/Shengguo.Tan");
MODULE_DESCRIPTION ("SPCT6100-V4L2-Driver");
MODULE_LICENSE("GPL");
