#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "lib.h"

#define OPTION_NUM 1

static int g_nTrace = 0;
static char *g_szOptions[OPTION_NUM];

static int
trace_log (const char *fmt, ...)
{
    if (g_nTrace) {
        int n;
        va_list vp;

        va_start (vp, fmt);
        n = vprintf (fmt, vp);
        va_end (vp);
        return n;
    }
    return 0;
}

static void
process_command_line (int argc, char *argv[])
{
    int i;
    int show_help = 0;
    int invalid_option = 0;
    int opt_idx = 0;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] != '-') {
            if (opt_idx < OPTION_NUM) {
                g_szOptions[opt_idx] = argv[i];
                opt_idx++;
                continue;
            }
            invalid_option = 1;
            break;
        }
        if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help")) {
            show_help = 1;
            break;
        } else if (!strcmp (argv[i], "-t") || !strcmp (argv[i], "--trace")) {
            g_nTrace = 1;
        } else {
            invalid_option = 1;
            break;
        }
    }

    if (show_help) {
        printf ("Usage: %s [OPTION] <config_file>\n", argv[0]);
        printf ("Init cards with <config_file>\n");
        printf ("OPTION:\n");
        printf ("-h,--help        show usage\n");
        printf ("-t,--trace       print trace log\n");
        exit (0);
    } else if (invalid_option) {
        printf ("%s: invalid option -- `%s`\n", argv[0], argv[i]);
        printf ("Try `%s --help' for more information.\n", argv[0]);
        exit (1);
    } else if (opt_idx != OPTION_NUM) {
        printf ("%s: necessary options missed\n", argv[0]);
        printf ("Try `%s --help' for more information.\n", argv[0]);
        exit (1);
    }
}

struct CardInfo {
    char live_name[32];
    int live_name_set;

    int live_status[4];
    int live_status_set[4];

    int live_showsize[4];
    int live_showsize_set[4];

    char comp_name[32];
    int comp_name_set;

    int comp_status[4];
    int comp_status_set[4];

    int comp_showsize[4];
    int comp_showsize_set[4];

    char std[4][32];
    int std_set[4];

    int audio[4];
    int audio_set[4];

    int motion[4];
    int motion_set[4];

    int vlost[4];
    int vlost_set[4];

    struct spct6100_gop gop[4];
    int gopp_set[4];
    int gopb_set[4];

    struct spct6100_equality equality[4];
    int minbitrate_set[4];
    int maxbitrate_set[4];
    int qpi_set[4];
    int qpp_set[4];
    int qpb_set[4];

    struct spct6100_cropset cropset[4];
    int start_x_set[4];
    int start_y_set[4];
    int width_set[4];
    int height_set[4];
};

static struct CardInfo card_info[4];

static int
is_spcaces (char ch)
{
    if (ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r' || ch == '\a') {
        return 1;
    }
    return 0;
}

static void
skip_spaces (char **str)
{
    while (is_spcaces (**str)) {
        (*str)++;
    }
}

static int
parse_line (char *str)
{
    int card = -1;
    int channel = -1;
    int type = -1;
    int i, j;
    int num = 0;

    skip_spaces (&str);
    if (*str == '#' || *str == '/' || *str == ';' || *str == 0) {
        return 0;
    }

    if (strncasecmp (str, "card", 4) == 0) {
        str += 4;
        if (*str >= '0' && *str <= '3') {
            card = *str - '0';
            str++;
        } else {
            return -1;
        }
        if (*str != '.') {
            return -1;
        }
        str++;
    }
    if (strncasecmp (str, "chan", 4) == 0) {
        str += 4;
        if (*str >= '0' && *str <= '3') {
            channel = *str - '0';
            str++;
        } else {
            return -1;
        }
        if (*str != '.') {
            return -1;
        }
        str++;
    }
    if (strncasecmp (str, "live", 4) == 0) {
        str += 4;
        type = 0;
        if (*str != '.') {
            return -1;
        }
        str++;
    }
    if (strncasecmp (str, "compress", 8) == 0) {
        str += 8;
        type = 1;
        if (*str != '.') {
            return -1;
        }
        str++;
    }

    if (strncasecmp (str, "name", 4) == 0) {
        if (card == -1 || type == -1) {
            return -1;
        }
        str += 4;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (type == 0) {
            i = 0;
            while (*str && !is_spcaces (*str)) {
                card_info[card].live_name[i++] = *str++;
                if (i == sizeof (card_info[card].live_name) - 1) {
                    break;
                }
            }
            card_info[card].live_name[i] = 0;
            card_info[card].live_name_set = 1;
        } else {
            i = 0;
            while (*str && !is_spcaces (*str)) {
                card_info[card].comp_name[i++] = *str++;
                if (i == sizeof (card_info[card].comp_name) - 1) {
                    break;
                }
            }
            card_info[card].comp_name[i] = 0;
            card_info[card].comp_name_set = 1;
        }
        return 0;
    }

    if (strncasecmp (str, "std", 3) == 0) {
        char *std;

        str += 3;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (strncasecmp (str, "PAL", 3) == 0) {
            std = "PAL";
        } else if (strncasecmp (str, "NTSC", 4) == 0) {
            std = "NTSC";
        } else {
            return -1;
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                strcpy (card_info[i].std[j], std);
                card_info[i].std_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "status", 6) == 0) {
        int status;

        str += 6;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (*str == '0' || *str == '1') {
            status = *str - '0';
            str++;
        } else {
            return -1;
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                if (type == 0) {
                    card_info[i].live_status[j] = status;
                    card_info[i].live_status_set[j] = 1;
                } else if (type == 1) {
                    card_info[i].comp_status[j] = status;
                    card_info[i].comp_status_set[j] = 1;
                } else {
                    card_info[i].live_status[j] = status;
                    card_info[i].live_status_set[j] = 1;
                    card_info[i].comp_status[j] = status;
                    card_info[i].comp_status_set[j] = 1;
                }
            }
        }
        return 0;
    }

    if (strncasecmp (str, "showsize", 8) == 0) {
        int showsize;

        str += 8;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (strncasecmp (str, "4CIF", 4) == 0) {
            showsize = 0;
            str += 4;
        } else if (strncasecmp (str, "CIF", 3) == 0) {
            showsize = 1;
            str += 3;
        } else if (strncasecmp (str, "QCIF", 4) == 0) {
            showsize = 2;
            str += 4;
        } else if (strncasecmp (str, "2CIF", 4) == 0) {
            showsize = 3;
            str += 4;
        } else {
            return -1;
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                if (type == 0) {
                    card_info[i].live_showsize[j] = showsize;
                    card_info[i].live_showsize_set[j] = 1;
                } else if (type == 1) {
                    card_info[i].comp_showsize[j] = showsize;
                    card_info[i].comp_showsize_set[j] = 1;
                } else {
                    card_info[i].live_showsize[j] = showsize;
                    card_info[i].live_showsize_set[j] = 1;
                    card_info[i].comp_showsize[j] = showsize;
                    card_info[i].comp_showsize_set[j] = 1;
                }
            }
        }
        return 0;
    }

    if (strncasecmp (str, "audio", 5) == 0) {
        int audio;

        str += 5;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (*str == '0' || *str == '1') {
            audio = *str - '0';
            str++;
        } else {
            return -1;
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].audio[j] = audio;
                card_info[i].audio_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "motion", 6) == 0) {
        int motion;

        str += 6;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (*str == '0' || *str == '1') {
            motion = *str - '0';
            str++;
        } else {
            return -1;
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].motion[j] = motion;
                card_info[i].motion_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "vlost", 5) == 0) {
        int vlost;

        str += 5;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (*str == '0' || *str == '1') {
            vlost = *str - '0';
            str++;
        } else {
            return -1;
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].vlost[j] = vlost;
                card_info[i].vlost_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "gopp", 4) == 0) {
        str += 4;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].gop[j].gopp = num;
                card_info[i].gopp_set[j] = 1;
            }
        }
        return 0;
    }
    if (strncasecmp (str, "gopb", 4) == 0) {
        str += 4;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].gop[j].gopb = num;
                card_info[i].gopb_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "minbitrate", 10) == 0) {
        str += 10;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].equality[j].minbitrate = num;
                card_info[i].minbitrate_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "maxbitrate", 10) == 0) {
        str += 10;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].equality[j].maxbitrate = num;
                card_info[i].maxbitrate_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "qpi", 3) == 0) {
        str += 3;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].equality[j].qpi = num;
                card_info[i].qpi_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "qpp", 3) == 0) {
        str += 3;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].equality[j].qpp = num;
                card_info[i].qpp_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "qpb", 3) == 0) {
        str += 3;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].equality[j].qpb = num;
                card_info[i].qpb_set[j] = 1;
            }
        }
        return 0;
    }

    if (strncasecmp (str, "start_x", 7) == 0) {
        str += 7;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].cropset[j].start_x = num;
                card_info[i].start_x_set[j] = 1;
            }
        }
        return 0;
    }
    if (strncasecmp (str, "start_y", 7) == 0) {
        str += 7;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].cropset[j].start_y = num;
                card_info[i].start_y_set[j] = 1;
            }
        }
        return 0;
    }
    if (strncasecmp (str, "width", 5) == 0) {
        str += 5;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].cropset[j].width = num;
                card_info[i].width_set[j] = 1;
            }
        }
        return 0;
    }
    if (strncasecmp (str, "height", 6) == 0) {
        str += 6;
        skip_spaces (&str);
        if (*str != '=') {
            return -1;
        }
        str++;
        skip_spaces (&str);
        if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
          sscanf (str, "%x", &num);
        } else {
          sscanf (str, "%d", &num);
        }
        for (i = 0; i < 4; i++) {
            if (card != -1 && card != i) {
               continue;
            }
            for (j = 0; j < 4; j++) {
                if (channel != -1 && channel != j) {
                    continue;
                }
                card_info[i].cropset[j].height = num;
                card_info[i].height_set[j] = 1;
            }
        }
        return 0;
    }

    return 0;
}

static int
init_devices (void)
{
    int card, chan;
    int fd;
    struct v4l2_capability cap;

    for (card = 0; card < 4; card++) {
        fd = -1;
        if (card_info[card].live_name_set) {
            fd = open (card_info[card].live_name, O_RDWR|O_SYNC);
            if (fd < 0) {
                trace_log ("Open device %s failed.\n",
                         card_info[card].live_name);
                continue;
            }
            memset (&cap, 0, sizeof (cap));
            if (ioctl (fd, VIDIOC_QUERYCAP, &cap) != 0) {
                trace_log ("%s: VIDIOC_QUERYCAP failed\n",
                         card_info[card].live_name);
                close (fd);
                continue;
            }
            if (strcmp ((char *)cap.card, "SPCT6100") != 0) {
                trace_log ("%s: error cap.card %s\n",
                         card_info[card].live_name, cap.card);
                close (fd);
                continue;
            }
            if (strcmp ((char *)cap.driver, "SPCT6100-V4L2-L") != 0) {
                trace_log ("%s: error cap.driver %s\n",
                         card_info[card].live_name, cap.driver);
                close (fd);
                continue;
            }
        }
        if (card_info[card].comp_name_set) {
            if (fd > 0) {
                close (fd);
            }
            fd = open (card_info[card].comp_name, O_RDWR|O_SYNC);
            if (fd < 0) {
                trace_log ("Open device %s failed.\n",
                         card_info[card].comp_name);
                continue;
            }
            memset (&cap, 0, sizeof (cap));
            if (ioctl (fd, VIDIOC_QUERYCAP, &cap) != 0) {
                trace_log ("%s: VIDIOC_QUERYCAP failed\n",
                         card_info[card].comp_name);
                close (fd);
                continue;
            }
            if (strcmp ((char *)cap.card, "SPCT6100") != 0) {
                trace_log ("%s: error cap.card %s\n",
                         card_info[card].comp_name, cap.card);
                close (fd);
                continue;
            }
            if (strcmp ((char *)cap.driver, "SPCT6100-V4L2-C") != 0) {
                trace_log ("%s: error cap.driver %s\n",
                         card_info[card].comp_name, cap.driver);
                close (fd);
                continue;
            }
        }
        if (fd < 0) {
            continue;
        }

        for (chan = 0; chan < 4; chan++) {
            int status;

            if (s_input (fd, chan) != 0) {
                break;
            }

            if (card_info[card].std_set[chan]) {
                trace_log ("card %d channel %d set std=%s ",
                         card, chan,
                         card_info[card].std[chan]);
                if (s_std (fd, card_info[card].std[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }
            if (card_info[card].audio_set[chan]) {
                trace_log ("card %d channel %d set audio=%d ",
                         card, chan,
                         card_info[card].audio[chan]);
                if (s_ctrl_AUDIO (fd, card_info[card].audio[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }
            if (card_info[card].motion_set[chan]) {
                trace_log ("card %d channel %d set motion=%d ",
                         card, chan,
                         card_info[card].motion[chan]);
                if (s_ctrl_MOTIONDETECT (fd, card_info[card].motion[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }
            if (card_info[card].vlost_set[chan]) {
                trace_log ("card %d channel %d set vlost=%d ",
                         card, chan,
                         card_info[card].vlost[chan]);
                if (s_ctrl_VLOSTEN (fd, card_info[card].vlost[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }

            status = 3;
            if (card_info[card].live_status_set[chan]) {
                status = (status & 0x1) | (card_info[card].live_status[chan] << 1);
            }
            if (card_info[card].comp_status_set[chan]) {
                status = (status & 0x2) | card_info[card].comp_status[chan];
            }
            if (card_info[card].live_status_set[chan]
                || card_info[card].comp_status_set[chan]) {
                trace_log ("card %d channel %d set status=%d ",
                         card, chan,
                         status);
                if (s_ctrl_STATUS (fd, status) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }

            if (card_info[card].live_showsize_set[chan]) {
                trace_log ("card %d channel %d set live_showsize=%d ",
                         card, chan,
                         card_info[card].live_showsize[chan]);
                if (s_ctrl_LSHOWSIZE (fd, card_info[card].live_showsize[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }

            if (card_info[card].comp_showsize_set[chan]) {
                trace_log ("card %d channel %d set comp_showsize=%d ",
                         card, chan,
                         card_info[card].comp_showsize[chan]);
                if (s_ctrl_CSHOWSIZE (fd, card_info[card].comp_showsize[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }

            if (card_info[card].gopp_set[chan]
                && card_info[card].gopb_set[chan]) {
                trace_log ("card %d channel %d set gopp=%d, gopb=%d ",
                         card, chan,
                         card_info[card].gop[chan].gopp,
                         card_info[card].gop[chan].gopb);
                if (s_ctrl_GOP (fd, &card_info[card].gop[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }

            if (card_info[card].minbitrate_set[chan]
                && card_info[card].maxbitrate_set[chan]) {
                card_info[chan].equality[chan].type = 0;
                trace_log ("card %d channel %d set minbitrate=%d, maxbitrate=%d ",
                         card, chan,
                         card_info[card].equality[chan].minbitrate,
                         card_info[card].equality[chan].maxbitrate);
                if (s_ctrl_EQUALITY (fd, &card_info[card].equality[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }
            if (card_info[card].qpi_set[chan] && card_info[card].qpp_set[chan]
                && card_info[card].qpb_set[chan]) {
                card_info[chan].equality[chan].type = 2;
                trace_log ("card %d channel %d set qpi=%d, qpp=%d, qpb=%d ",
                         card, chan,
                         card_info[card].equality[chan].qpi,
                         card_info[card].equality[chan].qpp,
                         card_info[card].equality[chan].qpb);
                if (s_ctrl_EQUALITY (fd, &card_info[card].equality[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }
            if (card_info[card].start_x_set[chan]
                && card_info[card].start_y_set[chan]
                && card_info[card].width_set[chan]
                && card_info[card].height_set[chan]) {
                trace_log ("card %d channel %d set start_x=0x%x, start_y=0x%x, width=0x%x, height=0x%x ",
                         card, chan,
                         card_info[card].cropset[chan].start_x,
                         card_info[card].cropset[chan].start_y,
                         card_info[card].cropset[chan].width,
                         card_info[card].cropset[chan].height);
                if (s_ctrl_CROPSET (fd, &card_info[card].cropset[chan]) != 0) {
                    trace_log ("failed\n");
                    break;
                }
                trace_log ("OK\n");
            }
        }
        close (fd);
    }
    return 0;
}

int
main (int argc, char *argv[])
{
    char *config_file_name;
    FILE *fp;
    char line_buf[1024];
    int line_num;
    int parse_error;

    process_command_line (argc, argv);
    config_file_name = g_szOptions[0];
    fp = fopen (config_file_name, "rb");
    if (fp == NULL) {
        trace_log ("Open file %s failed.\n", config_file_name);
        return -1;
    }

    line_num = 0;
    parse_error = 0;
    while (1) {
        fgets (line_buf, sizeof (line_buf), fp);
        if (feof (fp)) {
            break;
        }
        line_num++;
        if (parse_line (line_buf) != 0) {
            trace_log ("parse line %d error\n", line_num);
            parse_error = 1;
            break;
        }
    }
    fclose (fp);

    if (parse_error) {
        return -1;
    }
    return init_devices ();
}
