#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include "lib.h"

// ==== S_STD/G_STD ====
// szStd: "PAL"/"pal", "NTSC"/"ntsc"
int s_std (int fd, const char *szStd)
{
    v4l2_std_id std_id;
    int rc;

    if (szStd == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }

    if (strncasecmp (szStd, "PAL", 3) == 0) {
        std_id = V4L2_STD_PAL_M;
    } else if (strncasecmp (szStd, "NTSC", 4) == 0) {
        std_id = V4L2_STD_NTSC_M;
    } else {
        fprintf (stderr, "%s: Not support %s\n", __FUNCTION__, szStd);
        return -1;
    }

    rc = ioctl (fd, VIDIOC_S_STD, &std_id);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_STD rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_std (int fd, char *szStd)
{
    v4l2_std_id std_id;
    int rc;

    if (szStd == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }

    rc = ioctl (fd, VIDIOC_G_STD, &std_id);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_STD rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    switch (std_id) {
    case V4L2_STD_PAL_M:
        strcpy (szStd, "PAL-M");
        break;
    case V4L2_STD_NTSC_M:
        strcpy (szStd, "NTSC-M");
        break;
    default:
        fprintf (stderr, "%s: std 0x%016llx error\n", __FUNCTION__, std_id);
        return -1;
    }
    return 0;
}

// ==== S_INPUT/G_INPUT ====
// input: 0,1,2,3
int s_input (int fd, int input)
{
    int rc;

    rc = ioctl (fd, VIDIOC_S_INPUT, &input);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_INPUT rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_input (int fd, int *input)
{
    int rc;

    rc = ioctl (fd, VIDIOC_G_INPUT, input);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_INPUT rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

// ==== S_FMT/G_FMT ====
int s_fmt (int fd, int width, int height, const char *szFmt)
{
    struct v4l2_format format;
    int rc;

    memset (&format, 0, sizeof (format));
    format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    format.fmt.pix.width = width;
    format.fmt.pix.height = height;
    if (strncasecmp (szFmt, "UYVY", 4) == 0) {
        format.fmt.pix.pixelformat = V4L2_PIX_FMT_UYVY;
    } else if (strncasecmp (szFmt, "YUV420", 6) == 0
               || strncasecmp (szFmt, "IYUV", 4) == 0) {
        format.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
    }

    rc = ioctl (fd, VIDIOC_S_FMT, &format);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_FMT rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_fmt (int fd, int *width, int *height, char *szFmt)
{
    struct v4l2_format format;
    int rc;

    memset (&format, 0, sizeof (format));
    format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    rc = ioctl (fd, VIDIOC_G_FMT, &format);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_FMT rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }

    switch (format.fmt.pix.pixelformat) {
    case V4L2_PIX_FMT_UYVY:
        if (szFmt != NULL) {
            strcpy (szFmt, "UYVY422");
        }
        break;
    case V4L2_PIX_FMT_YUV420:
        if (szFmt != NULL) {
            strcpy (szFmt, "YUV420P");
        }
        break;
    default:
        fprintf (stderr, "%s: fmt 0x%08x error\n",
                 __FUNCTION__, format.fmt.pix.pixelformat);
        return -1;
    }
    if (width != NULL) {
        *width = format.fmt.pix.width;
    }
    if (height != NULL) {
        *height = format.fmt.pix.height;
    }
    return 0;
}

// ==== STREAMON/STREAMOFF ====
int stream_on (int fd)
{
    enum v4l2_buf_type type;
    int rc;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    rc = ioctl (fd, VIDIOC_STREAMON, &type);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_STREAMON rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int stream_off (int fd)
{
    enum v4l2_buf_type type;
    int rc;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    rc = ioctl (fd, VIDIOC_STREAMOFF, &type);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_STREAMOFF rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

// ==== S_CTRL/G_CTRL ====
#define S_CTRL_FUNCTION(CMD_ID, MIN_VALUE, MAX_VALUE, NAME) \
int s_ctrl_##NAME (int fd, int value) \
{  \
    struct v4l2_control control; \
    int rc; \
    control.id = CMD_ID; \
    control.value = value; \
    rc = ioctl (fd, VIDIOC_S_CTRL, &control); \
    if (rc != 0) { \
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc); \
        fprintf (stderr, "%s\n", strerror (errno)); \
        return -1; \
    } \
    return 0; \
}

S_CTRL_FUNCTION(V4L2_CID_MODE, 0, 1, MODE)
S_CTRL_FUNCTION(V4L2_CID_STATUS, 0, 3, STATUS)
S_CTRL_FUNCTION(V4L2_CID_EENSPEED, 1, 30, EENSPEED)
S_CTRL_FUNCTION(V4L2_CID_BRIGHTNESS, 0, 255, BRIGHTNESS)
S_CTRL_FUNCTION(V4L2_CID_CONTRAST, 0, 255, CONTRAST)
S_CTRL_FUNCTION(V4L2_CID_SATURATION, 0, 255, SATURATION)
S_CTRL_FUNCTION(V4L2_CID_HUE, 0, 255, HUE)
S_CTRL_FUNCTION(V4L2_CID_VLOSTEN, 0, 1, VLOSTEN)
S_CTRL_FUNCTION(V4L2_CID_LSHOWSIZE, 0, 3, LSHOWSIZE)
S_CTRL_FUNCTION(V4L2_CID_CSHOWSIZE, 0, 3, CSHOWSIZE)
S_CTRL_FUNCTION(V4L2_CID_MOTIONDETECT, 0, 1, MOTIONDETECT)
S_CTRL_FUNCTION(V4L2_CID_MTHRESHOLD, 0, 0x7fffffff, MTHRESHOLD)
S_CTRL_FUNCTION(V4L2_CID_MINTERVAL, 0, 0x7fffffff, MINTERVAL)
S_CTRL_FUNCTION(V4L2_CID_PREVIEW, 0, 1, PREVIEW)
S_CTRL_FUNCTION(V4L2_CID_AUDIO, 0, 1, AUDIO)
S_CTRL_FUNCTION(V4L2_CID_TIMEOUT, 0, 0xffffffff, TIMEOUT)
S_CTRL_FUNCTION(V4L2_CID_REQIFRAME, 0, 4, REQIFRAME)
S_CTRL_FUNCTION(V4L2_CID_REQISEQ, 1, 0x7fffffff, REQISEQ)
S_CTRL_FUNCTION(V4L2_CID_TIMESET, 0, 0xffffffff, TIMESET)

int s_ctrl_GOP (int fd, struct spct6100_gop *gop)
{
    struct v4l2_control control;
    int rc;

    if (gop == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_GOPSET;
    control.value = (int)gop;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_OSD (int fd, struct spct6100_osd *osd)
{
    struct v4l2_control control;
    int rc;

    if (osd == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_OSDSET;
    control.value = (int)osd;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_MOSAIC (int fd, struct spct6100_mosaic *mosaic)
{
    struct v4l2_control control;
    int rc;

    if (mosaic == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_MOSAIC;
    control.value = (int)mosaic;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_EQUALITY (int fd, struct spct6100_equality *equality)
{
    struct v4l2_control control;
    int rc;

    control.id = V4L2_CID_EQUALITY;
    control.value = (int)equality;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_GPIOD (int fd, struct spct6100_gpio *gpio)
{
    struct v4l2_control control;
    int rc;

    control.id = V4L2_CID_GPIOD;
    control.value = (int)gpio;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_CROPSET (int fd, struct spct6100_cropset *cropset)
{
    struct v4l2_control control;
    int rc;

    if (cropset == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_CROPSET;
    control.value = (int)cropset;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_WGPIO (int fd, struct spct6100_gpio *gpio)
{
    struct v4l2_control control;
    int rc;

    if (gpio == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_WGPIO;
    control.value = (int)gpio;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_RGPIO (int fd, struct spct6100_gpio *gpio)
{
    struct v4l2_control control;
    int rc;

    if (gpio == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_RGPIO;
    control.value = (int)gpio;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_DOWNLOAD (int fd, unsigned int dsp_addr, char *buf, int buf_size)
{
    struct v4l2_control control;
    struct spct6100_memory memory;
    int rc;

    if (buf == NULL || buf_size <= 0) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }

    memory.dsp_addr = dsp_addr;
    memory.buf = buf;
    memory.buf_size = buf_size;

    control.id = V4L2_CID_DOWNLOAD;
    control.value = (int)&memory;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_RUNCODE (int fd, unsigned int dsp_addr)
{
    struct v4l2_control control;
    int rc;

    control.id = V4L2_CID_RUNCODE;
    control.value = (int)dsp_addr;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_INITENV (int fd)
{
    struct v4l2_control control;
    int rc;

    control.id = V4L2_CID_INITENV;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

/*
int s_ctrl_WI2C (int fd, struct spct6100_eeprom *eeprom)
{
    struct v4l2_control control;
    int rc;

    if (eeprom == NULL || eeprom->buf == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_WI2C;
    control.value = (int)eeprom;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int s_ctrl_RI2C (int fd, struct spct6100_eeprom *eeprom)
{
    struct v4l2_control control;
    int rc;

    if (eeprom == NULL || eeprom->buf == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_RI2C;
    control.value = (int)eeprom;
    rc = ioctl (fd, VIDIOC_S_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}
*/

#define G_CTRL_FUNCTION(CMD_ID, MIN_VALUE, MAX_VALUE, NAME) \
int g_ctrl_##NAME (int fd, int *value) \
{ \
    struct v4l2_control control; \
    int rc; \
    if (value == NULL) { \
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__); \
        return -1; \
    } \
    control.id = CMD_ID; \
    control.value = -1; \
    rc = ioctl (fd, VIDIOC_G_CTRL, &control); \
    if (rc != 0) { \
        fprintf (stderr, "%s: VIDIOC_G_CTRL rc = %d\n", __FUNCTION__, rc); \
        fprintf (stderr, "%s\n", strerror (errno)); \
        return -1; \
    } \
    if ((unsigned int)control.value < MIN_VALUE \
        || (unsigned int)control.value > MAX_VALUE) { \
        fprintf (stderr, "%s: %u out of range [%u, %u]\n", \
                 __FUNCTION__, control.value, MIN_VALUE, MAX_VALUE); \
        return -1; \
    } \
    *value = control.value; \
    return 0; \
}

G_CTRL_FUNCTION(V4L2_CID_MODE, 0, 1, MODE)
G_CTRL_FUNCTION(V4L2_CID_STATUS, 0, 3, STATUS)
G_CTRL_FUNCTION(V4L2_CID_EENSPEED, 1, 30, EENSPEED)
G_CTRL_FUNCTION(V4L2_CID_BRIGHTNESS, 0, 255, BRIGHTNESS)
G_CTRL_FUNCTION(V4L2_CID_CONTRAST, 0, 255, CONTRAST)
G_CTRL_FUNCTION(V4L2_CID_SATURATION, 0, 255, SATURATION)
G_CTRL_FUNCTION(V4L2_CID_HUE, 0, 255, HUE)
G_CTRL_FUNCTION(V4L2_CID_VLOSTEN, 0, 1, VLOSTEN)
G_CTRL_FUNCTION(V4L2_CID_LSHOWSIZE, 0, 3, LSHOWSIZE)
G_CTRL_FUNCTION(V4L2_CID_CSHOWSIZE, 0, 3, CSHOWSIZE)
G_CTRL_FUNCTION(V4L2_CID_MOTIONDETECT, 0, 1, MOTIONDETECT)
G_CTRL_FUNCTION(V4L2_CID_MTHRESHOLD, 0, 0x7fffffff, MTHRESHOLD)
G_CTRL_FUNCTION(V4L2_CID_MINTERVAL, 0, 0x7fffffff, MINTERVAL)
G_CTRL_FUNCTION(V4L2_CID_PREVIEW, 0, 1, PREVIEW)
G_CTRL_FUNCTION(V4L2_CID_AUDIO, 0, 1, AUDIO)
G_CTRL_FUNCTION(V4L2_CID_TIMEOUT, 0, 0xffffffff, TIMEOUT)
G_CTRL_FUNCTION(V4L2_CID_CARDNUM, 0, 0xffffffff, CARDNUM)
G_CTRL_FUNCTION(V4L2_CID_DSPSTATUS, 0, 0xffffffff, DSPSTATUS)

int g_ctrl_GOP (int fd, struct spct6100_gop *gop)
{
    struct v4l2_control control;
    int rc;

    if (gop == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_GOPSET;
    control.value = (int)gop;
    rc = ioctl (fd, VIDIOC_G_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_ctrl_OSD (int fd, struct spct6100_osd *osd)
{
    struct v4l2_control control;
    int rc;

    if (osd == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_OSDSET;
    control.value = (int)osd;
    rc = ioctl (fd, VIDIOC_G_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_ctrl_MOSAIC (int fd, struct spct6100_mosaic *mosaic)
{
    struct v4l2_control control;
    int rc;

    if (mosaic == NULL) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }
    control.id = V4L2_CID_MOSAIC;
    control.value = (int)mosaic;
    rc = ioctl (fd, VIDIOC_G_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_ctrl_EQUALITY (int fd, struct spct6100_equality *equality)
{
    struct v4l2_control control;
    int rc;

    control.id = V4L2_CID_EQUALITY;
    control.value = (int)equality;
    rc = ioctl (fd, VIDIOC_G_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_ctrl_GPIOD (int fd, struct spct6100_gpio *gpio)
{
    struct v4l2_control control;
    int rc;

    control.id = V4L2_CID_GPIOD;
    control.value = (int)gpio;
    rc = ioctl (fd, VIDIOC_G_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_G_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}

int g_ctrl_UPLOAD (int fd, unsigned int dsp_addr, char *buf, int buf_size)
{
    struct v4l2_control control;
    struct spct6100_memory memory;
    int rc;

    if (buf == NULL || buf_size <= 0) {
        fprintf (stderr, "%s: Bad argument\n", __FUNCTION__);
        return -1;
    }

    memory.dsp_addr = dsp_addr;
    memory.buf = buf;
    memory.buf_size = buf_size;

    control.id = V4L2_CID_UPLOAD;
    control.value = (int)&memory;
    rc = ioctl (fd, VIDIOC_G_CTRL, &control);
    if (rc != 0) {
        fprintf (stderr, "%s: VIDIOC_S_CTRL rc = %d\n", __FUNCTION__, rc);
        fprintf (stderr, "%s\n", strerror (errno));
        return -1;
    }
    return 0;
}
