#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include "lib.h"

#define OPTION_NUM 3

static int g_nCheck = 0;
static int g_nTrace = 0;
static char *g_szOptions[OPTION_NUM];

#define BUFFER_SIZE 4096
static char g_strDownBuf[BUFFER_SIZE];
static char g_strUpBuf[BUFFER_SIZE];

static int
trace_log (const char *fmt, ...)
{
    if (g_nTrace) {
        int n;
        va_list vp;

        va_start (vp, fmt);
        n = vprintf (fmt, vp);
        va_end (vp);
        return n;
    }
    return 0;
}

static void
process_command_line (int argc, char *argv[])
{
    int i;
    int show_help = 0;
    int invalid_option = 0;
    int opt_idx = 0;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] != '-') {
            if (opt_idx < OPTION_NUM) {
                g_szOptions[opt_idx] = argv[i];
                opt_idx++;
                continue;
            } 
            invalid_option = 1;
            break;
        }
        if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help")) {
            show_help = 1;
            break;
        } else if (!strcmp (argv[i], "-t") || !strcmp (argv[i], "--trace")) {
            g_nTrace = 1;
        } else if (!strcmp (argv[i], "-c") || !strcmp (argv[i], "--check")) {
            g_nCheck = 1;
        } else {
            invalid_option = 1;
            break;
        }
    }

    if (show_help) {
        printf ("Usage: %s [OPTION] <DEVICE> <ADDRESS> <FILE>\n", argv[0]);
        printf ("Download <FILE> to <DEVICE> <ADDRESS>\n");
        printf ("OPTION:\n");
        printf ("-h,--help        show usage\n");
        printf ("-t,--trace       print trace log\n");
        printf ("-c,--check       download then upload and comprare them\n");
        exit (0);
    } else if (invalid_option) {
        printf ("%s: invalid option -- `%s`\n", argv[0], argv[i]);
        printf ("Try `%s --help' for more information.\n", argv[0]);
        exit (1);
    } else if (opt_idx != OPTION_NUM) {
        printf ("%s: necessary options missed\n", argv[0]);
        printf ("Try `%s --help' for more information.\n", argv[0]);
        exit (1);
    }
}

static int
download (int fd, unsigned int memaddr, FILE *fp, unsigned int insize)
{
    unsigned int down_size, left_size, nbytes;

    if (insize == 0) {
        return 0;
    }
    left_size = insize;
    while (left_size > 0) {
        if (left_size < sizeof (g_strDownBuf)) {
            down_size = left_size;
        } else {
            down_size = sizeof (g_strDownBuf);
        }
        nbytes = fread (g_strDownBuf, 1, down_size, fp);
        if (nbytes != down_size) {
            trace_log ("fread fail\n");
            return -1;
        }

        if (s_ctrl_DOWNLOAD (fd, memaddr, g_strDownBuf, down_size) != 0) {
            trace_log ("s_ctrl_DOWNLOAD fail\n");
            return -1;
        }

        if (g_nCheck) {
            if (g_ctrl_UPLOAD (fd, memaddr, g_strUpBuf, down_size) != 0) {
                trace_log ("g_ctrl_UPLOAD fail\n");
                return -1;
            }
            if (memcmp (g_strUpBuf, g_strDownBuf, down_size) != 0) {
                trace_log ("download check fail\n");
                return -1;
            }
        }

        left_size -= down_size;
        memaddr += down_size;
    }
    if (g_nCheck) {
        trace_log ("download size %d, check pass\n", insize);
    }
    return 0;
}

static int
LoadDSPCode (int fd, char *szFilename, unsigned long addr)
{
    FILE *fp;
    int fsize;

    fp = fopen (szFilename, "rb");
    if (fp == NULL) {
        trace_log("Open H.264 decoder binary file %s fail!!\n", szFilename);
        return -1;
    }

    fseek (fp, 0L, SEEK_END);
    fsize = ftell (fp);
    fseek (fp, 0L, SEEK_SET);
    trace_log ("File=%s, size=%d, address=0x%08lx\n", szFilename, fsize, addr);
    if (download (fd, addr, fp, fsize) != 0) {
        trace_log ("download %s fail\n", szFilename);
        return -1;
    }
    fclose (fp);

    return 0;
}

int
main (int argc, char *argv[])
{
    char *device_name;
    unsigned long address;
    char *file_name;
    int fd;
    int rc;

    process_command_line (argc, argv);
    device_name = g_szOptions[0];
    if (strncmp (g_szOptions[1], "0x", 2) == 0) {
        sscanf (g_szOptions[1], "%lx", &address);
    } else {
        sscanf (g_szOptions[1], "%lu", &address);
    }
    file_name = g_szOptions[2];

    trace_log ("Write %s to %s <0x%08lx>\n",
               file_name, device_name, address);
    fd = open (device_name, O_RDWR|O_SYNC);
    if (fd < 0) {
        trace_log ("Open device: %s failed\n", device_name);
        return -1;
    }
    rc = LoadDSPCode (fd, file_name, address);
    close (fd);
    return rc;
}
